//---------------------------------------------------------------------------
/*
ADPCM Program for Windows

 Copyright 2020 by Mark DiVecchio

    This file is part of Remote Train Control.

    Remote Train Control is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Remote Train Control is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Remote Train Control.  If not, see <http://www.gnu.org/licenses/>.

*/
// Home Page: http://www.silogic.com/trains/ADPCM.html
//---------------------------------------------------------------------------
// Borland C++ Builder v4.0 version
//---------------------------------------------------------------------------
#ifndef ADPCM_PlayerH
#define ADPCM_PlayerH
//---------------------------------------------------------------------------
#include <Classes.hpp>
#include <Controls.hpp>
#include <StdCtrls.hpp>
#include <Forms.hpp>
#include <Dialogs.hpp>
#include <Grids.hpp>
#include <ExtCtrls.hpp>
#include <ComCtrls.hpp>
#include <Registry.hpp>
#include <Menus.hpp>
#include <Graphics.hpp>
//---------------------------------------------------------------------------
// Define NYBBLEADDRESSES to see the actual nybble addresses in the index.
// Only [Analyze] is supported, not playing or editing.
// This was done only to see the actual nybble address.
// and should not be used in released versions
//#define NYBBLEADDRESSES

#ifdef NYBBLEADDRESSES
// DEBUG used to be the actual nybble addresses in the index
#define EMPTY_FIELD 0x00FFFFFF
#else
#define EMPTY_FIELD 0x007FFFFF
#endif

#if __BORLANDC__ < 0x720
// types missing in BCB4.0
typedef short int int16_t;
typedef int int32_t;
typedef unsigned char uint8_t;
typedef char int8_t;
#endif
//---------------------------------------------------------------------------
// start of the Quill index in the sound file
#define QUILL_INDEX 0x3200
//---------------------------------------------------------------------------

// Overall, clips play better without CLICKING defined
//#define CLICKING

#ifdef CLICKING
struct ImaState {
   short step;    		// Index into step size table (can go negative)
   int predValue;		// Predicted sample value
   };
#else
struct ImaState {
   short step;    		// Index into step size table (can go negative)
   short predValue;		// Predicted sample value
   };
#endif
//---------------------------------------------------------------------------

class TMyDragObject : public TDragObject {
public:
	int SourceClipNo;
	unsigned char IndexSource[16];
	String SourceFilename;
	__fastcall TMyDragObject(void);
	};

// one Sound File Index entry
struct SFIndex {
	int Clip;
	int StartAddr;
	int StartRepAddr;
	int EndRepAddr;
	int SMInit;
	int EndAddr;
	unsigned long ClipLength;
	unsigned long ClipRepLength;

	int databyte;
    //
	byte SA_LOBit;
	byte ERA_LOBit;
	byte SRA_LOBit;
	byte EA_LOBit;
	//
	int DiskFileLength;
	};

// Quill Index structure
struct QIndex {
	int Clip;
	int dataword;   // unknown function
	int StartAddr;  // Starting Address
	byte B7; 		// always 0x00
	int EndAddr;    // Ending Address
	byte B11; 		// always 0x80
	byte B12; 		// always 0x00
	int SMInit;		// ADPCM State Machine Initialization
	};

struct SRSetting {
	int index;
	TCHAR Hz[10];
	TCHAR Kz[10];
	int Value;
	};

#include "Edit_the_Index.h"
#include "Quilling_Whistle.h"

// Clip Index String Grid
#define ICLIPNO 0
#define ISTARTADDR 1
#define IENDREPADDR 2
#define ISTARTREPADDR 3
#define ISMINIT 4
#define IENDADDR 5
#define ILENGTH 6
#define IDATABYTE 7

// Quill Index String Grid
#define QCLIPNO 0
#define QDATAWORD 1   		// first four bytes
#define QSTARTADDR 2
#define QENDADDR 3
#define QSMINIT 4
#define QLENGTH 5

#define ADPCM 0
#define PCM24 1
#define PCM24MASK 0x80
#define PCM16 2
#define EMPTY_INDEX 3

#define PS2 0
#define PS3 1

#define HZ5512 0
#define HZ8000 1
#define HZ11025 2
#define HZ16000 3
#define HZ22050 4
#define HZ44100 5

#define SPLASH_OFF  0
#define SPLASH_ON  1

#define EngineSound 0
#define AccentSound 1
#define HornWhistleSound 2
#define BellSound 3

#define IDIGITSTODISPLAY 6 		// number if digits to display in address fields
#define IOFFSETDIGITSTODISPLAY 4	// number of digits to display in offset field
#define HEXFIELDPREFIX TEXT("0x")

#define DEBUGGING (ADPCM_Main->bDebug > 0)

//---------------------------------------------------------------------------
class TADPCM_Main : public TForm
{
__published:	// IDE-managed Components
	TMenuItem *AboutADPCM;
	TMenuItem *AboutADPCM1;
	TMenuItem *ADPCMWebPage1;
	TMenuItem *Bell_Index;
	TMenuItem *CopyClipIndextoClipboard1;
	TMenuItem *CopyRowtoClipboard;
	TMenuItem *DecodeandPlayClip;
	TMenuItem *PFA_Index;
	TMenuItem *Clickity_Clack_Index;
	TMenuItem *PlayRepeatingSegmentofClip1;
	TMenuItem *ToggleOffsetClipNo;
	TMenuItem *Toot_Index;
	TOpenDialog *OpenDialog;
	TPanel *AnalyzePanel;
	TPanel *ButtonPanel;
	TPanel *FileNamePanel;
	TPopupMenu *ADPCMPopupMenu;
	TPopupMenu *IndexPopupMenu;
	TLabel *Label3;
	TButton *BrowseforFileButton;
	TEdit *ClipEdit;
	TLabel *Label1;
	TButton *PlayEntireFileButton;
	TButton *StopButton2;
	TRadioGroup *PSGroup;
	TRadioGroup *FormatGroup;
	TMemo *DebugBuffer;
	TRadioGroup *SampleGroup;
	TCheckBox *AutoCheckBox;
	TPopupMenu *MessagesPopupMenu;
	TMenuItem *MessagesCopytoClipboard;
	TMenuItem *Clear1;
	TMenuItem *AboutADPCM2;
	TButton *AnalyzeButton;
	TButton *ClearButton;
	TLabel *SoundCategory;
	TLabel *Label4;
	TEdit *VolumeEdit;
	TUpDown *VolumeUpDown;
	TStringGrid *StringGrid1;
	TMemo *MemoBuffer;
	TLabel *Label6;
	TButton *RestartButton;
	TButton *PauseButton;
	TButton *PlayClipButton;
	TButton *StopButton1;
	TButton *GetDevCapsButton;
	TEdit *Rep;
	TUpDown *RepUpDown;
	TButton *LongPlayButton;
	TButton *PlayRepeatingSegmentButton;
	TImage *Copyright;
	TLabel *VersionLabel;
	TButton *ExitButton;
	TButton *ExportOneCliptomp3;
	TMenuItem *ClipInfo;
	TLabel *Label5;
	TMenuItem *Export1Cliptomp3;
	TButton *Mod3Button;
	TMenuItem *EditClip;
	TMenuItem *EngineSounds_Index;
	TMenuItem *EditIndex;
	TMenuItem *SFS_Index;
	TMenuItem *MinimizeFileLen;
	TMenuItem *N1;
	TMenuItem *N2;
	TMenuItem *N3;
	TMenuItem *N4;
	TButton *EditIndexButton;
	TButton *EditClipButton;
	TMenuItem *SXS_Index;
	TMenuItem *EditIndexnewversion1;
	TMenuItem *EditSoftKeys;
	TButton *EditSoftkeysButton;
	TMenuItem *CopyIndex;
	TMenuItem *ClearthisIndex;
	TMenuItem *NoSoundFileSizeChecks;
	TLabel *EngineEEPROMSize;
	TMenuItem *AvailableTimeRemaining;
	TMenuItem *DumpClip1;
	TMenuItem *DumpEntireCliptoDisk;
	TMenuItem *DumpRepeatingSegmentofCliptoDisk;
	TMenuItem *ExportAllClipstomp3;
	TButton *QuillButton;
	TComboBox *MRUComboBox;
	TMenuItem *ExportAllClipsforDFPlayer1;
	TPopupMenu *FileNamePopupMenu;
	TMenuItem *RemoveFromList;
	TMenuItem *TruncateSoundFileforFasterDownloading;
	TMenuItem *N5;
	TMenuItem *ChangePCMClipVolume1;
	TMenuItem *EdittheScaleFactor1;
	TMenuItem *CleartheList1;
	TLabel *Hidden;
	TMenuItem *EditEngineParameters1;
	TButton *EditParamsButton;
	TMenuItem *Copylower256bytes;
	TRadioButton *Overflow;

	void __fastcall PlayEntireFileButtonClick(TObject *Sender);
	void __fastcall AnalyzeButtonClick(TObject *Sender);
	void __fastcall BrowseforFileButtonClick(TObject *Sender);
	void __fastcall StopButton2Click(TObject *Sender);
	void __fastcall PlayClipButtonClick(TObject *Sender);
	void __fastcall PauseButtonClick(TObject *Sender);
	void __fastcall RestartButtonClick(TObject *Sender);
	void __fastcall StringGrid1DblClick(TObject *Sender);
	void __fastcall GetDevCapsButtonClick(TObject *Sender);
	void __fastcall ExitButtonClick(TObject *Sender);
	void __fastcall StringGrid1Click(TObject *Sender);
	void __fastcall PlayRepeatingSegmentButtonClick(TObject *Sender);
	void __fastcall CopyRowtoClipboardClick(TObject *Sender);
	void __fastcall VolumeUpDownClick(TObject *Sender, TUDBtnType Button);
	void __fastcall AutoCheckBoxClick(TObject *Sender);
	void __fastcall ClearButtonClick(TObject *Sender);
	void __fastcall StringGrid1KeyPress(TObject *Sender, char &Key);
	void __fastcall CopyClipIndextoClipboard1Click(TObject *Sender);
	void __fastcall MessagesCopytoClipboardClick(TObject *Sender);
	void __fastcall AboutADPCMClick(TObject *Sender);
	void __fastcall ADPCMWebPage1Click(TObject *Sender);
	void __fastcall FormCreate(TObject *Sender);
   	void __fastcall StringGrid1FixedCellClick(TObject *Sender, int ACol, int ARow);
	void __fastcall ToggleOffsetClipNoClick(TObject *Sender);
	void __fastcall StringGrid1DrawCell(TObject *Sender, int ACol, int ARow, TRect &Rect,
          TGridDrawState State);
	void __fastcall StringGrid1SelectCell(TObject *Sender, int ACol, int ARow, bool &CanSelect);
	void __fastcall LongPlayButtonClick(TObject *Sender);
	void __fastcall Bell_IndexClick(TObject *Sender);
	void __fastcall Clear1Click(TObject *Sender);
	void __fastcall Toot_IndexClick(TObject *Sender);
	void __fastcall PFA_IndexClick(TObject *Sender);
	void __fastcall Clickity_Clack_IndexClick(TObject *Sender);
	void __fastcall ClipInfoClick(TObject *Sender);
	void __fastcall ExportOneCliptomp3Click(TObject *Sender);
	void __fastcall Export1Cliptomp3Click(TObject *Sender);
	void __fastcall Mod3ButtonClick(TObject *Sender);
	void __fastcall EditClipClick(TObject *Sender);
	void __fastcall EngineSounds_IndexClick(TObject *Sender);
	void __fastcall EditIndexClick(TObject *Sender);
	void __fastcall PSGroupClick(TObject *Sender);
	void __fastcall SFS_IndexClick(TObject *Sender);
	void __fastcall MinimizeFileLenClick(TObject *Sender);
	void __fastcall CopyrightClick(TObject *Sender);
	void __fastcall SXS_IndexClick(TObject *Sender);
	void __fastcall EditIndexButtonClick(TObject *Sender);
	void __fastcall EditSoftKeysClick(TObject *Sender);
	void __fastcall CopyIndexClick(TObject *Sender);
	void __fastcall ClearthisIndexClick(TObject *Sender);
	void __fastcall NoSoundFileSizeChecksClick(TObject *Sender);
	void __fastcall AvailableTimeRemainingClick(TObject *Sender);
	void __fastcall DumpEntireCliptoDiskClick(TObject *Sender);
	void __fastcall DumpRepeatingSegmentofCliptoDiskClick(TObject *Sender);
	void __fastcall ExportAllClipstomp3Click(TObject *Sender);
	void __fastcall StringGrid1MouseDown(TObject *Sender, TMouseButton Button, TShiftState Shift,
          int X, int Y);
	void __fastcall StringGrid1DragOver(TObject *Sender, TObject *Source, int X, int Y,
          TDragState State, bool &Accept);
	void __fastcall StringGrid1DragDrop(TObject *Sender, TObject *Source, int X, int Y);
	void __fastcall StringGrid1EndDrag(TObject *Sender, TObject *Target, int X, int Y);
	void __fastcall StringGrid1StartDrag(TObject *Sender, TDragObject *&DragObject);
	void __fastcall QuillButtonClick(TObject *Sender);
	void __fastcall StringGrid1MouseMove(TObject *Sender, TShiftState Shift, int X,
          int Y);
	void __fastcall FormDestroy(TObject *Sender);
	void __fastcall ExportAllClipsforDFPlayer1Click(TObject *Sender);
	void __fastcall RemoveFromListClick(TObject *Sender);
	void __fastcall TruncateSoundFileforFasterDownloadingClick(TObject *Sender);
	void __fastcall MRUComboBoxChange(TObject *Sender);
	void __fastcall ChangePCMClipVolume1Click(TObject *Sender);
	void __fastcall EdittheScaleFactor1Click(TObject *Sender);
	void __fastcall CleartheList1Click(TObject *Sender);
	void __fastcall HiddenDblClick(TObject *Sender);
	void __fastcall PlayClipButtonMouseEnter(TObject *Sender);
	void __fastcall EditEngineParameters1Click(TObject *Sender);
	void __fastcall EditParamsButtonClick(TObject *Sender);
	void __fastcall Copylower256bytesClick(TObject *Sender);

private:	// User declarations
	bool Decoded;
	void virtual __fastcall WMDropFiles(TWMDropFiles &message);
	AnsiString ProgramName;
	void __fastcall ExportClip(int index, bool DFPlayerFormat);
	TMyDragObject * tm;
	void __fastcall EditOneByte(TObject *Sender);
	bool IndexFormat;
	bool ExportOneClipFlag;

public:		// User declarations
	__fastcall TADPCM_Main(TComponent* Owner);
	__fastcall ~TADPCM_Main();
	void __fastcall getFileDetails(HANDLE hFile);
	void __fastcall ClearStringGrid(void);
	void __fastcall MyShowMessage(String Msg, bool Popup);
	void __fastcall MyShowLastError(String Msg, bool Popup);
	String RegFileRootKey;
	char cwd[256];
	int SS;		// Splash Window if == 1
	int BytesSent, BuffersSent;
	HANDLE hRawFile;
	int SoundFileSize, EEPROMSize;
	int AvailableSpace;
	int EndAddrMax;
	int16_t AudioMax, AudioMin;
	int32_t AudioMax24, AudioMin24;

	bool __fastcall CheckEEPROMSize(int Movement);
	void __fastcall FileSetDateNow(String FileName);
	void __fastcall getFileString(HANDLE hFile, long offset, int Len, String Caption);
    #ifdef CLICKING
	// previousValue was declared short
	void __fastcall PlayClip(int StartAddr, int StartRepAddr, int EndRepAddr, int EndAddr, int DisplayAddr, short index, int previousValue);
    #else
	void __fastcall PlayClip(int StartAddr, int StartRepAddr, int EndRepAddr, int EndAddr, int DisplayAddr, short index, short previousValue);
    #endif
	int iDebug;
	bool RecordFlag;
	bool __fastcall TADPCM_Main::ChecksumCheck(String FileName);

BEGIN_MESSAGE_MAP
	MESSAGE_HANDLER(WM_DROPFILES, TWMDropFiles, WMDropFiles);
END_MESSAGE_MAP (TForm);
};
//---------------------------------------------------------------------------
extern PACKAGE TADPCM_Main *ADPCM_Main;
//---------------------------------------------------------------------------

// Decode/Encode a single sample and update state
#ifdef CLICKING
// function was declared short
int ImaAdpcmDecode(unsigned char deltaCode, ImaState &);
// sample was declared short
//unsigned char ImaAdpcmEncode(short sample, ImaState &);
unsigned char ImaAdpcmEncode(int sample, ImaState &);
#else
short ImaAdpcmDecode(unsigned char deltaCode, ImaState &);
unsigned char ImaAdpcmEncode(short sample, ImaState &);
#endif

// Microsoft-compatible decoder
class DecompressImaAdpcmMs {
private:
   int  _channels;
   short *_samples[2];  // Left and right sample buffers
   short *_samplePtr[2]; // Pointers to current samples
   size_t   _samplesRemaining; // Samples remaining in each channel
   size_t   _samplesPerPacket; // Total samples per packet
public:
   DecompressImaAdpcmMs(int packetLength, int channels);
   ~DecompressImaAdpcmMs();
private:
   unsigned char *_packet;   // Temporary buffer for packets
   size_t   _bytesPerPacket; // Size of a packet
};
#endif
