//---------------------------------------------------------------------------
/*
ADPCM Program for Windows

 Copyright 2020 by Mark DiVecchio

    This file is part of Remote Train Control.

    Remote Train Control is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Remote Train Control is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
    along with Remote Train Control.  If not, see <http://www.gnu.org/licenses/>.

*/
// Home Page: http://www.silogic.com/trains/ADPCM.html
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include <tchar.h>
#include <windows.h>
#include <mmsystem.h>
#include <stdio.h>
#include "assert.h"
#include "alloc.h"

#include "ADPCM_Player.h"
#include "Edit_the_Clip.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"

TClip_Editor *Clip_Editor;

extern bool ParseIndex(SFIndex &Index, HANDLE hParseFile, int clip);
extern bool WriteIndex(SFIndex &Index, HANDLE hParseFile, int clip);
extern bool WriteQuillIndex(QIndex &Index, HANDLE hParseFile, int index);
extern bool ParseQuillIndex(QIndex &Index, HANDLE hParseFile, int index);
int MyGetFileSizeBytes(HANDLE hFile);

/*

Here are the 6 cases supported:

	REPLACE_FROM_MTH_SOUND_FILE:		// replace clip from another sound file
	ADD_FROM_MTH_SOUND_FILE: 			// add clip from another sound file
	REPLACE_FROM_AUDIO_FILE:			// replace clip from an audio file
	ADD_FROM_AUDIO_FILE:				// add clip from an audio file
	DELETE_CLIP_FROM_MTH_SOUND_FILE:	// delete clip from a sound file
	REPLACE_CLIP_WITH_SILENCE:    		// replace clip with 27 bytes of silence
*/

//---------------------------------------------------------------------------
__fastcall TClip_Editor::TClip_Editor(TComponent* Owner)
	: TForm(Owner)
{
assert(false);		// never called
}
//---------------------------------------------------------------------------
__fastcall TClip_Editor::TClip_Editor(TComponent* Owner, int ClipNumber, String FN)
	: TForm(Owner)
{
// NOTE: possible bug in BCB 4.0 - overloaded constructors cannot have
// exactly 2 operands. Need 3 or more
ClipNo = abs(ClipNumber);     // default values
assert(ClipNo > 0 && ClipNo < 256);
SoundFileName = FN;
TargetDataByte = 0x00;
EndAddrMax = EMPTY_FIELD;
//
DataByteEdit = new THexEdit(this);
assert(DataByteEdit != NULL);
DataByteEdit->Parent = this;
DataByteEdit->Height = XDataByte->Height;
DataByteEdit->Width = XDataByte->Width;
DataByteEdit->Top = XDataByte->Top;
DataByteEdit->Left = XDataByte->Left;
DataByteEdit->MinValue = 0x00000000;
DataByteEdit->MaxValue = 0x000000FF;
DataByteEdit->Digits = 2;
DataByteEdit->MaxLength = 2;
DataByteEdit->Cursor = crHandPoint;
DataByteEdit->ShowHint = true;
DataByteEdit->Hint = "Databyte specifying format, sample rate, and category of clip";
DataByteEdit->Show();

Caption = "Editing Clip #" + String(ClipNo);
// User Registry Data
RegFileRootKey = ADPCM_Main->RegFileRootKey;

OpenDialog->DefaultExt = "mth";
OpenDialog->Filter = "Sound files (*.mth)|*.MTH|Audio Clip files (*.ima *.raw)|*.ima;*.IMA;*.raw;*.RAW";

// Experimental - use Drop Down Menu style in CB10.1
#if __BORLANDC__ >= 0x720
enum DECLSPEC_DENUM TButtonStyle : unsigned char { bsPushButton, bsCommandLink, bsSplitButton };
// bsSplitButton defined in Vcl.StdCtrls.hpp is really not defined
BrowseforFileButton->Style = bsSplitButton;
BrowseforFileButton->DropDownMenu = FileNamePopupMenu;
#endif

QWCheckBox->Checked = false;
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::FormClose(TObject *Sender, TCloseAction &Action)
{
// write preferences to the Registry
TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
// User Registry Data
RegFile->WriteString("Options", "SourceClipNoEdit", SourceClipNoEdit->Text);
RegFile->WriteInteger("Options", "OpGroup", OpGroup->ItemIndex);
//
// Deallocate RegFile
delete RegFile;
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::ShowIt(int clip, String FN, int DB, int EAM, int PS)
{
SampleGroup->ItemIndex = HZ5512;   // default sample rate
ClipNo = abs(clip);
assert(ClipNo > 0 && ClipNo < 256);
SoundFileName = FN;

#ifdef OLDWAY
// insert FN into MRU List
// don't insert a duplicate string
int idx = MRUComboBox->Items->IndexOf(FN);
if (idx == -1) {
	MRUComboBox->Items->Insert(0, FN);
	MRUComboBox->ItemIndex = 0;
	if (MRUComboBox->Items->Count > MRUMAX) MRUComboBox->Items->Delete(MRUMAX);
	} else {
	// set ItemIndex
	MRUComboBox->ItemIndex = idx;
    }
#else
// Set hint for MRUComboBox
MRUComboBox->Hint = MRUComboBox->Text;
#endif
TargetDataByte = DB;
EndAddrMax = EAM;
assert(EndAddrMax > 0x00080000 && EndAddrMax < 0x400000);
PSGroup->ItemIndex = PS;
Caption = "Editing Clip #" + String(ClipNo);

if (clip > 0) {
	Caption = "Editing Clip #" + String(ClipNo);
	} else {
	Caption = "Adding Clip #" + String(ClipNo);
	}
EngineEEPROMSize->Caption = ADPCM_Main->EngineEEPROMSize->Caption;
DataByteEdit->Text = IntToHex(DB | DATABYTEBASE, 2);
Show();
OpGroupClick(NULL);
SetDBCheckBoxes(TargetDataByte);
QWCheckBox->Checked = false;
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::OpGroupClick(TObject *Sender)
{
// choice?
//SourceFileNameEdit->Enabled = false;
SourceClipNoEdit->Enabled = false;
ClipNoLabel->Enabled = false;
BrowseforFileButton->Enabled = false;
DoItButton->Enabled = false;
DataByteEdit->Enabled = false;
DataLabel->Enabled = false;

switch(OpGroup->ItemIndex) {
//
//-----------------------------------------------------------------------------------------------------------------------------------------
//
	case REPLACE_FROM_MTH_SOUND_FILE:	// replace clip from another sound file
	case ADD_FROM_MTH_SOUND_FILE: 		// add clip from another sound file
		SourceClipNoEdit->Enabled = true;
		ClipNoLabel->Enabled = true;
		BrowseforFileButton->Enabled = true;
		SoftkeyPanel->Enabled = true;
		DoItButton->Enabled = true;
		DataByteEdit->Enabled = false;
		DataLabel->Enabled = false;
		DataLabel->Visible = false;
		XDataByte->Visible = false;
		DataByteEdit->Visible = false;
		Label1->Visible = false;
		ClipNoLabel->Visible = true;
		SourceClipNoEdit->Visible = true;
		Height = 236;
		Copyright->Top = 165;
		MRUComboBox->Enabled = true;
		OpenDialog->DefaultExt = "mth";
		OpenDialog->Filter = "Sound files (*.mth)|*.MTH";
		QWCheckBox->Enabled = true;
		break;
//
//-----------------------------------------------------------------------------------------------------------------------------------------
//
	case REPLACE_FROM_AUDIO_FILE:	// replace clip from an audio file
	case ADD_FROM_AUDIO_FILE:		// add clip from an audio file
		SourceClipNoEdit->Enabled = false;
		ClipNoLabel->Enabled = false;
		BrowseforFileButton->Enabled = true;
		SoftkeyPanel->Enabled = true;
		DoItButton->Enabled = true;
		DataByteEdit->Enabled = true;
		DataLabel->Enabled = true;
		DataLabel->Visible = true;
		XDataByte->Visible = true;
		DataByteEdit->Visible = true;
		Label1->Visible = true;
		ClipNoLabel->Visible = false;
		SourceClipNoEdit->Visible = false;
		Height = 344;
		Copyright->Top = 273;
		MRUComboBox->Enabled = true;
		// setup correct file query
		if (FormatGroup->ItemIndex == PCM24) {
			OpenDialog->DefaultExt = "raw";
			OpenDialog->Filter = "Audio PCM24 Clip files (*.raw)|*.raw;*.RAW";
			} else {
			OpenDialog->DefaultExt = "ima";
			OpenDialog->Filter = "Audio ADPCM Clip files (*.ima)|*.ima;*.IMA";
			}
		QWCheckBox->Enabled = false;
		QWCheckBox->Checked = false;
		break;
//
//-----------------------------------------------------------------------------------------------------------------------------------------
//
	case DELETE_CLIP_FROM_MTH_SOUND_FILE:   // delete clip from a sound file
	case REPLACE_CLIP_WITH_SILENCE:    		// replace clip with 27 bytes of silence
		SourceClipNoEdit->Enabled = false;
		ClipNoLabel->Enabled = false;
		BrowseforFileButton->Enabled = false;
		SoftkeyPanel->Enabled = false;
		DoItButton->Enabled = true;
		DataByteEdit->Enabled = false;
		DataLabel->Enabled = false;
		DataLabel->Enabled = true;
		DataLabel->Visible = false;
		XDataByte->Visible = false;
		DataByteEdit->Visible = false;
		Label1->Visible = false;
		ClipNoLabel->Visible = false;
		SourceClipNoEdit->Visible = false;
		Height = 236;
		Copyright->Top = 158;
		MRUComboBox->Enabled = false;
		QWCheckBox->Enabled = false;
		QWCheckBox->Checked = false;
		break;
	default:
		return;
	}
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::DoItButtonClick(TObject *Sender)
{
// Do the function
HANDLE hSoundFile, hSourceFile;
int sourceclipNo, Adjustment, fHandle;
SFIndex IndexSound, IndexSource;
long  fp;
DWORD readBytes, writtenBytes;
DWORD memblocksize;
char * membuf;
// validate
if (!FileExists(SoundFileName)) {
	ADPCM_Main->MyShowMessage("Sound File Not Found", true);
	return;
	}
if (ClipNo < 1 || ClipNo > 255) {
	ADPCM_Main->MyShowMessage("Sound Clip Number must be between 1 and 255", true);
	return;
	}
//
/*

Here are the 6 cases supported:

	REPLACE_FROM_SOUND_FILE:	// replace clip from another sound file
	ADD_FROM_SOUND_FILE: 		// add clip from another sound file
	REPLACE_FROM_AUDIO_FILE:	// replace clip from an audio file
	ADD_FROM_AUDIO_FILE:		// add clip from an audio file
	DELETE_CLIP_FROM_SOUND_FILE:   // delete clip from a sound file
	REPLACE_CLIP_WITH_SILENCE:    // replace clip with 27 bytes of silence
*/
switch (OpGroup->ItemIndex) {
//
//-----------------------------------------------------------------------------------------------------------------------------------------
//
	//-------------------------------------------------------------------- REPLACE_FROM_MTH_SOUND_FILE
	case REPLACE_FROM_MTH_SOUND_FILE:
		// insert a clip into this MTH sound file from another MTH sound file
		sourceclipNo = SourceClipNoEdit->Text.ToInt();
		if (sourceclipNo < 1 || sourceclipNo > 255) {
			ADPCM_Main->MyShowMessage("Source Clip Number must be between 1 and 255", true);
			return;
			}
		if (MRUComboBox->Text != "c:\\" && MRUComboBox->Text != "" && SourceClipNoEdit->Text != "0") {
			// read the sound file index entry
			// try and open the file
			if((hSoundFile = CreateFile(SoundFileName.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
				ADPCM_Main->MyShowLastError("DoIt: Unable to open sound file : " + SoundFileName, true);
				// cancel the operation
				return;
				} else {
				ADPCM_Main->MyShowMessage("DoIt: Sound file to be edited : " + SoundFileName, false);
				}
			FileLength = SetFilePointer(hSoundFile, 0, NULL, FILE_END);
			if (!ParseIndex(IndexSound, hSoundFile, ClipNo)) {
				ADPCM_Main->MyShowMessage("DoIt: Parse Index failed", true);
				CloseHandle(hSoundFile);
				return;
				}
			// There must be a clip there
			if ((IndexSound.StartAddr == 0) || (IndexSound.StartAddr == EMPTY_FIELD) || (IndexSound.ClipLength == 0)) {
				ADPCM_Main->MyShowMessage("DoIt: Sound File does not have a clip at that index", true);
				CloseHandle(hSoundFile);
				return;
				}
            // IndexSound.StartAddr does not change, replacement clip goes into StartAddr
			// read the editor file index entry
			// try and open the file
			if((hSourceFile = CreateFile(MRUComboBox->Text.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
				ADPCM_Main->MyShowLastError("DoIt: Unable to open file : " + MRUComboBox->Text, true);
				CloseHandle(hSoundFile);
				// cancel the operation
				return;
				} else {
				ADPCM_Main->MyShowMessage("DoIt: Source of replacement clip : " + MRUComboBox->Text, false);
				}
			if (!ParseIndex(IndexSource, hSourceFile, sourceclipNo)) {
					ADPCM_Main->MyShowMessage("DoIt: Parse Index failed", true);
					CloseHandle(hSourceFile);
					CloseHandle(hSoundFile);
					return;
					}
			// There must be a clip there
			if ((IndexSource.StartAddr == 0) || (IndexSource.StartAddr == EMPTY_FIELD) || (IndexSource.ClipLength == 0)) {
				ADPCM_Main->MyShowMessage("DoIt: Source Sound File does not have a clip at that index", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			//-------------------------------------------------------------------- REPLACE_FROM_MTH_SOUND_FILE
			Adjustment = IndexSound.ClipLength - IndexSource.ClipLength;	// movement amount
			// Plus means file has gotten smaller
			// Minus means file has gotten bigger
			ADPCM_Main->MyShowMessage("DoIt: Adjustment : " + String(Adjustment) + " bytes", false);
			//
			if (IndexSource.ClipLength == 0) {
				// zero length source not allowed
				ADPCM_Main->MyShowMessage("DoIt: Source clip sound has zero length", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			// Check that enough space is available in the EEPROM
			if (!ADPCM_Main->CheckEEPROMSize(Adjustment)) {
				// not enough space
				ADPCM_Main->MyShowMessage("DoIt: Not enough space to insert the clip", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			//
			// Clip index and/or quill index adjustment may be required
			//
			//-------------------------------------------------------------------- REPLACE_FROM_MTH_SOUND_FILE
			if (Adjustment != 0) {    // then following sounds must be moved
				// move all sounds away to/from EOF to increase/decrease the space used in the sound file
				memblocksize = FileLength - (IndexSound.EndAddr + 1);
				//
				membuf = (char *) malloc(memblocksize);
				if (membuf == NULL) {
					ADPCM_Main->MyShowMessage("DoIt: membuf malloc failed", true);
					CloseHandle(hSourceFile);
					CloseHandle(hSoundFile);
					return;
					}
				// set file pointer to first byte after the end of this clip
				fp = IndexSound.EndAddr + 1;
				SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
				// read length bytes
				ReadFile(hSoundFile, membuf, memblocksize, &readBytes, NULL);
				if (readBytes != memblocksize) {
					ADPCM_Main->MyShowMessage("DoIt: Could not read sound file", true);
					CloseHandle(hSourceFile);
					CloseHandle(hSoundFile);
					free (membuf);
					return;
					}
				// then write it with space for the new (larger or smaller) clip
				// set file pointer to first byte of the seqment
				fp = IndexSound.StartAddr + IndexSource.ClipLength;
				SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
				// write length bytes
				WriteFile(hSoundFile, membuf, memblocksize, &writtenBytes, NULL);
				if (writtenBytes != memblocksize) {
					ADPCM_Main->MyShowMessage("DoIt: Could not adjust sound file size", true);
					CloseHandle(hSourceFile);
					CloseHandle(hSoundFile);
					free (membuf);
					return;
					}
				free (membuf);
				//-------------------------------------------------------------------- REPLACE_FROM_MTH_SOUND_FILE
				// adjust all index entries for all moved sounds starting at the next clip in the sound file
				for (int indexclipNo = 1 ; indexclipNo <= 255 ; indexclipNo ++) {
					SFIndex AdjustClip;
					ParseIndex(AdjustClip, hSoundFile, indexclipNo);
					if (AdjustClip.StartAddr == 0 || AdjustClip.StartAddr == EMPTY_FIELD) continue; // on to next index entry if this clip is empty
					#ifdef OLDWAY
					// if starting address of this clip is after the end of the moved clip, adjust it because it was moved
					if (AdjustClip.StartAddr < IndexSound.EndAddr) {
						// no adjustment needed
						continue;
						}
					#else			// TBD  25 Jan 22 change due to bug reported by Brenden
					// if starting address of this clip is after the starting address of the moved clip, adjust it because clip was moved
					if (AdjustClip.StartAddr <= IndexSound.StartAddr) {
						// no adjustment needed
						continue;
						}

					#endif
					// adjust the addresses
					AdjustClip.StartAddr -= Adjustment;
					if (AdjustClip.StartRepAddr != EMPTY_FIELD) AdjustClip.StartRepAddr -= Adjustment;
					AdjustClip.EndAddr -= Adjustment;
					if (AdjustClip.EndRepAddr != EMPTY_FIELD) AdjustClip.EndRepAddr -= Adjustment;
					WriteIndex(AdjustClip, hSoundFile, indexclipNo);
					}
				// adjust quill data index if the quill clip was moved
				if (!QWCheckBox->Checked) { // we don't need to do this if we are processing a Quilled clip
					// read the first quill index
					QIndex QuillIndex;
					ParseQuillIndex(QuillIndex, hSoundFile, 1);
					if (QuillIndex.StartAddr != 0x000000 && QuillIndex.StartAddr != 0xFFFFFF) {   // make sure there is an index entry
						ADPCM_Main->MyShowMessage("DoIt: Quill Adjustment : " + String(Adjustment) + " bytes", false);
						// is it greater than the end address of the clip being replaced or deleted?
						if (QuillIndex.StartAddr >= IndexSound.EndAddr) {
							// yes adjust the Quill data
							QuillAddressFixup(hSoundFile, Adjustment); // Quill Data starts at offset 0x3200
							}
						}
					}
				}
			//-------------------------------------------------------------------- REPLACE_FROM_MTH_SOUND_FILE
			// put the new clip over top of the old clip
			// allocate memory of the correct size for a buffer
			membuf = (char *) malloc(IndexSource.ClipLength);
			if (membuf == NULL) {
				ADPCM_Main->MyShowMessage("DoIt: membuf malloc failed", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			// read from Source from IndexSource.StartAddr for IndexSource.ClipLength
			fp = IndexSource.StartAddr;
			SetFilePointer(hSourceFile, fp, NULL, FILE_BEGIN);
			// read length bytes
			ReadFile(hSourceFile, membuf, IndexSource.ClipLength, &readBytes, NULL);
			if (readBytes != IndexSource.ClipLength) {
				ADPCM_Main->MyShowMessage("DoIt: Could not read clip from source file", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				free(membuf);
				return;
				}
			//-------------------------------------------------------------------- REPLACE_FROM_MTH_SOUND_FILE
			// Calculate quill fixup for use later if we are copying a Quill clip
			int QuillFixup =  IndexSource.StartAddr - IndexSound.StartAddr;
			//
			// write to Sound File at IndexSound.StartAddr
			// set file pointer to first byte of the seqment
			fp = IndexSound.StartAddr;
			SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
			// write length bytes
			WriteFile(hSoundFile, membuf, IndexSource.ClipLength, &writtenBytes, NULL);
			if (writtenBytes != IndexSource.ClipLength) {
				ADPCM_Main->MyShowMessage("DoIt: Could not write clip to sound file", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				free (membuf);
				return;
				}
			free (membuf);
			//-------------------------------------------------------------------- REPLACE_FROM_MTH_SOUND_FILE
			// create and write the index entry
			// transfer and update sound file index
			// new addresses, keep length, keep data byte, keep SMInit
			// IndexSound.Clip is not changed
			//
			// IndexSound.StartAddr is not changed because we are putting the new clip at the same StartAddr
			if (IndexSource.StartRepAddr != EMPTY_FIELD) IndexSound.StartRepAddr = (IndexSource.StartRepAddr - IndexSource.StartAddr) + IndexSound.StartAddr;
			if (IndexSource.EndRepAddr != EMPTY_FIELD) IndexSound.EndRepAddr = (IndexSource.EndRepAddr - IndexSource.StartAddr) + IndexSound.StartAddr;

			IndexSound.SMInit = IndexSource.SMInit;
			IndexSound.ClipLength = IndexSource.ClipLength;
			IndexSound.EndAddr = IndexSound.StartAddr + IndexSource.ClipLength - 1;
			IndexSound.databyte = IndexSource.databyte;
			//
			IndexSound.SA_LOBit = IndexSource.SA_LOBit;
			IndexSound.SRA_LOBit = IndexSource.SRA_LOBit;
			IndexSound.ERA_LOBit = IndexSource.ERA_LOBit;
			IndexSound.EA_LOBit = IndexSource.EA_LOBit;
			//
			// write the sound file index entry
			if (!WriteIndex(IndexSound, hSoundFile, ClipNo)) {
				ADPCM_Main->MyShowMessage("DoIt: Could not write index to sound file", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			// search the index to see if this clip appears more than once
			for (int indexclipNo = 1 ; indexclipNo <= 255 ; indexclipNo ++) {
				if (indexclipNo == ClipNo) continue;	// don't reprocess ClipNo
				SFIndex AdjustClip;
				ParseIndex(AdjustClip, hSoundFile, indexclipNo);
				// if the start addresses match, write the new IndexSound into that clip index location
				if (IndexSound.StartAddr == AdjustClip.StartAddr) {
					// print message that another index entry must be changed
					ADPCM_Main->MyShowMessage("DoIt: Adjusting additional clip index : " + String(indexclipNo), false);
					// keep original volume control regime bits
					IndexSound.databyte = (IndexSound.databyte & 0xCF) | (AdjustClip.databyte & 0x30);
					WriteIndex(IndexSound, hSoundFile, indexclipNo);
					}
				}
			// Set new EOF if needed
			if (Adjustment != 0) {
				fp = FileLength - Adjustment;
				//
				SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
				SetEndOfFile(hSoundFile);
				}
			//
			// Copy Quillable Whistle Index if requested, then fixup the addresses in the Index
			if (QWCheckBox->Checked) {
				CopyBlock(hSourceFile, hSoundFile, 0x3200, 0x3EFF, 0x3200);
				QuillAddressFixup(hSoundFile, QuillFixup);
				}
			//
			CloseHandle(hSourceFile);
			//-------------------------------------------------------------------- REPLACE_FROM_MTH_SOUND_FILE
			if (SXSCheckBox->Checked) SetORFileByte(hSoundFile, 0x190C, 0x02);
			if (SAHCheckBox->Checked) SetORFileByte(hSoundFile, 0x190C, 0x01);
			if (CDxCheckBox->Checked) SetORFileByte(hSoundFile, 0x191F, 0x03);
			CloseHandle(hSoundFile);
			ADPCM_Main->FileSetDateNow(SoundFileName);
			//
			if (SoundFileName == ADPCM_Main->MRUComboBox->Text) {
				// Analyze only if filename has not changed since we started.
				ADPCM_Main->AnalyzeButtonClick(Sender);
				}
			}
		break;
//
//-----------------------------------------------------------------------------------------------------------------------------------------
//
	//-------------------------------------------------------------------- REPLACE_FROM_AUDIO_FILE
	case REPLACE_FROM_AUDIO_FILE:
		// replace clip with a clip from a disk file - IMA/DVI ADPCM or 24 bit big endian signed PCM
		if (MRUComboBox->Text != "c:\\" && MRUComboBox->Text != "" ) {
			// read the sound file index entry
			// try and open the file
			if((hSoundFile = CreateFile(SoundFileName.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
				ADPCM_Main->MyShowLastError("DoIt: Unable to open sound file : " + SoundFileName, true);
				// cancel the operation
				return;
				} else {
				ADPCM_Main->MyShowMessage("DoIt: Sound file to be edited : " + SoundFileName, false);
				}
			FileLength = SetFilePointer(hSoundFile, 0, NULL, FILE_END);
			if (!ParseIndex(IndexSound, hSoundFile, ClipNo)) {
				ADPCM_Main->MyShowMessage("DoIt: Parse Index failed", true);
				CloseHandle(hSoundFile);
				return;
				}
			// There must be a clip there
			if ((IndexSound.StartAddr == 0) || (IndexSound.StartAddr == EMPTY_FIELD) || (IndexSound.ClipLength == 0)) {
				ADPCM_Main->MyShowMessage("DoIt: Sound File does not have a clip at that index", true);
				CloseHandle(hSoundFile);
				return;
				}
            // IndexSound.StartAddr does not change, replacement clip goes into StartAddr
			// read the editor file index entry
			// try and open the file
			if((hSourceFile = CreateFile(MRUComboBox->Text.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
				ADPCM_Main->MyShowLastError("DoIt: Unable to open file : " + MRUComboBox->Text, true);
				CloseHandle(hSoundFile);
				// cancel the operation
				return;
				} else {
				ADPCM_Main->MyShowMessage("DoIt: Source of replacement clip : " + MRUComboBox->Text, false);
				IndexSource.DiskFileLength = MyGetFileSizeBytes(hSourceFile);
				}
			//-------------------------------------------------------------------- REPLACE_FROM_AUDIO_FILE
			// create source file clip Index entry
			IndexSource.Clip = ClipNo;
			IndexSource.StartAddr = 0;
			IndexSource.EndRepAddr = EMPTY_FIELD;
			IndexSource.StartRepAddr = EMPTY_FIELD;
			IndexSource.SMInit = 0x00;

			IndexSource.ClipLength = IndexSource.DiskFileLength;	// number of bytes in file/clip
			IndexSource.EndAddr = IndexSource.DiskFileLength - 1;	// Offset of last byte in file (0 based)
			IndexSource.databyte = _tcstol(DataByteEdit->Text.c_str(), NULL, 16);
			//
			IndexSource.SA_LOBit = 0;
			IndexSource.SRA_LOBit = 1;
			IndexSource.ERA_LOBit = 1;
			IndexSource.EA_LOBit = 0;
			//
			//-------------------------------------------------------------------- REPLACE_FROM_AUDIO_FILE
			Adjustment = IndexSound.ClipLength - IndexSource.ClipLength;	// movement amount
			// Plus means file has gotten smaller
			// Minus means file has gotten bigger
			ADPCM_Main->MyShowMessage("DoIt: Adjustment : " + String(Adjustment) + " bytes", false);
			//
			if (IndexSource.ClipLength == 0) {
				// zero length source not allowed
				ADPCM_Main->MyShowMessage("DoIt: Source clip file has zero length", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			// Check that enough space is available in the EEPROM
			if (!ADPCM_Main->CheckEEPROMSize(Adjustment)) {
				// not enough space
				ADPCM_Main->MyShowMessage("DoIt: Not enough space to insert the clip", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			//
			// Clip index and/or quill index adjustment may be required
			//
			//-------------------------------------------------------------------- REPLACE_FROM_AUDIO_FILE
			if (Adjustment != 0) {    // then following sounds must be moved
				// move all sounds away to/from EOF to increase/decrease the space used in the sound file
				memblocksize = FileLength - (IndexSound.EndAddr + 1);
				//
				membuf = (char *) malloc(memblocksize);
				if (membuf == NULL) {
					ADPCM_Main->MyShowMessage("DoIt: membuf malloc failed", true);
					CloseHandle(hSourceFile);
					CloseHandle(hSoundFile);
					return;
					}
				// set file pointer to first byte after the end of this clip
				fp = IndexSound.EndAddr + 1;
				SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
				// read length bytes
				ReadFile(hSoundFile, membuf, memblocksize, &readBytes, NULL);
				if (readBytes != memblocksize) {
					ADPCM_Main->MyShowMessage("DoIt: Could not read sound file", true);
					CloseHandle(hSourceFile);
					CloseHandle(hSoundFile);
					free (membuf);
					return;
					}
				// then write it with space for the new (larger or smaller) clip
				// set file pointer to first byte of the seqment
				fp = IndexSound.StartAddr + IndexSource.ClipLength;
				SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
				// write length bytes
				WriteFile(hSoundFile, membuf, memblocksize, &writtenBytes, NULL);
				if (writtenBytes != memblocksize) {
					ADPCM_Main->MyShowMessage("DoIt: Could not adjust sound file size", true);
					CloseHandle(hSourceFile);
					CloseHandle(hSoundFile);
					free (membuf);
					return;
					}
				free (membuf);
				//-------------------------------------------------------------------- REPLACE_FROM_AUDIO_FILE
				// adjust all index entries for all moved sounds starting at the next clip in the sound file
				for (int indexclipNo = 1 ; indexclipNo <= 255 ; indexclipNo ++) {
					SFIndex AdjustClip;
					ParseIndex(AdjustClip, hSoundFile, indexclipNo);
					if (AdjustClip.StartAddr == 0 || AdjustClip.StartAddr == EMPTY_FIELD) continue; // on to next index entry if this clip is empty
					#ifdef OLDWAY
					// if starting address of this clip is after the end of the moved clip, adjust it because it was moved
					if (AdjustClip.StartAddr < IndexSound.EndAddr) {
						// no adjustment needed
						continue;
						}
					#else			// TBD  25 Jan 22 change due to bug reported by Brenden
					// if starting address of this clip is after the starting address of the moved clip, adjust it because clip was moved
					if (AdjustClip.StartAddr <= IndexSound.StartAddr) {
						// no adjustment needed
						continue;
						}

					#endif
					// adjust the addresses
					AdjustClip.StartAddr -= Adjustment;
					if (AdjustClip.StartRepAddr != EMPTY_FIELD) AdjustClip.StartRepAddr -= Adjustment;
					AdjustClip.EndAddr -= Adjustment;
					if (AdjustClip.EndRepAddr != EMPTY_FIELD) AdjustClip.EndRepAddr -= Adjustment;
					WriteIndex(AdjustClip, hSoundFile, indexclipNo);
					}
				// adjust quill data index if the quill clip was moved
				if (!QWCheckBox->Checked) { // we don't need to do this if we are processing a Quilled clip
					// read the first quill index
					QIndex QuillIndex;
					ParseQuillIndex(QuillIndex, hSoundFile, 1);
					if (QuillIndex.StartAddr != 0x000000 && QuillIndex.StartAddr != 0xFFFFFF) {   // make sure there is an index entry
						ADPCM_Main->MyShowMessage("DoIt: Quill Adjustment : " + String(Adjustment) + " bytes", false);
						// is it greater than the end address of the clip being replaced or deleted?
						if (QuillIndex.StartAddr >= IndexSound.EndAddr) {
							// yes adjust the Quill data
							QuillAddressFixup(hSoundFile, Adjustment); // Quill Data starts at offset 0x3200
							}
						}
					}
				}
			//-------------------------------------------------------------------- REPLACE_FROM_AUDIO_FILE
			// put the new clip over top of the old clip
			// allocate memory of the correct size for a buffer
			membuf = (char *) malloc(IndexSource.ClipLength);
			if (membuf == NULL) {
				ADPCM_Main->MyShowMessage("DoIt: membuf malloc failed", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			// fill new buffer with all zeros
			memset(membuf, '\0', IndexSource.ClipLength);
			//
			// read from Source from start of file to EOF
			fp = 0;
			SetFilePointer(hSourceFile, fp, NULL, FILE_BEGIN);
			// read length bytes
			ReadFile(hSourceFile, membuf, IndexSource.ClipLength, &readBytes, NULL);
			if (readBytes != IndexSource.ClipLength) {
				ADPCM_Main->MyShowMessage("DoIt: Could not read clip from source file", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				free(membuf);
				return;
				}
			// write to Sound File at IndexSound.StartAddr
			// set file pointer to first byte of the seqment
			fp = IndexSound.StartAddr;
			SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
			// write length bytes
			WriteFile(hSoundFile, membuf, IndexSource.ClipLength, &writtenBytes, NULL);
			if (writtenBytes != IndexSource.ClipLength) {
				ADPCM_Main->MyShowMessage("DoIt: Could not write clip to sound file", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				free (membuf);
				return;
				}
			free (membuf);
			//-------------------------------------------------------------------- REPLACE_FROM_AUDIO_FILE
			// create and write the index entry
			// transfer and update sound file index
			// new addresses, keep length, keep data byte, keep SMInit
			IndexSound.Clip = IndexSource.Clip;
			//
			// IndexSound.StartAddr is not changed because we are putting the new clip at the same StartAddr
			// Repeat not supported on read clip from file
			IndexSound.EndRepAddr = IndexSource.EndRepAddr;
			IndexSound.StartRepAddr = IndexSource.StartRepAddr;
			// IndexSound.SMInit is not changed
			//IndexSound.SMInit = IndexSource.SMInit;
			IndexSound.ClipLength = IndexSource.ClipLength;
			IndexSound.EndAddr = IndexSound.StartAddr + IndexSource.ClipLength - 1;
			IndexSound.databyte = IndexSource.databyte;
			//
			IndexSound.SA_LOBit = IndexSource.SA_LOBit;
			IndexSound.SRA_LOBit = IndexSource.SRA_LOBit;
			IndexSound.ERA_LOBit = IndexSource.ERA_LOBit;
			IndexSound.EA_LOBit = IndexSource.EA_LOBit;
			//
			// write the sound file index entry
			if (!WriteIndex(IndexSound, hSoundFile, ClipNo)) {
				ADPCM_Main->MyShowMessage("DoIt: Could not write index to sound file", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			// search the index to see if this clip appears more than once
			for (int indexclipNo = 1 ; indexclipNo <= 255 ; indexclipNo ++) {
				if (indexclipNo == ClipNo) continue;	// don't reprocess ClipNo
				SFIndex AdjustClip;
				ParseIndex(AdjustClip, hSoundFile, indexclipNo);
				// if the start addresses match, write the new IndexSound into that clip index location
				if (IndexSound.StartAddr == AdjustClip.StartAddr) {
					// print message that another index entry must be changed
					ADPCM_Main->MyShowMessage("DoIt: Adjusting additional clip index : " + String(indexclipNo), false);
					// keep original volume control regime bits
					IndexSound.databyte = (IndexSound.databyte & 0xCF) | (AdjustClip.databyte & 0x30);
					WriteIndex(IndexSound, hSoundFile, indexclipNo);
					}
				}
			// Set new EOF if needed
			if (Adjustment != 0) {
				fp = FileLength - Adjustment;
				//
				SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
				SetEndOfFile(hSoundFile);
				}
			//
			CloseHandle(hSourceFile);
			//-------------------------------------------------------------------- REPLACE_FROM_MTH_SOUND_FILE
			if (SXSCheckBox->Checked) SetORFileByte(hSoundFile, 0x190C, 0x02);
			if (SAHCheckBox->Checked) SetORFileByte(hSoundFile, 0x190C, 0x01);
			if (CDxCheckBox->Checked) SetORFileByte(hSoundFile, 0x191F, 0x03);
			CloseHandle(hSoundFile);
			ADPCM_Main->FileSetDateNow(SoundFileName);
			//
			if (SoundFileName == ADPCM_Main->MRUComboBox->Text) {
				// Analyze only if filename has not changed since we started.
				ADPCM_Main->AnalyzeButtonClick(Sender);
				}
			}
		break;
//
//-----------------------------------------------------------------------------------------------------------------------------------------
//
	//-------------------------------------------------------------------- ADD_FROM_MTH_SOUND_FILE
	case ADD_FROM_MTH_SOUND_FILE:
		// add a clip from another sound file into an empty index location
		sourceclipNo = SourceClipNoEdit->Text.ToInt();
		if (sourceclipNo < 1 || sourceclipNo > 255) {
			ADPCM_Main->MyShowMessage("Source Clip Number must be between 1 and 255", true);
			return;
			}
		if (MRUComboBox->Text != "c:\\" && MRUComboBox->Text != "" && SourceClipNoEdit->Text != "0") {
			// read the sound file index entry
			// try and open the file
			if((hSoundFile = CreateFile(SoundFileName.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
				ADPCM_Main->MyShowLastError("DoIt: Unable to open sound file : " + SoundFileName, true);
				// cancel the operation
				return;
				} else {
				ADPCM_Main->MyShowMessage("DoIt: Sound file to be edited : " + SoundFileName, false);
				}
			FileLength = SetFilePointer(hSoundFile, 0, NULL, FILE_END);
			if (!ParseIndex(IndexSound, hSoundFile, ClipNo)) {
				ADPCM_Main->MyShowMessage("DoIt: Parse Index failed", true);
				CloseHandle(hSoundFile);
				return;
				}
			// There must NOT be a clip there
			if ((IndexSound.StartAddr != 0) && (IndexSound.StartAddr != EMPTY_FIELD) && (IndexSound.ClipLength != 0)) {
				ADPCM_Main->MyShowMessage("DoIt: Sound File already has a clip at that index", true);
				CloseHandle(hSoundFile);
				return;
				}
			// create source file clip Index entry
			IndexSound.StartAddr = EndAddrMax + 1;  // starting address for added clip
			// read the source file index entry
			// try and open the file
			if((hSourceFile = CreateFile(MRUComboBox->Text.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
				ADPCM_Main->MyShowLastError("DoIt: Unable to open file : " + MRUComboBox->Text, true);
				CloseHandle(hSoundFile);
				// cancel the operation
				return;
				} else {
				ADPCM_Main->MyShowMessage("DoIt: Source of replacement clip : " + MRUComboBox->Text, false);
				}
			//-------------------------------------------------------------------- ADD_FROM_AUDIO_FILE
			//
			// create Source file clip Index entry
			if (!ParseIndex(IndexSource, hSourceFile, sourceclipNo)) {
				ADPCM_Main->MyShowMessage("DoIt: Parse Index failed", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			// There must be a clip there
			if ((IndexSource.StartAddr == 0) || (IndexSource.StartAddr == EMPTY_FIELD) || (IndexSource.ClipLength == 0)) {
				ADPCM_Main->MyShowMessage("DoIt: Source Sound File does not have a clip at that index", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			//-------------------------------------------------------------------- ADD_FROM_MTH_SOUND_FILE
			Adjustment = 0 - IndexSource.ClipLength;	// file growth amount
			// Plus means file has gotten smaller
			// Minus means file has gotten bigger
			ADPCM_Main->MyShowMessage("DoIt: Adjustment : " + String(Adjustment) + " bytes", false);
			//
			if (IndexSource.ClipLength == 0) {
				// zero length source not allowed
				ADPCM_Main->MyShowMessage("DoIt: Source clip has zero length", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			// Check that enough space is available in the EEPROM
			if (!ADPCM_Main->CheckEEPROMSize(Adjustment)) {
				// not enough space
				ADPCM_Main->MyShowMessage("DoIt: Not enough space to add the clip", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			//
			// Put the new clip at the end of the file
			// therefore, no clip index or quill index adjustment is needed
			//
			//-------------------------------------------------------------------- ADD_FROM_MTH_SOUND_FILE
			// allocate memory of the correct size for a buffer
			membuf = (char *) malloc(IndexSource.ClipLength);		// ClipLength includes the extra byte
			if (membuf == NULL) {
				ADPCM_Main->MyShowMessage("DoIt: membuf malloc failed", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			// read from Source from IndexSource.StartAddr for IndexSource.ClipLength
			fp = IndexSource.StartAddr;
			SetFilePointer(hSourceFile, fp, NULL, FILE_BEGIN);
			// read length bytes
			ReadFile(hSourceFile, membuf, IndexSource.ClipLength, &readBytes, NULL);
			if (readBytes != IndexSource.ClipLength) {
				ADPCM_Main->MyShowMessage("DoIt: Could not read clip from source file", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				free(membuf);
				return;
				}
			//-------------------------------------------------------------------- ADD_FROM_MTH_SOUND_FILE
			// Calculate quill fixup for use later if we are copying a Quill clip
			int QuillFixup = IndexSource.StartAddr - IndexSound.StartAddr;
			//
			// write to Sound File at IndexSound.StartAddr
			// set file pointer to first byte of the seqment
			fp = IndexSound.StartAddr;	// really (EndAddrMax + 1)
			SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
			// write length bytes
			WriteFile(hSoundFile, membuf, IndexSource.ClipLength, &writtenBytes, NULL);
			if (writtenBytes != IndexSource.ClipLength) {
				ADPCM_Main->MyShowMessage("DoIt: Could not write clip to sound file", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				free (membuf);
				return;
				}
			free (membuf);
			//-------------------------------------------------------------------- ADD_FROM_MTH_SOUND_FILE
			// create and write the index entry
			// transfer and update sound file index
			// new addresses, keep length, keep data byte, keep SMInit
			// IndexSound.Clip is not changed
			// IndexSound.StartAddr is not changed because we are putting the new clip at the same StartAddr
			if (IndexSource.StartRepAddr != EMPTY_FIELD) IndexSound.StartRepAddr = (IndexSource.StartRepAddr - IndexSource.StartAddr) + IndexSound.StartAddr;
			if (IndexSource.EndRepAddr != EMPTY_FIELD) IndexSound.EndRepAddr = (IndexSource.EndRepAddr - IndexSource.StartAddr) + IndexSound.StartAddr;

			IndexSound.EndAddr = (IndexSource.EndAddr - IndexSource.StartAddr) + IndexSound.StartAddr;
			IndexSound.SMInit = IndexSource.SMInit;
			IndexSound.ClipLength = IndexSource.ClipLength;
			IndexSound.databyte = IndexSource.databyte;
			//
			IndexSound.SA_LOBit = IndexSource.SA_LOBit;
			IndexSound.SRA_LOBit = IndexSource.SRA_LOBit;
			IndexSound.ERA_LOBit = IndexSource.ERA_LOBit;
			IndexSound.EA_LOBit = IndexSource.EA_LOBit;
			//
			// write the sound file index entry
			if (!WriteIndex(IndexSound, hSoundFile, ClipNo)) {
				ADPCM_Main->MyShowMessage("DoIt: Could not write index to sound file", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			// Set new EOF if needed
			if (Adjustment != 0) {
				fp = FileLength - Adjustment;
				//
				SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
				SetEndOfFile(hSoundFile);
				}
			//
			// Copy Quillable Whistle Index if requested, then fixup the addresses in the Index
			if (QWCheckBox->Checked) {
				CopyBlock(hSourceFile, hSoundFile, 0x3200, 0x3EFF, 0x3200);
				QuillAddressFixup(hSoundFile, QuillFixup);
				}
			//
			CloseHandle(hSourceFile);
			//-------------------------------------------------------------------- ADD_FROM_MTH_SOUND_FILE
			if (SXSCheckBox->Checked) SetORFileByte(hSoundFile, 0x190C, 0x02);
			if (SAHCheckBox->Checked) SetORFileByte(hSoundFile, 0x190C, 0x01);
			if (CDxCheckBox->Checked) SetORFileByte(hSoundFile, 0x191F, 0x03);
			CloseHandle(hSoundFile);
			ADPCM_Main->FileSetDateNow(SoundFileName);
			//
			if (SoundFileName == ADPCM_Main->MRUComboBox->Text) {
				// Analyze only if filename has not changed since we started.
				ADPCM_Main->AnalyzeButtonClick(Sender);
				}
			}
		break;
//
//-----------------------------------------------------------------------------------------------------------------------------------------
//
	//-------------------------------------------------------------------- ADD_FROM_AUDIO_FILE
	case ADD_FROM_AUDIO_FILE:
		// add a clip from a disk file into an empty index location - IMA/DVI ADPCM or 24 bit big endian signed PCM
		// ADD_FROM_AUDIO_FILE does not put in repeat segment. If you need one, edit the index manually
		if (MRUComboBox->Text != "c:\\" && MRUComboBox->Text != "" ) {
			// read the sound file index entry
			// try and open the file
			if((hSoundFile = CreateFile(SoundFileName.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
				ADPCM_Main->MyShowLastError("DoIt: Unable to open sound file : " + SoundFileName, true);
				// cancel the operation
				return;
				} else {
				ADPCM_Main->MyShowMessage("DoIt: Sound file to be edited : " + SoundFileName, false);
				}
			FileLength = SetFilePointer(hSoundFile, 0, NULL, FILE_END);
			if (!ParseIndex(IndexSound, hSoundFile, ClipNo)) {
				ADPCM_Main->MyShowMessage("DoIt: Parse Index failed", true);
				CloseHandle(hSoundFile);
				return;
				}
			// There must NOT be a clip there
			if ((IndexSound.StartAddr != 0) && (IndexSound.StartAddr != EMPTY_FIELD) && (IndexSound.ClipLength != 0)) {
				ADPCM_Main->MyShowMessage("DoIt: Sound File already has a clip at that index", true);
				CloseHandle(hSoundFile);
				return;
				}
			// read the editor file index entry
			// try and open the file
			if((hSourceFile = CreateFile(MRUComboBox->Text.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
				ADPCM_Main->MyShowLastError("DoIt: Unable to open file : " + MRUComboBox->Text, true);
				CloseHandle(hSoundFile);
				// cancel the operation
				return;
				} else {
				ADPCM_Main->MyShowMessage("DoIt: Source of replacement clip : " + MRUComboBox->Text, false);
				IndexSource.DiskFileLength = MyGetFileSizeBytes(hSourceFile);
				}
			//-------------------------------------------------------------------- ADD_FROM_AUDIO_FILE
			//
			// create Source file clip Index entry
			IndexSource.Clip = ClipNo;
			IndexSource.StartAddr = 0;
			IndexSource.EndRepAddr = EMPTY_FIELD;
			IndexSource.StartRepAddr = EMPTY_FIELD;
			IndexSource.SMInit = 0x00;
			IndexSource.ClipLength = IndexSource.DiskFileLength;	// number of bytes in file

			IndexSource.EndAddr = IndexSource.ClipLength - 1;	// Offset of last byte in file (0 based)
			IndexSource.databyte = _tcstol(DataByteEdit->Text.c_str(), NULL, 16);
			//
			IndexSource.SA_LOBit = 0;
			IndexSource.SRA_LOBit = 1;
			IndexSource.ERA_LOBit = 1;
			IndexSource.EA_LOBit = 0;
			//-------------------------------------------------------------------- ADD_FROM_AUDIO_FILE
			// Create the New Sound File index entry
			// new addresses, keep length and databyte
			IndexSound.Clip = IndexSource.Clip;
			IndexSound.StartAddr = EndAddrMax + 1;  // starting address for added clip
			// Repeat Segment not supported on clip read from file - if you need this, manually edit the index
			IndexSound.EndRepAddr = IndexSource.EndRepAddr;
			IndexSound.StartRepAddr = IndexSource.StartRepAddr;
			// IndexSound.SMInit is not changed
			//IndexSound.SMInit = IndexSource.SMInit;
			IndexSound.ClipLength = IndexSource.ClipLength;
			IndexSound.EndAddr = IndexSound.StartAddr + IndexSource.ClipLength - 1;
			IndexSound.databyte = IndexSource.databyte;
			//
			IndexSound.SA_LOBit = IndexSource.SA_LOBit;
			IndexSound.SRA_LOBit = IndexSource.SRA_LOBit;
			IndexSound.ERA_LOBit = IndexSource.ERA_LOBit;
			IndexSound.EA_LOBit = IndexSource.EA_LOBit;
			//
			//-------------------------------------------------------------------- ADD_FROM_AUDIO_FILE 																									ADD_FROM_AUDIO_FILE
			Adjustment = 0 - IndexSource.ClipLength;	// file growth amount
			// Plus means file has gotten smaller
			// Minus means file has gotten bigger
			ADPCM_Main->MyShowMessage("DoIt: Adjustment : " + String(Adjustment) + " bytes", false);
			//
			if (IndexSource.ClipLength == 0) {
				// zero length source not allowed
				ADPCM_Main->MyShowMessage("DoIt: Source clip file has zero length", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			// Check that enough space is available in the EEPROM
			if (!ADPCM_Main->CheckEEPROMSize(Adjustment)) {
				// not enough space
				ADPCM_Main->MyShowMessage("DoIt: Not enough space to add the clip", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			//
			// Put the new clip at the end of the file
			// therefore, no clip index or quill index adjustment is needed
			//
			//-------------------------------------------------------------------- ADD_FROM_AUDIO_FILE
			// allocate memory of the correct size for a buffer
			membuf = (char *) malloc(IndexSource.ClipLength);		// ClipLength includes the extra byte
			if (membuf == NULL) {
				ADPCM_Main->MyShowMessage("DoIt: membuf malloc failed", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			// fill new buffer with all zeros
			memset(membuf, '\0', IndexSource.ClipLength);
			//
			// read from Source from start of file to EOF
			fp = 0;
			SetFilePointer(hSourceFile, fp, NULL, FILE_BEGIN);
			// read length bytes
			ReadFile(hSourceFile, membuf, IndexSource.ClipLength, &readBytes, NULL);
			if (readBytes != IndexSource.ClipLength) {
				ADPCM_Main->MyShowMessage("DoIt: Could not read clip from source file", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				free(membuf);
				return;
				}
			//-------------------------------------------------------------------- ADD_FROM_AUDIO_FILE
			// write to Sound File at IndexSound.StartAddr
			// set file pointer to first byte of the seqment
			fp = IndexSound.StartAddr;	// really (EndAddrMax + 1)
			SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
			// write length bytes
			WriteFile(hSoundFile, membuf, IndexSource.ClipLength, &writtenBytes, NULL);
			if (writtenBytes != IndexSource.ClipLength) {
				ADPCM_Main->MyShowMessage("DoIt: Could not write clip to sound file", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				free (membuf);
				return;
				}
			free (membuf);
			//-------------------------------------------------------------------- ADD_FROM_AUDIO_FILE
			// write the sound file index entry
			if (!WriteIndex(IndexSound, hSoundFile, ClipNo)) {
				ADPCM_Main->MyShowMessage("DoIt: Could not write index to sound file", true);
				CloseHandle(hSourceFile);
				CloseHandle(hSoundFile);
				return;
				}
			// Set new EOF if needed
			if (Adjustment != 0) {
				fp = FileLength - Adjustment;
				//
				SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
				SetEndOfFile(hSoundFile);
				}
			//
			CloseHandle(hSourceFile);
			//-------------------------------------------------------------------- ADD_FROM_AUDIO_FILE
			if (SXSCheckBox->Checked) SetORFileByte(hSoundFile, 0x190C, 0x02);
			if (SAHCheckBox->Checked) SetORFileByte(hSoundFile, 0x190C, 0x01);
			if (CDxCheckBox->Checked) SetORFileByte(hSoundFile, 0x191F, 0x03);
			CloseHandle(hSoundFile);
			ADPCM_Main->FileSetDateNow(SoundFileName);
			//
			if (SoundFileName == ADPCM_Main->MRUComboBox->Text) {
				// Analyze only if filename has not changed since we started.
				ADPCM_Main->AnalyzeButtonClick(Sender);
				}
			}
		break;
//
//-----------------------------------------------------------------------------------------------------------------------------------------
//
	//-------------------------------------------------------------------- DELETE_CLIP_FROM_MTH_SOUND_FILE
	case DELETE_CLIP_FROM_MTH_SOUND_FILE:
		// delete clip and recover space in MTH sound file
		// read the sound file index entry
		// try and open the file
		if((hSoundFile = CreateFile(SoundFileName.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
			ADPCM_Main->MyShowLastError("DoIt: Unable to open sound file : " + SoundFileName, true);
			// cancel the operation
			return;
			} else {
			ADPCM_Main->MyShowMessage("DoIt: Sound file with clip #" + String(ClipNo) +  " to be deleted : " + SoundFileName, false);
			}
		FileLength = SetFilePointer(hSoundFile, 0, NULL, FILE_END);
		if (!ParseIndex(IndexSound, hSoundFile, ClipNo)) {
			ADPCM_Main->MyShowMessage("DoIt: Parse Index failed", true);
			CloseHandle(hSoundFile);
			return;
			}
		// There must be a clip there
		if ((IndexSound.StartAddr == 0) || (IndexSound.StartAddr == EMPTY_FIELD) || (IndexSound.ClipLength == 0)) {
			ADPCM_Main->MyShowMessage("DoIt: Sound File does not have a clip at that index", true);
			CloseHandle(hSoundFile);
			return;
			}
		//-------------------------------------------------------------------- DELETE_CLIP_FROM_MTH_SOUND_FILE
		// clear and write the index entry
		IndexSource.Clip = ClipNo;
		IndexSource.StartAddr = EMPTY_FIELD;
		IndexSource.EndRepAddr = EMPTY_FIELD;
		IndexSource.StartRepAddr = EMPTY_FIELD;
		IndexSource.SMInit = 0x000000;
		IndexSource.EndAddr = EMPTY_FIELD;
		IndexSource.ClipLength = 0;
		IndexSource.databyte = 0x00;
		//
		IndexSource.SA_LOBit = 1;
		IndexSource.SRA_LOBit = 1;
		IndexSource.ERA_LOBit = 1;
		IndexSource.EA_LOBit = 1;
		//
		// write the sound file index entry
		if (!WriteIndex(IndexSource, hSoundFile, ClipNo)) {
			ADPCM_Main->MyShowMessage("DoIt: Could not write index to sound file", true);
			CloseHandle(hSoundFile);
			return;
			}
		// search the index to see if this clip appears more than once
		for (int indexclipNo = 1 ; indexclipNo <= 255 ; indexclipNo ++) {
			if (indexclipNo == ClipNo) continue;	// don't reprocess ClipNo
			SFIndex AdjustClip;
			ParseIndex(AdjustClip, hSoundFile, indexclipNo);
			// if the start addresses match, write the cleared IndexSource into that clip index location
			if (IndexSound.StartAddr == AdjustClip.StartAddr) {
				// print message that another index entry must be changed
				ADPCM_Main->MyShowMessage("DoIt: Adjusting additional clip index : " + String(indexclipNo), false);
				WriteIndex(IndexSource, hSoundFile, indexclipNo);
				}
			}
		//-------------------------------------------------------------------- DELETE_CLIP_FROM_MTH_SOUND_FILE
		Adjustment = IndexSound.ClipLength;	// movement amount - entire clip is deleted
		// Plus means file has gotten smaller
		// Minus means file has gotten bigger
		ADPCM_Main->MyShowMessage("DoIt: Adjustment : " + String(Adjustment) + " bytes", false);
		//
		// Clip index and/or quill index adjustment may be required
		//
		if (Adjustment != 0) {    // then following sounds must be moved
			// move all sounds away from EOF to decrease the space used in the sound file
			memblocksize = FileLength - (IndexSound.EndAddr + 1);
			//			
			membuf = (char *) malloc(memblocksize);
			if (membuf == NULL) {
				ADPCM_Main->MyShowMessage("DoIt: membuf malloc failed", true);
				CloseHandle(hSoundFile);
				return;
				}
			// set file pointer to first byte of the next clip
			fp = IndexSound.EndAddr + 1;
			SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
			// read length bytes
			ReadFile(hSoundFile, membuf, memblocksize, &readBytes, NULL);
			if (readBytes != memblocksize) {
				ADPCM_Main->MyShowMessage("DoIt: Could not read sound file", true);
				CloseHandle(hSoundFile);
				free (membuf);
				return;
				}
			// then write it closed up
			// set file pointer to first byte of the deleted clip
			fp = IndexSound.StartAddr;
			SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
			// write length bytes
			WriteFile(hSoundFile, membuf, memblocksize, &writtenBytes, NULL);
			if (writtenBytes != memblocksize) {
				ADPCM_Main->MyShowMessage("DoIt: Could not adjust sound file size", true);
				CloseHandle(hSoundFile);
				free (membuf);
				return;
				}
			free (membuf);
			// adjust all index entries for all moved sounds starting at the next clip in the sound file
			for (int indexclipNo = 1 ; indexclipNo <= 255 ; indexclipNo ++) {
				SFIndex AdjustClip;
				ParseIndex(AdjustClip, hSoundFile, indexclipNo);
				if (AdjustClip.StartAddr == 0 || AdjustClip.StartAddr == EMPTY_FIELD) continue; // on to next index entry if this clip is empty
				#ifdef OLDWAY
				// if starting address of this clip is after the end of the moved clip, adjust it because it was moved
				if (AdjustClip.StartAddr < IndexSound.EndAddr) {
					// no adjustment needed
					continue;
					}
				#else			// TBD  25 Jan 22 change due to bug reported by Brenden
				// if starting address of this clip is after the starting address of the moved clip, adjust it because clip was moved
				if (AdjustClip.StartAddr <= IndexSound.StartAddr) {
					// no adjustment needed
					continue;
					}
				#endif
				// adjust the addresses
				AdjustClip.StartAddr -= Adjustment;
				if (AdjustClip.StartRepAddr != EMPTY_FIELD) AdjustClip.StartRepAddr -= Adjustment;
				AdjustClip.EndAddr -= Adjustment;
				if (AdjustClip.EndRepAddr != EMPTY_FIELD) AdjustClip.EndRepAddr -= Adjustment;
				WriteIndex(AdjustClip, hSoundFile, indexclipNo);
				}
			// adjust quill data index if the quill clip was moved
			if (!QWCheckBox->Checked) { // we don't need to do this if we are processing a Quilled clip
				// read the first quill index
				QIndex QuillIndex;
				ParseQuillIndex(QuillIndex, hSoundFile, 1);
				if (QuillIndex.StartAddr != 0x000000 && QuillIndex.StartAddr != 0xFFFFFF) {   // make sure there is an index entry
					ADPCM_Main->MyShowMessage("DoIt: Quill Adjustment : " + String(Adjustment) + " bytes", false);
					// is it greater than the end address of the clip being replaced or deleted?
					if (QuillIndex.StartAddr >= IndexSound.EndAddr) {
						// yes adjust the Quill data
						QuillAddressFixup(hSoundFile, Adjustment); // Quill Data starts at offset 0x3200
						}
					}
				}
			//
			}
		// Set new EOF if needed
		if (Adjustment != 0) {
			fp = FileLength - Adjustment;
			//
			SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
			SetEndOfFile(hSoundFile);
			}
		//
		CloseHandle(hSoundFile);
		ADPCM_Main->FileSetDateNow(SoundFileName);
		//
		if (SoundFileName == ADPCM_Main->MRUComboBox->Text) {
			// Analyze only if filename has not changed since we started.
			ADPCM_Main->AnalyzeButtonClick(Sender);
			}
		break;
//
//-----------------------------------------------------------------------------------------------------------------------------------------
//
	//-------------------------------------------------------------------- REPLACE_CLIP_WITH_SILENCE
	case REPLACE_CLIP_WITH_SILENCE:
		// replace clip with silence - IMA/DVI 4 bit ADPCM or 24 bit big endian signed PCM
        // replacement clip is always the same format as the original sound clip.
		// read the sound file index entry
		// try and open the file
		if((hSoundFile = CreateFile(SoundFileName.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
			ADPCM_Main->MyShowLastError("DoIt: Unable to open sound file : " + SoundFileName, true);
			// cancel the operation
			return;
			} else {
			ADPCM_Main->MyShowMessage("DoIt: Sound file to be edited : " + SoundFileName, false);
			}
		FileLength = SetFilePointer(hSoundFile, 0, NULL, FILE_END);
		if (!ParseIndex(IndexSound, hSoundFile, ClipNo)) {
			ADPCM_Main->MyShowMessage("DoIt: Parse Index failed", true);
			CloseHandle(hSoundFile);
			return;
			}
		// There must be a clip there
		if ((IndexSound.StartAddr == 0) || (IndexSound.StartAddr == EMPTY_FIELD) || (IndexSound.ClipLength == 0)) {
			ADPCM_Main->MyShowMessage("DoIt: Sound File does not have a clip at that index", true);
			CloseHandle(hSoundFile);
			return;
			}
		// create silent clip Index entry
		IndexSource.Clip = ClipNo;
		IndexSource.StartAddr = IndexSound.StartAddr;
		IndexSource.SA_LOBit = 0;
		if (IndexSound.StartRepAddr == EMPTY_FIELD) {
			IndexSource.EndRepAddr = EMPTY_FIELD;  // no repeat segment
			IndexSource.StartRepAddr = EMPTY_FIELD;
			IndexSource.SRA_LOBit = 1;
			IndexSource.ERA_LOBit = 1;
			} else {
			IndexSource.EndRepAddr = IndexSource.StartAddr + (27 * 2)/3; // fake a repeat segment
			IndexSource.StartRepAddr = IndexSource.StartAddr + (27 * 1)/3;
			IndexSource.SRA_LOBit = 0;
			IndexSource.ERA_LOBit = 0;
			}
		IndexSource.ClipLength = 27;

		IndexSource.SMInit = 0x00;
		IndexSource.EndAddr = (IndexSound.StartAddr + IndexSource.ClipLength) - 1;	// offset of last byte in silence clip (0 based)
		IndexSource.EA_LOBit = 0;
		IndexSource.databyte = IndexSound.databyte;    // data byte does not change
		/*
		// create the index entry
		// IndexSound.clipNo is not changed
		// IndexSound.StartAddr is not changed because we are putting the new clip at the same StartAddr
		IndexSound.StartRepAddr = IndexSource.StartRepAddr;
		IndexSound.EndRepAddr = IndexSource.EndRepAddr;
		IndexSound.SMInit = IndexSource.SMInit;
		IndexSound.EndAddr = IndexSource.EndAddr;
		IndexSound.ClipLength = IndexSource.ClipLength;
		IndexSound.databyte = IndexSource.databyte;
		//
		IndexSound.SA_LOBit = IndexSource.SA_LOBit;
		IndexSound.SRA_LOBit = IndexSource.SRA_LOBit;
		IndexSound.ERA_LOBit = IndexSource.ERA_LOBit;
		IndexSound.EA_LOBit = IndexSource.EA_LOBit;
		*/
		//-------------------------------------------------------------------- REPLACE_CLIP_WITH_SILENCE
		Adjustment = IndexSound.ClipLength - IndexSource.ClipLength;	// movement amount - entire clip minus silence clip length
		// Plus means file has gotten smaller
		// Minus means file has gotten bigger
		ADPCM_Main->MyShowMessage("DoIt: Adjustment : " + String(Adjustment) + " bytes", false);
		// Check that enough space is available in the EEPROM
		if (!ADPCM_Main->CheckEEPROMSize(Adjustment)) {
				// not enough space
				ADPCM_Main->MyShowMessage("DoIt: Not enough space to insert the silent clip", true);
				CloseHandle(hSoundFile);
				return;
				}
		//
		// Clip index and/or quill index adjustment may be required
		//
		//-------------------------------------------------------------------- REPLACE_CLIP_WITH_SILENCE
		if (Adjustment != 0) {    // then following sounds must be moved
			// move all sounds away to/from EOF to increase/decrease the space used in the sound file
			memblocksize = FileLength - (IndexSound.EndAddr + 1);
			//			
			membuf = (char *) malloc(memblocksize);
			if (membuf == NULL) {
				ADPCM_Main->MyShowMessage("DoIt: membuf malloc failed", true);
				CloseHandle(hSoundFile);
				return;
				}
			// set file pointer to first byte of the next clip
			fp = IndexSound.EndAddr + 1;
			SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
			// read length bytes
			ReadFile(hSoundFile, membuf, memblocksize, &readBytes, NULL);
			if (readBytes != memblocksize) {
				ADPCM_Main->MyShowMessage("DoIt: Could not read sound file", true);
				CloseHandle(hSoundFile);
				free (membuf);
				return;
				}
			// then write it closed up
			// set file pointer to first byte of the seqment
			fp = IndexSound.StartAddr + IndexSource.ClipLength;
			SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
			// write length bytes
			WriteFile(hSoundFile, membuf, memblocksize, &writtenBytes, NULL);
			if (writtenBytes != memblocksize) {
				ADPCM_Main->MyShowMessage("DoIt: Could not adjust sound file size", true);
				CloseHandle(hSoundFile);
				free (membuf);
				return;
				}
			free (membuf);
			// adjust all index entries for all moved sounds starting at the next clip in the sound file
			for (int indexclipNo = 1 ; indexclipNo <= 255 ; indexclipNo ++) {
				SFIndex AdjustClip;
				ParseIndex(AdjustClip, hSoundFile, indexclipNo);
				if (AdjustClip.StartAddr == 0 || AdjustClip.StartAddr == EMPTY_FIELD) continue; // on to next index entry if this clip is empty
				#ifdef OLDWAY
				// if starting address of this clip is after the end of the moved clip, adjust it because it was moved
				if (AdjustClip.StartAddr < IndexSound.EndAddr) {
					// no adjustment needed
					continue;
					}
				#else			// TBD  25 Jan 22 change due to bug reported by Brenden
				// if starting address of this clip is after the starting address of the moved clip, adjust it because clip was moved
				if (AdjustClip.StartAddr <= IndexSound.StartAddr) {
					// no adjustment needed
					continue;
					}
				#endif
				// adjust the addresses
				AdjustClip.StartAddr -= Adjustment;
				if (AdjustClip.StartRepAddr != EMPTY_FIELD) AdjustClip.StartRepAddr -= Adjustment;
				AdjustClip.EndAddr -= Adjustment;
				if (AdjustClip.EndRepAddr != EMPTY_FIELD) AdjustClip.EndRepAddr -= Adjustment;
				WriteIndex(AdjustClip, hSoundFile, indexclipNo);
				}
			// adjust quill data index if the quill clip was moved
			if (!QWCheckBox->Checked) { // we don't need to do this if we are processing a Quilled clip
				// read the first quill index
				QIndex QuillIndex;
				ParseQuillIndex(QuillIndex, hSoundFile, 1);
				if (QuillIndex.StartAddr != 0x000000 && QuillIndex.StartAddr != 0xFFFFFF) {   // make sure there is an index entry
					ADPCM_Main->MyShowMessage("DoIt: Quill Adjustment : " + String(Adjustment) + " bytes", false);
					// is it greater than the end address of the clip being replaced or deleted?
					if (QuillIndex.StartAddr >= IndexSound.EndAddr) {
						// yes adjust the Quill data
						QuillAddressFixup(hSoundFile, Adjustment); // Quill Data starts at byte offset 0x3200
						}
					}
				}
			//
			}
		// put the new clip over top of the old clip
		// allocate memory of the correct size for a buffer
		membuf = (char *) malloc(IndexSource.ClipLength);
		if (membuf == NULL) {
			ADPCM_Main->MyShowMessage("DoIt: membuf malloc failed", true);
			CloseHandle(hSoundFile);
			return;
			}
		// create the bytes of silence
		memset(membuf, '\0', IndexSource.ClipLength);
		// write to Sound at IndexSound.StartAddr
		// write the clip to the sound file
		// set file pointer to first byte of the seqment
		fp = IndexSound.StartAddr;
		SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
		// write length bytes
		WriteFile(hSoundFile, membuf, IndexSource.ClipLength, &writtenBytes, NULL);
		if (writtenBytes != IndexSource.ClipLength) {
			ADPCM_Main->MyShowMessage("DoIt: Could not write silent clip to sound file", true);
			CloseHandle(hSoundFile);
			free (membuf);
			return;
			}
		free (membuf);
		//-------------------------------------------------------------------- REPLACE_CLIP_WITH_SILENCE
		// create and write the index entry
		// transfer and update sound file index
		// new addresses, keep length, keep data byte, keep SMInit
		// IndexSound.Clip is not changed
		//
		// IndexSound.StartAddr is not changed because we are putting the new clip at same StartAddr
		if (IndexSource.StartRepAddr != EMPTY_FIELD) IndexSound.StartRepAddr = (IndexSource.StartRepAddr - IndexSource.StartAddr) + IndexSound.StartAddr;
		if (IndexSource.EndRepAddr != EMPTY_FIELD) IndexSound.EndRepAddr = (IndexSource.EndRepAddr - IndexSource.StartAddr) + IndexSound.StartAddr;

		IndexSound.EndAddr = IndexSound.StartAddr + IndexSource.ClipLength - 1;
		IndexSound.SMInit = IndexSource.SMInit;
		IndexSound.ClipLength = IndexSource.ClipLength;
		IndexSound.databyte = IndexSource.databyte;
		//
		IndexSound.SA_LOBit = IndexSource.SA_LOBit;
		IndexSound.SRA_LOBit = IndexSource.SRA_LOBit;
		IndexSound.ERA_LOBit = IndexSource.ERA_LOBit;
		IndexSound.EA_LOBit = IndexSource.EA_LOBit;
		//
		//
		// write the sound file index entry
		if (!WriteIndex(IndexSound, hSoundFile, ClipNo)) {
			ADPCM_Main->MyShowMessage("DoIt: Could not write silent clip index to sound file", true);
			CloseHandle(hSoundFile);
			return;
			}
		// search the index to see if this clip appears more than once
		for (int indexclipNo = 1 ; indexclipNo <= 255 ; indexclipNo ++) {
			if (indexclipNo == ClipNo) continue;	// don't reprocess ClipNo
			SFIndex AdjustClip;
			ParseIndex(AdjustClip, hSoundFile, indexclipNo);
			// if the start addresses match, write the new IndexSound into that clip index location
			if (IndexSound.StartAddr == AdjustClip.StartAddr) {
				// print message that another index entry must be changed
				ADPCM_Main->MyShowMessage("DoIt: Adjusting additional clip index : " + String(indexclipNo), false);
				// keep original volume control regime bits
				IndexSound.databyte = (IndexSound.databyte & 0xCF) | (AdjustClip.databyte & 0x30);
				WriteIndex(IndexSound, hSoundFile, indexclipNo);
				}
			}
		// Set new EOF if needed
		if (Adjustment != 0) {
			fp = FileLength - Adjustment;
			//			
			SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
			SetEndOfFile(hSoundFile);
			}
		//
		CloseHandle(hSoundFile);
		ADPCM_Main->FileSetDateNow(SoundFileName);
		//
		if (SoundFileName == ADPCM_Main->MRUComboBox->Text) {
			// Analyze only if filename has not changed since we started.
			ADPCM_Main->AnalyzeButtonClick(Sender);
			}
		break;
	default:
		ShowMessage("DoIt: This feature not implemented yet");
		break;
	}
Close();
return;
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::CancelButtonClick(TObject *Sender)
{
// cancel the operation
Close();
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::BrowseforFileButtonClick(TObject *Sender)
{
OpenDialog->Options.Clear();
OpenDialog->Options << ofFileMustExist;

#ifdef DONEELSEWHERE
// setup correct file query
if (FormatGroup->ItemIndex == PCM24) {
	OpenDialog->DefaultExt = "raw";
	OpenDialog->Filter = "Audio PCM24 Clip file (*.raw)|*.raw;*.RAW";
	} else {
	OpenDialog->DefaultExt = "ima";
	OpenDialog->Filter = "Audio ADPCM Clip file (*.ima)|*.ima;*.IMA";
	}
#endif

if (OpenDialog->Execute()) {
	// file picked
	MRUComboBox->Text = OpenDialog->FileName;
	MRUComboBox->Hint = MRUComboBox->Text;
	OpenDialog->InitialDir = ExtractFilePath(OpenDialog->FileName);

    // insert into MRU List
	// don't insert a duplicate string
    int idx = MRUComboBox->Items->IndexOf(MRUComboBox->Text);
    if (idx == -1) {
        MRUComboBox->Items->Insert(0, MRUComboBox->Text);
        MRUComboBox->ItemIndex = 0;
		if (MRUComboBox->Items->Count > MRUComboBox->DropDownCount) MRUComboBox->Items->Delete(MRUComboBox->DropDownCount);
		} else {
        // set ItemIndex
        MRUComboBox->ItemIndex = idx;
        }
	}
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::FormCreate(TObject *Sender)
{
// read Preferences from registry if they exist
TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
SourceClipNoEdit->Text = RegFile->ReadString("Options","SourceClipNoEdit", SourceClipNoEdit->Text);
OpGroup->ItemIndex = RegFile->ReadInteger("Options","OpGroup", OpGroup->ItemIndex);
//
//http://bcbjournal.org/articles/vol1/9711/Implementing_a_recent-files_list.htm
for (int idx = 0 ; idx < MRUComboBox->DropDownCount ; idx++) {
	String S = RegFile->ReadString("MRU Source File", "MRU" + String(idx + 1), "");
	if (S != "") {
        MRUComboBox->Items->Append(S);
        } else {
        break;
        }
    }
// next line changes MRUComboBox->Text
MRUComboBox->ItemIndex = RegFile->ReadInteger("MRU Source File", "MRUItemIndex", -1);
MRUComboBox->Hint = MRUComboBox->Text;
// Deallocate RegFile
delete RegFile;

//
if (MRUComboBox->Text.IsEmpty()) {
    // use WorkingDir
	OpenDialog->InitialDir = ADPCM_Main->cwd;
	MRUComboBox->Text = OpenDialog->InitialDir + "\\*.mth";
   	MRUComboBox->Hint = MRUComboBox->Text;
    } else {
	// use directory from Script FileName
	OpenDialog->InitialDir = ExtractFilePath(MRUComboBox->Text);
	MRUComboBox->Hint = MRUComboBox->Text;
	}

if (ADPCM_Main->NoSoundFileSizeChecks->Checked) {
	EngineEEPROMSize->Font->Color = clRed;
	} else {
	EngineEEPROMSize->Font->Color = clGreen;
	}

DragAcceptFiles (Handle, True);		//  Accept drag and drop from the Explorer
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::FormDestroy(TObject *Sender)
{
// write preferences to the Registry
TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
// User Registry Data
RegFile->WriteString("Options", "SourceClipNoEdit", SourceClipNoEdit->Text);
RegFile->WriteInteger("Options", "OpGroup", OpGroup->ItemIndex);

// erase MRU PC key
RegFile->EraseSection("MRU Source File");
 // then resave the MRU List
int Cnt = MRUComboBox->Items->Count;
if (Cnt > 0) {
    for (int idx = 0 ; idx < Cnt ; idx++) {
		RegFile->WriteString("MRU Source File", "MRU" + String(idx + 1), MRUComboBox->Items->Strings[idx]);
        }
    }
RegFile->WriteInteger("MRU Source File", "MRUItemIndex", MRUComboBox->ItemIndex);

// Deallocate RegFile
delete RegFile;
}
//---------------------------------------------------------------------------
void __fastcall TClip_Editor::WMDropFiles(TWMDropFiles &message)
{
// Get the number of files dropped
UINT filecount = DragQueryFile((HDROP) message.Drop, 0xFFFFFFFF, NULL, 0);
if (filecount > 0) {
	String FullFileName;
	for (UINT ii = 0 ; ii < filecount ; ++ii){
    	Application->ProcessMessages();
		// Get the file name
		TCHAR FileName[MAX_PATH];
		DragQueryFile ((HDROP) message.Drop, ii, FileName, MAX_PATH);
		FullFileName = FileName;
		//Application->MessageBox(FullFileName.c_str(), TEXT("Result of DragQueryFile()"), MB_OKCANCEL | MB_DEFBUTTON1);
		if (FileExists(FullFileName.c_str())) {
			// do something for each file
			ADPCM_Main->MyShowMessage("Dropped Edit Source file " + FullFileName, false);
			} else {
			continue;
			}
		}
	//String FileName = LowerCase(ExtractFileName(FullFileName));
	// use last dropped path/filename
	MRUComboBox->Text = FullFileName;
	MRUComboBox->Hint = MRUComboBox->Text;

    // Add new filename to DropDown List
    // insert into MRU List
	// don't insert a duplicate string
    int idx = MRUComboBox->Items->IndexOf(MRUComboBox->Text);
    if (idx == -1) {
        MRUComboBox->Items->Insert(0, MRUComboBox->Text);
        MRUComboBox->ItemIndex = 0;
        if (MRUComboBox->Items->Count > MRUComboBox->DropDownCount)
        MRUComboBox->Items->Delete(MRUComboBox->DropDownCount);
		} else {
        // set ItemIndex
        MRUComboBox->ItemIndex = idx;
		}
	}

Update();	// redraw Form
DragFinish ((HDROP) message.Drop);
QWCheckBox->Checked = false;
return;
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::SetDBCheckBoxes(int DB)
{
// Set the Format based on DB
if (DB & PCM24MASK)
	FormatGroup->ItemIndex = PCM24;
	else
	FormatGroup->ItemIndex = ADPCM;

// Set the Sample Rate based on DB
int Sample = -1;
if (PSGroup->ItemIndex == PS2) {
	// PS2
	if (DB & PCM24MASK) ADPCM_Main->MyShowMessage("PS3 clip in a PS2 sound file", true);
	//Sample = SampleRatePS2[DB & 0x7F]; // index 0-127
	// ADPCM
	if ((DB & 0x48) == 0x48) Sample = HZ5512;
	if ((DB & 0x48) == 0x08) Sample = HZ11025;
	if ((DB & 0x48) == 0x00) Sample = HZ22050;
	if (Sample == -1) ADPCM_Main->MyShowMessage("Can't decode the PS2 Sample Rate for Data Byte : 0x" + IntToHex(DB, 2), true);
	} else {
	// PS3
	//Sample = SampleRatePS3[DB & 0xFF]; // index 0-255
	// ADPCM
	if ((DB & 0xC8) == 0x48) Sample = HZ5512;
	if ((DB & 0xC8) == 0x08) Sample = HZ11025;
	if ((DB & 0xC8) == 0x00) Sample = HZ22050;
	// PCM
	if ((DB & 0xC8) == 0x88) Sample = HZ5512;
	if ((DB & 0xC8) == 0x80) Sample = HZ11025;
	if (Sample == -1) ADPCM_Main->MyShowMessage("Can't decode the PS3 Sample Rate for Data Byte : 0x" + IntToHex(DB, 2), true);
	}

if (Sample != -1) {
	SampleGroup->ItemIndex = Sample;
	} else {
	SampleGroup->ItemIndex = -1;    // can't decode the sample rate
	}
CategoryGroup->ItemIndex = (DB & 0x30) >> 4;
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::CopyrightClick(TObject *Sender)
{
PSGroup->Visible = true;
FormatGroup->Visible = true;
SampleGroup->Visible = true;
CategoryGroup->Visible = true;
BorderStyle = bsSizeable;
}
//---------------------------------------------------------------------------
/*
Data Byte

		bit 7 (high order bit)  0 = IMA ADPCM clip
								1 = 24 bit big endian signed PCM clip
		bit 6 - possibly, with bits 7 & 3, sets the sample rate
		bit 5-4 - appears to connect the clip with the volume control that controls it:
					00 - Engine sound clip
					01 - Accent sound clip
					10 - Horn/Whistle sound clip
					11 - Bell sound clip
		bit 3 - possibly, with bits 7 & 6, sets the sample rate
		bit 2 - unknown
		bits 1-0 - unknown, either "00" or "11", but mostly "11".
*/
void __fastcall TClip_Editor::FormatGroupClick(TObject *Sender)
{
if (FormatGroup->ItemIndex == -1) return;
int DB = _tcstol(DataByteEdit->Text.c_str(), NULL, 16);
switch(FormatGroup->ItemIndex) {
	case ADPCM:	// 4 bit IMA ADPCM
		DB &= 0x7F;
		OpenDialog->DefaultExt = "ima";
		OpenDialog->Filter = "Audio ADPCM Clip files (*.ima)|*.ima;*.IMA";
		break;
	case PCM24:	// 24 bit big endian signed PCM
		if (PSGroup->ItemIndex == PS3) {
			DB |= PCM24MASK;
			OpenDialog->DefaultExt = "raw";
			OpenDialog->Filter = "Audio PCM24 Clip files (*.raw)|*.raw;*.RAW";
			} else {
			ShowMessage("24 bit big endian signed PCM is not supported for PS2 clips");
			}
		break;
	}

DataByteEdit->Text = IntToHex(DB, 2);
SampleGroupClick(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::CategoryGroupClick(TObject *Sender)
{
if (CategoryGroup->ItemIndex == -1) return;
int DB = _tcstol(DataByteEdit->Text.c_str(), NULL, 16);
DB &= 0xCF;	// clear the two bit field
switch(CategoryGroup->ItemIndex) {
	case 0: // Engine
		DB |= 0x00;
		break;
	case 1:	// Accent
		DB |= 0x10;
		break;
	case 2:	// Horn/Whistle
		DB |= 0x20;
		break;
	case 3: // Bell
		DB |= 0x30;
		break;
	}
DataByteEdit->Text = IntToHex(DB, 2);
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::SampleGroupClick(TObject *Sender)
{
if (SampleGroup->ItemIndex == -1) return;
int DB = _tcstol(DataByteEdit->Text.c_str(), NULL, 16);
DB &= 0xB7; // clear the two bits that are not used in the Sample Rate determination
if (PSGroup->ItemIndex == PS3) {
	if (FormatGroup->ItemIndex == PCM24) {
		// 24 bit big endian signed PCM -------------------------------------
		switch(SampleGroup->ItemIndex) {
			case HZ5512: 	// 5512Hz
				DB |= 0x88;
				break;
			case HZ11025:	// 11025 Hz
				DB |= 0x80;
				break;
			case HZ22050:	// 22050 Hz
				ShowMessage("22,050 Hx sample rate is not supported for 24 bit PCM clips");
				break;
			default:
				ShowMessage("That sample rate is not supported");
				//SampleGroup->ItemIndex = -1;
				break;
			}
		} else {
		// 4 bit IMA ADPCM and default_-----------------_---------------------
		switch(SampleGroup->ItemIndex) {
			case HZ5512: 	// 5512Hz
				DB |= 0x48;
				break;
			case HZ11025:	// 11025 Hz
				DB |= 0x08;
				break;
			case HZ22050:	// 22050 Hz
				DB |= 0x00;
				break;
			default:
				ShowMessage("That sample rate is not supported");
				//SampleGroup->ItemIndex = -1;
				break;
			}
		}
	} else {
	if (FormatGroup->ItemIndex == ADPCM) {
		// 4 bit IMA ADPCM
		switch(SampleGroup->ItemIndex) {
			case HZ5512: 	// 5512Hz
				DB |= 0x48;
				break;
			case HZ11025:	// 11025 Hz
				DB |= 0x08;
				break;
			case HZ22050:	// 22050 Hz
				DB |= 0x00;
				break;
			default:
				ShowMessage("That sample rate is not supported");
				//SampleGroup->ItemIndex = -1;
				break;
			}
		}
	}
DataByteEdit->Text = IntToHex(DB, 2);
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::PSGroupClick(TObject *Sender)
{
// Click
int DB = _tcstol(DataByteEdit->Text.c_str(), NULL, 16);
if ((FormatGroup->ItemIndex == PCM24) && (PSGroup->ItemIndex == PS3)) {
	DB |= PCM24MASK;
	DataByteEdit->Text = IntToHex(DB, 2);
	SampleGroupClick(Sender);
	return;
	}
if ((FormatGroup->ItemIndex == PCM24) && (PSGroup->ItemIndex == PS2)) {
	ShowMessage("24 bit big endian signed PCM is not supported for PS2 clips");
	return;
	}
// else ADPCM
DB &= 0x7F;
DataByteEdit->Text = IntToHex(DB, 2);
SampleGroupClick(Sender);
return;
}
//---------------------------------------------------------------------------
void __fastcall TClip_Editor::SetORFileByte(HANDLE hFile, int Address, unsigned char Value) {
// OR in the Value at the offset Address
unsigned char membuf[2];
unsigned long readBytes = 0, writtenBytes = 0;
// read the byte
SetFilePointer(hFile, Address, NULL, FILE_BEGIN);
ReadFile(hFile, membuf, 1L, &readBytes, NULL);
if (readBytes != 1L) {
	ADPCM_Main->MyShowMessage("SetORFileByte: Could not read byte from sound file", true);
	}
membuf[0] |= Value;
// write the byte
SetFilePointer(hFile, Address, NULL, FILE_BEGIN);
WriteFile(hFile, membuf, 1L, &writtenBytes, NULL);
if (writtenBytes != 1L) {
	ADPCM_Main->MyShowMessage("SetORFileByte: Could not write byte to sound file", true);
	}
}
//---------------------------------------------------------------------------
void __fastcall TClip_Editor::SetANDFileByte(HANDLE hFile, int Address, unsigned char Value) {
// AND in the Value at the offset Address
unsigned char membuf[2];
unsigned long readBytes = 0, writtenBytes = 0;
// read the byte
SetFilePointer(hFile, Address, NULL, FILE_BEGIN);
ReadFile(hFile, membuf, 1L, &readBytes, NULL);
if (readBytes != 1L) {
	ADPCM_Main->MyShowMessage("SetANDFileByte: Could not read byte from sound file", true);
	}
membuf[0] &= Value;
// write the byte
SetFilePointer(hFile, Address, NULL, FILE_BEGIN);
WriteFile(hFile, membuf, 1L, &writtenBytes, NULL);
if (writtenBytes != 1L) {
	ADPCM_Main->MyShowMessage("SetANDFileByte: Could not write byte to sound file", true);
	}
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::CopyBlock(HANDLE hSourceFile, HANDLE hDestFile, int StartAddr1, int EndAddr1, int StartAddr2)
{
// Copy a block from one sound file to another
// Copy bytes from StartAdd1 to EndAddr inclusive to StartAddr2
DWORD readBytes, writtenBytes;
DWORD Len = EndAddr1 - StartAddr1 + 1;
char * membuf;
// allocate memory of the correct size for a buffer
membuf = (char *) malloc(Len);
if (membuf == NULL) {
	ADPCM_Main->MyShowMessage("CopyBlock: membuf malloc failed", true);
	return;
	}
// read from Source from StartAddr1 for Len
SetFilePointer(hSourceFile, StartAddr1, NULL, FILE_BEGIN);
// read Len bytes
ReadFile(hSourceFile, membuf, Len, &readBytes, NULL);
if (readBytes != Len) {
	ADPCM_Main->MyShowMessage("CopyBlock: Could not read clip from source file", true);
	free(membuf);
	return;
	}
// write to Sound at StartAddr2
// set file pointer to first byte of the seqment
SetFilePointer(hDestFile, StartAddr2, NULL, FILE_BEGIN);
// write Len bytes
WriteFile(hDestFile, membuf, Len, &writtenBytes, NULL);
if (writtenBytes != Len) {
	ADPCM_Main->MyShowMessage("CopyBlock: Could not write clip to sound file", true);
	free (membuf);
	return;
	}
free (membuf);
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::SourceClipNoEditDblClick(TObject *Sender)
{
// on double click, use the same clip number for the source clip
SourceClipNoEdit->Text = ClipNo;
}
//---------------------------------------------------------------------------
void __fastcall TClip_Editor::QuillAddressFixup(HANDLE hSoundFile, int Adjustment)
{
ADPCM_Main->MyShowMessage("QuillAddressFixup: Adjustment " + String(Adjustment), false);
if (Adjustment == 0) return;	// nothing to do
//
// Fixup the addresses in the Quill Data by Adjustment bytes
//
DWORD readBytes, writtenBytes;
const DWORD Len = 0x0D00;
byte * membuf;
// allocate memory of the correct size for a buffer
membuf = (byte *) malloc(Len);
if (membuf == NULL) {
	ADPCM_Main->MyShowMessage("QuillAddressFixup: membuf malloc failed", true);
	return;
	}
// read from Source from Quill Data Offset for Len
SetFilePointer(hSoundFile, 0x3200, NULL, FILE_BEGIN);
// read Len bytes
ReadFile(hSoundFile, membuf, Len, &readBytes, NULL);
if (readBytes != Len) {
	ADPCM_Main->MyShowMessage("QuillAddressFixup: Could not read quill data from sound file", true);
	free(membuf);
	return;
	}
// Fixup up the Quill pointers in the memory buffer
//00003200  00 01 17 2B 00 B1 7E 00 00 B1 E1 80 00 00 00 00  ...+.~......
//                      --------    --------
//					     0xB17E      0xB1E1   <--- Actual Sound file byte addresss

// There are 0xD0 index entries
for (DWORD offset = 0 ; offset < Len ; offset += 0x10) {
	int StartAddr = membuf[offset + 4] << 16 | membuf[offset + 5] << 8 | membuf[offset + 6];
	if (StartAddr == 0x000000 || StartAddr == 0xFFFFFF) continue;   // 0x00 or 0xFFFFFF doesn't get fixup
	StartAddr -= Adjustment;
	membuf[offset + 4] = (StartAddr & 0xFF0000) >> 16;
	membuf[offset + 5] = (StartAddr & 0x00FF00) >> 8;
	membuf[offset + 6] = (StartAddr & 0x0000FF);
    //
	int EndAddr = membuf[offset + 8] << 16 | membuf[offset + 9] << 8 | membuf[offset + 10];
	EndAddr -= Adjustment;
	membuf[offset + 8] = (EndAddr & 0xFF0000) >> 16;
	membuf[offset + 9] = (EndAddr & 0x00FF00) >> 8;
	membuf[offset + 10] = (EndAddr & 0x0000FF);
	}

// write to Sound at Quill Data Offset
// set file pointer to first byte of the seqment
SetFilePointer(hSoundFile, 0x3200, NULL, FILE_BEGIN);
// write Len bytes
WriteFile(hSoundFile, membuf, Len, &writtenBytes, NULL);
if (writtenBytes != Len) {
	ADPCM_Main->MyShowMessage("QuillAddressFixup: Could not write quill data to sound file", true);
	free (membuf);
	return;
	}
free (membuf);
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::RemoveFromListClick(TObject *Sender)
{
MRUComboBox->Items->Delete(MRUComboBox->ItemIndex);
MRUComboBox->ItemIndex = -1;	// nothing selected	
}
//---------------------------------------------------------------------------

void __fastcall TClip_Editor::CleartheList1Click(TObject *Sender)
{
// clear the list
MRUComboBox->Text = "";
MRUComboBox->Items->Clear();
}
//---------------------------------------------------------------------------

