//---------------------------------------------------------------------------
/*
ADPCM Program for Windows

 Copyright 2020 by Mark DiVecchio

    This file is part of Remote Train Control.

    Remote Train Control is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Remote Train Control is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
    along with Remote Train Control.  If not, see <http://www.gnu.org/licenses/>.

*/
// Home Page: http://www.silogic.com/trains/ADPCM.html
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include <tchar.h>
#include "ADPCM_Player.h"
#include "Edit_the_Engine_Parameters.h"
#include "assert.h"
#include <stdio.h>
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TEdit_Engine_Parameters *Edit_Engine_Parameters;
//---------------------------------------------------------------------------
__fastcall TEdit_Engine_Parameters::TEdit_Engine_Parameters(TComponent* Owner)
	: TForm(Owner)
{
assert(false);	// never called
}
//---------------------------------------------------------------------------

THexEdit * __fastcall TEdit_Engine_Parameters::CreateHexEdit(TMaskEdit * TemplateEdit, int minval, int maxval)
{
THexEdit * NewHexEdit = new THexEdit(this);
assert(NewHexEdit != NULL);
NewHexEdit->Parent = this;
NewHexEdit->Height = TemplateEdit->Height;
NewHexEdit->Width = TemplateEdit->Width;
NewHexEdit->Top = TemplateEdit->Top;
NewHexEdit->Left = TemplateEdit->Left;
NewHexEdit->MinValue = minval;
NewHexEdit->MaxValue = maxval;
NewHexEdit->Digits = 4;
NewHexEdit->MaxLength = 4;
NewHexEdit->Text = "00.00";
NewHexEdit->Cursor = crHandPoint;
NewHexEdit->ShowHint = true;
NewHexEdit->Hint = TemplateEdit->Hint;
if (!NewHexEdit->Hint.IsEmpty()) {
	NewHexEdit->ShowHint = true;
    }
NewHexEdit->TabOrder = TabOrder++;
NewHexEdit->OnChange = TemplateEdit->OnChange;
NewHexEdit->Font = TemplateEdit->Font;
NewHexEdit->ReadOnly = TemplateEdit->ReadOnly;
NewHexEdit->Show();

return NewHexEdit;
}
//---------------------------------------------------------------------------

__fastcall TEdit_Engine_Parameters::TEdit_Engine_Parameters(TComponent* Owner, String FN, int p1, int p2, int PSType)
	: TForm(Owner)
{
TargetFileName = FN;
assert(FN != "");
assert(PSType == PS2 || PSType == PS3);
MyPSType = PSType;

TabOrder = 0;

// 0x04-0x05
XUnknown0405Edit = CreateHexEdit(XUnknown0405, 0x00000000, 0x0000FFFF);
// 0x06-0x07
XCountsperRevEdit = CreateHexEdit(XCountsperRev, 0x00000000, 0x0000FFFF);
// 0x08-0x09
XCountsper10feetEdit = CreateHexEdit(XCountsper10feet, 0x00000000, 0x0000FFFF);
// 0x18-0x19
XCountsperChuffEdit = CreateHexEdit(XCountsperChuff, 0x00000000, 0x0000FFFF);
// 0x0A-0x0B
XUnknown0A0BEdit = CreateHexEdit(XUnknown0A0B, 0x00000000, 0x0000FFFF);
// 0x0C-0x0D
XEngineNumberEdit = CreateHexEdit(XEngineNumber, 0x00000000, 0x0000FFFF);
// 0x64-0x65
XChecksumEdit = CreateHexEdit(XChecksum, 0x00000000, 0x0000FFFF);
// 0x22-0x23
XEngineTypeEdit = CreateHexEdit(XEngineType, 0x00000000, 0x0000FFFF);
XEngineTypeEdit->Visible = EngTypeLabel->Visible;

ReloadButtonClick(NULL);
}
//---------------------------------------------------------------------------

__fastcall TEdit_Engine_Parameters::~TEdit_Engine_Parameters()
{
// parent of dyanmically added control will free it
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::ShowIt(String FN, int p1, int p2, int PSType)
{
TargetFileName = FN;
assert(FN != "");
assert(PSType == PS2 || PSType == PS3);
MyPSType = PSType;

Caption = "Editing Engine Parameters";
//
Show();
Repaint();
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::CancelButtonClick(TObject *Sender)
{
Close();    // cancel the operation
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::ReloadButtonClick(TObject *Sender)
{
// clear the calculated values and restore original Sound File values

iNStripes->Text = 24;
ChangesMade = false;

XCountsper10feetEdit->Text = "";
XUnknown0405Edit->Text = "";
XCountsperChuffEdit->Text = "";
XCountsperRevEdit->Text = "";
XUnknown0A0BEdit->Text = "";
XChecksumEdit->Text = "";
XCountsperChuffEdit->Text = "0";
XEngineNumberEdit->Text = "";

WheelDiameter->Text = "00.00";
GearRatio->Text = "00.00";
Countsperfoot->Text = "00.00";
ChuffsperRev->Text = "0";
fClockRate->Text = "00.00000";

LoadParameters(TargetFileName);	
Repaint();
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::LoadParameters(String TargetFileName)
{

HANDLE hSoundFile;
// try and open the file
if((hSoundFile = CreateFile(TargetFileName.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	ADPCM_Main->MyShowLastError("LoadParameters: Unable to open file : " + TargetFileName, true);
	Close();    // cancel the operation
	} else {
	ADPCM_Main->MyShowMessage("Edit Engine Parameters: Sound file to be edited : " + TargetFileName, false);
	}

unsigned char buffer16[2];
DWORD readBytes;
//-------------------------------------------------------------
// read the 2 bytes at 0x04-0x05 Wheel Diameter (tach reload)
SetFilePointer(hSoundFile, 0x0004, NULL, FILE_BEGIN);
// read 2 bytes
ReadFile(hSoundFile, buffer16, sizeof(buffer16), &readBytes, NULL);
if (readBytes != 2) {
	ADPCM_Main->MyShowMessage("LoadParameters: Read did not read 2 bytes 0x04-0x05", true);
	return;
	}
// convert big endian to little endian
XUnknown0405Edit->Text = IntToHex((buffer16[0] << 8) | buffer16[1], 4);
// Convert the hex value to decimal and display it
// done in CalcWheelDiameterClick()
//-------------------------------------------------------------
// read the 2 bytes at 0x06-0x07  Counts/Revolution (gr constant) - used to calculate Gear Ratio
SetFilePointer(hSoundFile, 0x0006, NULL, FILE_BEGIN);
ReadFile(hSoundFile, buffer16, sizeof(buffer16), &readBytes, NULL);
if (readBytes != 2) {
	ADPCM_Main->MyShowMessage("LoadParameters: Read did not read 2 bytes 0x06-0x07", true);
	return;
	}
// convert big endian to little endian
XCountsperRevEdit->Text = IntToHex((buffer16[0] << 8) | buffer16[1], 4);
// Convert the hex value to decimal and display it
int NS = _tcstol(iNStripes->Text.c_str(), NULL, 10);	// 24
float fGR = _tcstol(XCountsperRevEdit->Text.c_str(), NULL, 16) / (2 * NS);	// eg 0x504/48 = 10.5
GearRatio->Text = Format("%5.2f", ARRAYOFCONST((fGR)));

//-------------------------------------------------------------
// read the 2 bytes at 0x08-0x09 Counts/foot * 10
SetFilePointer(hSoundFile, 0x0008, NULL, FILE_BEGIN);
ReadFile(hSoundFile, buffer16, sizeof(buffer16), &readBytes, NULL);
if (readBytes != 2) {
	ADPCM_Main->MyShowMessage("LoadParameters: Read did not read 2 bytes 0x08-0x09", true);
	return;
	}
// convert big endian to little endian
XCountsper10feetEdit->Text = IntToHex((buffer16[0] << 8) | buffer16[1], 4);
// Convert the hex value to decimal and display it
float fCount = _tcstol(XCountsper10feetEdit->Text.c_str(), NULL, 16) / 10.0;
Countsperfoot->Text = Format("%8.1f", ARRAYOFCONST((fCount)));
//
//-------------------------------------------------------------
// read the 2 bytes at 0x18-0x19 Counts/Chuff
SetFilePointer(hSoundFile, 0x0018, NULL, FILE_BEGIN);
ReadFile(hSoundFile, buffer16, sizeof(buffer16), &readBytes, NULL);
if (readBytes != 2) {
	ADPCM_Main->MyShowMessage("LoadParameters: Read did not read 2 bytes 0x18-0x19", true);
	return;
	}
// convert big endian to little endian
XCountsperChuffEdit->Text = IntToHex((buffer16[0] << 8) | buffer16[1], 4);
// Convert the hex value to decimal and display it
int Chuffs_Rev = _tcstol(XCountsperChuffEdit->Text.c_str(), NULL, 16);
if (Chuffs_Rev != 0) {
	// Convert the hex value to decimal and display it
	int iChuff = _tcstol(XCountsperRevEdit->Text.c_str(), NULL, 16) / Chuffs_Rev; 		// eg 0x360/0xD8 = 4
	ChuffsperRev->Text = Format("%02d", ARRAYOFCONST((iChuff)));
	} else {
	ChuffsperRev->Text = "00.00";
	}
//-------------------------------------------------------------

// read the 2 bytes at 0x0A-0x0B Interrupt Rate
SetFilePointer(hSoundFile, 0x000A, NULL, FILE_BEGIN);
ReadFile(hSoundFile, buffer16, sizeof(buffer16), &readBytes, NULL);
if (readBytes != 2) {
	ADPCM_Main->MyShowMessage("LoadParameters: Read did not read 2 bytes 0x0A-0x0B", true);
	return;
	}
// convert big endian to little endian
XUnknown0A0BEdit->Text = IntToHex((buffer16[0] << 8) | buffer16[1], 4);
// Convert the hex value to decimal and display it
float Clock = 1000000.0 / _tcstol(XUnknown0A0BEdit->Text.c_str(), NULL, 16);
fClockRate->Text = Format("%8.5f", ARRAYOFCONST((Clock)));

//-------------------------------------------------------------
// read the 2 bytes at 0x0C-0x0D Engine Number
SetFilePointer(hSoundFile, 0x000C, NULL, FILE_BEGIN);
ReadFile(hSoundFile, buffer16, sizeof(buffer16), &readBytes, NULL);
if (readBytes != 2) {
	ADPCM_Main->MyShowMessage("LoadParameters: Read did not read 2 bytes 0x0C-0x0D", true);
	return;
	}
// convert big endian to little endian
XEngineNumberEdit->Text = IntToHex((buffer16[0] << 8) | buffer16[1], 4);
// Convert the hex value to decimal and display it
int EngNum = _tcstol(XEngineNumberEdit->Text.c_str(), NULL, 16) - 1;
fEngNum->Text = Format("%d", ARRAYOFCONST((EngNum)));

//-------------------------------------------------------------
// read the 2 bytes at 0x22-0x23 engine type
SetFilePointer(hSoundFile, 0x0022, NULL, FILE_BEGIN);
ReadFile(hSoundFile, buffer16, sizeof(buffer16), &readBytes, NULL);
if (readBytes != 2) {
	ADPCM_Main->MyShowMessage("LoadParameters: Read did not read 2 bytes 0x22-0x23", true);
	return;
	}
// convert big endian to little endian
XEngineTypeEdit->Text = IntToHex((buffer16[0] << 8) | buffer16[1], 4);
//-------------------------------------------------------------

// read the 2 bytes at 0x64-0x64 checksum
SetFilePointer(hSoundFile, 0x0064, NULL, FILE_BEGIN);
ReadFile(hSoundFile, buffer16, sizeof(buffer16), &readBytes, NULL);
if (readBytes != 2) {
	ADPCM_Main->MyShowMessage("LoadParameters: Read did not read 2 bytes 0x64-0x65", true);
	return;
	}
// convert big endian to little endian
XChecksumEdit->Text = IntToHex((buffer16[0] << 8) | buffer16[1], 4);
//-------------------------------------------------------------
CloseHandle(hSoundFile);

CalcWheelDiameter();
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::CalcWheelDiameter(void)
{
// Create the Wheel Diameter from
// gear ratio, and number of stripes

double Counts = _tcstod(Countsperfoot->Text.c_str(), NULL);	// bytes 0x08-0x09
int NS = _tcstol(iNStripes->Text.c_str(), NULL, 10);	// 24
double GRraw = _tcstol(XCountsperRevEdit->Text.c_str(), NULL, 16);	// bytes 0x06-0x07
int K;
double TS;
float WD;

// This equation calculates the Wheel Diameter in mm, eg 22.60
// Counts is counts per foot, the value at 0x08-0x09 divided by 10
// 5280 feet per mile
// 12 inches per foot
// 48 "O" Scale
// 25.4 mm per inch
// M_PI is 3.14159265358979323846264338327950288
// NS Number of stripes on the tach strip, usually 24
// GR Gear Ratio, eg  10.50, 18.00, etc. Calculated by GRraw/(2 * NS)
// GRraw is the number of counts for one revolution of driving wheel, the value at 0x06-0x07
//
//
K = (12 * 5280/48);	// 1320 inches/Smile
// multiplier of 2 seems to be because both the
// leading edge and the trailing edge of each stripe generates a count
//TS = 2 * ((K)/((WD / 25.4) * M_PI)) * NS * GR;
TS = Counts * 11.0;
if (TS != 0.0) {
	WD = (2 * NS * (GRraw/(2 * NS)) * 25.4 * K)/(TS * M_PI);
    } else {
    WD = 0.0;
    }
WD /= 10.0;
/*
WD = ((2 * NS * (GRraw/(2 * NS)) * 25.4 * (12 * 5280/48))/((Counts * 11.0) * M_PI))/10
*/
WheelDiameter->Text = Format("%6.2f", ARRAYOFCONST((WD)));

float WD1;
// Second way to calculate the Wheel Diameter
if (Counts != 0) {
	WD1 = (GRraw * 12 * 25.4)/(Counts * M_PI);
    } else {
    WD1 = 0.0;
    }
WheelDiameter->Text = Format("%6.2f", ARRAYOFCONST((WD1)));
//
//ADPCM_Main->MyShowMessage(Format("Wheel Diameter : %5.2f", ARRAYOFCONST((WD1))), false);
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::CalcCountsperFoot(void)
{
// Calculate (Count/foot)*10 from a change in Wheel Diameter and/or Gear Ratio

float fWD, fCounts;
int iGRraw;
try {
	iGRraw = _tcstol(XCountsperRevEdit->Text.c_str(), NULL, 16);
	fWD = WheelDiameter->Text.ToDouble();
    }
catch(...) {
    iGRraw = 0;
    fWD = 0.0;
	}
if (iGRraw != 0.0 && fWD != 0.0) fCounts = (iGRraw * 12 * 25.4)/(fWD * M_PI);
// set Counts/foot
if (fCounts != 0.0) Countsperfoot->Text = Format("%8.1f", ARRAYOFCONST((fCounts)));
// set (Counts/foot)*10
int iCount = (fCounts * 10.0) + 0.5;
if (fCounts != 0.0) XCountsper10feetEdit->Text = IntToHex(iCount, 4);
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::DoItButtonClick(TObject *Sender)
{

if (!ChangesMade) {
	ADPCM_Main->MyShowMessage("DoIt: No parameters were changed, nothing new to write", true);
	return;
	}
HANDLE hSoundFile;
// try and open the file
if((hSoundFile = CreateFile(TargetFileName.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	ADPCM_Main->MyShowLastError("DoIt: Unable to open file : " + TargetFileName, true);
	Close(); // cancel the operation
	}

unsigned char buffer16[2];
DWORD byteswritten;
//-------------------------------------------------------------
// write the 2 bytes at 0x04-0x05 Unknown
SetFilePointer(hSoundFile, 0x0004, NULL, FILE_BEGIN);
// convert little endian to big endian
buffer16[1] =  _tcstol(XUnknown0405Edit->Text.SubString(3,2).c_str(), NULL, 16);	// LSB
buffer16[0] =  _tcstol(XUnknown0405Edit->Text.SubString(1,2).c_str(), NULL, 16);	// MSB

WriteFile(hSoundFile, buffer16, sizeof(buffer16), &byteswritten, NULL);
if (byteswritten != sizeof(buffer16)) {
	ADPCM_Main->MyShowMessage("DoIt: Write did not write 2 bytes of 0x04-0x05", true);
	return;
	}
//-------------------------------------------------------------
// write the 2 bytes at 0x06-0x07  Counts/Revolution used to calculate Gear Ratio
SetFilePointer(hSoundFile, 0x0006, NULL, FILE_BEGIN);
// convert little endian to big endian
buffer16[1] =  _tcstol(XCountsperRevEdit->Text.SubString(3,2).c_str(), NULL, 16);	// LSB
buffer16[0] =  _tcstol(XCountsperRevEdit->Text.SubString(1,2).c_str(), NULL, 16);	// MSB

WriteFile(hSoundFile, buffer16, sizeof(buffer16), &byteswritten, NULL);
if (byteswritten != sizeof(buffer16)) {
	ADPCM_Main->MyShowMessage("DoIt: Write did not write 2 bytes 0x06-0x07", true);
	return;
	}
//-------------------------------------------------------------
// write the 2 bytes at 0x08-0x09 Counts/foot * 10
SetFilePointer(hSoundFile, 0x0008, NULL, FILE_BEGIN);
// convert little endian to big endian
buffer16[1] =  _tcstol(XCountsper10feetEdit->Text.SubString(3,2).c_str(), NULL, 16);	// LSB
buffer16[0] =  _tcstol(XCountsper10feetEdit->Text.SubString(1,2).c_str(), NULL, 16);	// MSB

WriteFile(hSoundFile, buffer16, sizeof(buffer16), &byteswritten, NULL);
if (byteswritten != sizeof(buffer16)) {
	ADPCM_Main->MyShowMessage("DoIt: Write did not write 2 bytes 0x08-0x09", true);
	return;
	}
//-------------------------------------------------------------
// write the 2 bytes at 0x18-0x19 Counts per Chuff
SetFilePointer(hSoundFile, 0x0018, NULL, FILE_BEGIN);
// convert little endian to big endian
buffer16[1] =  _tcstol(XCountsperChuffEdit->Text.SubString(3,2).c_str(), NULL, 16);	// LSB
buffer16[0] =  _tcstol(XCountsperChuffEdit->Text.SubString(1,2).c_str(), NULL, 16);	// MSB

WriteFile(hSoundFile, buffer16, sizeof(buffer16), &byteswritten, NULL);
if (byteswritten != sizeof(buffer16)) {
	ADPCM_Main->MyShowMessage("DoIt: Write did not write 2 bytes 0x18-0x19", true);
	return;
	}
//-------------------------------------------------------------
// write the 2 bytes at 0x0A-0x0B Unknown
SetFilePointer(hSoundFile, 0x000A, NULL, FILE_BEGIN);
// convert little endian to big endian
buffer16[1] =  _tcstol(XUnknown0A0BEdit->Text.SubString(3,2).c_str(), NULL, 16);	// LSB
buffer16[0] =  _tcstol(XUnknown0A0BEdit->Text.SubString(1,2).c_str(), NULL, 16);	// MSB

WriteFile(hSoundFile, buffer16, sizeof(buffer16), &byteswritten, NULL);
if (byteswritten != sizeof(buffer16)) {
	ADPCM_Main->MyShowMessage("DoIt: Write did not write 2 bytes 0x0A-0x0B", true);
	return;
	}
//-------------------------------------------------------------
// write the 2 bytes at 0x0C-0x0D DCS Engine #
SetFilePointer(hSoundFile, 0x000C, NULL, FILE_BEGIN);
// convert little endian to big endian
buffer16[1] =  _tcstol(XEngineNumberEdit->Text.SubString(3,2).c_str(), NULL, 16);	// LSB
buffer16[0] =  _tcstol(XEngineNumberEdit->Text.SubString(1,2).c_str(), NULL, 16);	// MSB

WriteFile(hSoundFile, buffer16, sizeof(buffer16), &byteswritten, NULL);
if (byteswritten != sizeof(buffer16)) {
	ADPCM_Main->MyShowMessage("DoIt: Write did not write 2 bytes 0x0C-0x0D", true);
	return;
	}
//-------------------------------------------------------------// write the 2 bytes at 0x22-0x23 Engine Type
SetFilePointer(hSoundFile, 0x0022, NULL, FILE_BEGIN);
// convert little endian to big endian
buffer16[1] =  _tcstol(XEngineTypeEdit->Text.SubString(3,2).c_str(), NULL, 16);	// LSB
buffer16[0] =  _tcstol(XEngineTypeEdit->Text.SubString(1,2).c_str(), NULL, 16);	// MSB

WriteFile(hSoundFile, buffer16, sizeof(buffer16), &byteswritten, NULL);
if (byteswritten != sizeof(buffer16)) {
	ADPCM_Main->MyShowMessage("DoIt: Write did not write 2 bytes 0x22-0x23", true);
	return;
	}
//-------------------------------------------------------------
ChecksumGenerate(hSoundFile);	// and write it to the file
//
CloseHandle(hSoundFile);
ADPCM_Main->FileSetDateNow(TargetFileName);
ChangesMade = false;
return;
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::CountsperfootChange(TObject *Sender)
{
int iCount;
if (Countsperfoot->Modified) {
// convert the decimal value to hex and display it
    try {
	    float fCount = Countsperfoot->Text.ToDouble();
        iCount = (fCount * 10.0) + 0.5;
        ChangesMade = true;
        }
    catch (...) {
        iCount = 0;
        }
    if (iCount != 0) XCountsper10feetEdit->Text = IntToHex(iCount, 4);
    CalcWheelDiameter();
    }
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::XCountsper10feetEditChange(TObject *Sender)
{
float fCount;
if (XCountsper10feetEdit->Modified) {
	// Convert the hex value to decimal and display it
    try {
        fCount = _tcstol(XCountsper10feetEdit->Text.c_str(), NULL, 16) / 10.0;
        ChangesMade = true;
        }
    catch(...) {
        fCount = 0.0;
        }
    if (fCount != 0.0) Countsperfoot->Text = Format("%8.1f", ARRAYOFCONST((fCount)));
    CalcWheelDiameter();
    }
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::WheelDiameterChange(TObject *Sender)
{
float iWD;
if (WheelDiameter->Modified) {
    CalcCountsperFoot();
    }
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::Xunknown0405EditChange(TObject *Sender)
{
// convert to decimal and display
if (XUnknown0405Edit->Modified) {
    ChangesMade = true;
    }
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::GearRatioChange(TObject *Sender)
{
int iCR;
if (GearRatio->Modified) {
    try {
		// convert the decimal value to hex and display it
        iCR = (GearRatio->Text.ToDouble() * 48.0) + 0.5;	// eg 10.5 * 48 = 504
        ChangesMade = true;
        }
    catch (...) {
        iCR = 0;
        }
    if (iCR != 0) XCountsperRevEdit->Text = IntToHex(iCR, 4);
    CalcWheelDiameter();
    }
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::XCountsperRevEditChange(TObject *Sender)
{
//
float fCR;
if (XCountsperRevEdit->Modified) {
	// Convert the hex value to decimal and display it
	try {
		fCR = _tcstol(XCountsperRevEdit->Text.c_str(), NULL, 16) / 48.0;	// eg 0x504/48 = 10.5
        ChangesMade = true;
    	}
	catch (...) {
    	fCR = 0.0;
        }
	if (fCR != 0.0) GearRatio->Text = Format("%5.2f", ARRAYOFCONST((fCR)));
	CalcWheelDiameter();
    }
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::ChuffsperRevChange(TObject *Sender)
{
int iCountsperChuff;
if (ChuffsperRev->Modified) {
    try {
		// convert the decimal value to hex and display it
		int iCountsperRrev = _tcstol(XCountsperRevEdit->Text.c_str(), NULL, 16);
	    ChangesMade = true;
        int iChuffsRev = ChuffsperRev->Text.ToInt();
        if (iChuffsRev != 0) {
	        iCountsperChuff = iCountsperRrev/iChuffsRev;
            } else {
            iCountsperChuff = 0;
            }
        }
    catch (...) {
        iCountsperChuff = 0;
        }
    if (iCountsperChuff != 0) XCountsperChuffEdit->Text = IntToHex(iCountsperChuff, 4);
    }
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::XCountsperChuffChange(
      TObject *Sender)
{
//
int iChuffs_Rev, iChuff;
if (XCountsperChuffEdit->Modified) {
	try {
        // Convert the hex value to decimal and display it
        iChuffs_Rev = _tcstol(XCountsperChuffEdit->Text.c_str(), NULL, 16);
        ChangesMade = true;
        if (iChuffs_Rev != 0) {
            // Convert the hex value to decimal and display it
            iChuff = _tcstol(XCountsperRevEdit->Text.c_str(), NULL, 16) / iChuffs_Rev; 		// eg 0x360/0xD8 = 4
            } else {
			iChuff = 0;
            }
		}
    catch(...){
    	iChuff = 0;
        }
    if (iChuff != 0) ChuffsperRev->Text = Format("%02d", ARRAYOFCONST((iChuff)));
    }
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::fClockRateChange(TObject *Sender)
{
//
int iClock;
if (fClockRate->Modified) {
	// convert the decimal value to hex and display it
    try {
        iClock = (fClockRate->Text.ToDouble() * 1000000.0) + 0.5;
        ChangesMade = true;
        }
    catch (...) {
        iClock = 0;
        }
    if (iClock != 0) XUnknown0A0BEdit->Text = IntToHex(iClock, 4);
    }
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::XUnknown0A0BChange(TObject *Sender)
{
//
float fClock;
if (XUnknown0A0BEdit->Modified) {
	// Convert the hex value to decimal and display it
    try {
	    long iClock = _tcstol(XUnknown0A0BEdit->Text.c_str(), NULL, 16);
	    ChangesMade = true;
	    if (iClock != 0) {
			fClock = 1000000.0 / iClock;
	        } else {
	        fClock = 0.0;
	        }
        }
    catch(...) {
    	fClock = 0.0;
        }
	if (fClock != 0.0) fClockRate->Text = Format("%8.5f", ARRAYOFCONST((fClock)));
    }
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::iNStripesChange(TObject *Sender)
{
//
if (iNStripes->Modified) {
	CalcWheelDiameter();
    CalcCountsperFoot();
    ChangesMade = true;
    }
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::XEngineTypeEditChange(TObject *Sender)
{
//
if (XEngineTypeEdit->Modified) {
    ChangesMade = true;
    }
}
//---------------------------------------------------------------------------
void __fastcall TEdit_Engine_Parameters::ChecksumGenerate(HANDLE hSoundFile)
{
WORD cksum = 0, value;
// sum up the words from 0x00 to 0x63, subtract from 0xFFFF - result is the checksum
// 50 words
unsigned char buffer16[2];
DWORD byteswritten;
for (int idx = 0; idx <= 49 ; idx++) {
	SetFilePointer(hSoundFile, idx * 2, NULL, FILE_BEGIN);

	DWORD readBytes;
	// read 2 bytes
	ReadFile(hSoundFile, buffer16, sizeof(buffer16), &readBytes, NULL);
	if (readBytes != 2) {
		ADPCM_Main->MyShowMessage("Checksum Generate: Read did not read 2 bytes", true);
		return;
		}
	value = ((buffer16[0] << 8) | buffer16[1]);
	cksum += value;
	//ADPCM_Main->MyShowMessage("Checksum index " + String(idx * 2) + " : " + IntToHex( cksum, 4), false);
	}
cksum = 0xFFFF - cksum;
XChecksumEdit->Text = IntToHex(cksum, 4);

//-------------------------------------------------------------
// write the 2 bytes at 0x64-0x65 Checksum
SetFilePointer(hSoundFile, 0x0064, NULL, FILE_BEGIN);
// convert little endian to big endian
buffer16[1] =  cksum & 0x00FF;		// LSB
buffer16[0] =  (cksum & 0xFF00) >> 8;	// MSB

WriteFile(hSoundFile, buffer16, sizeof(buffer16), &byteswritten, NULL);
if (byteswritten != sizeof(buffer16)) {
	ADPCM_Main->MyShowMessage("Checksum Generate: Write did not write 2 bytes 0x64-0x65", true);
	return;
	}

return;	
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::fEngNumChange(TObject *Sender)
{

if (fEngNum->Modified) {
	// convert the decimal value to hex and display it
    XEngineNumberEdit->Text = IntToHex(fEngNum->Text.ToInt() + 1, 4);
    ChangesMade = true;
    }
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::XEngineNumberChange(
      TObject *Sender)
{
//
int iEngNum;
if (XEngineNumberEdit->Modified) {
	// Convert the hex value to decimal and display it
    iEngNum  = _tcstol(XEngineNumberEdit->Text.c_str(), NULL, 16) - 1;
    fEngNum->Text = Format("%02d", ARRAYOFCONST((iEngNum)));
    ChangesMade = true;
	}
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Engine_Parameters::Copy256ButtonClick(
      TObject *Sender)
{
// Copy 256 bytes from another sound file into this sound file
ADPCM_Main->Copylower256bytesClick(Sender);
// reload the Engine Parameters
ReloadButtonClick(Sender);	
}
//---------------------------------------------------------------------------

