//---------------------------------------------------------------------------
/*
ADPCM Program for Windows

 Copyright 2019 by Mark DiVecchio

	This file is part of Remote Train Control.

	Remote Train Control is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Remote Train Control is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with Remote Train Control.  If not, see <http://www.gnu.org/licenses/>.

*/
// Home Page: http://www.silogic.com/trains/ADPCM.html
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include <tchar.h>
#include "ADPCM_Player.h"
#include "Edit_the_Index.h"
#include "assert.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"

extern bool ParseIndex(SFIndex &Index, HANDLE hParseFile, int clip);
extern bool WriteIndex(SFIndex &Index, HANDLE hParseFile, int clip);
//---------------------------------------------------------------------------
TEdit_Index *Edit_Index;
//---------------------------------------------------------------------------
__fastcall TEdit_Index::TEdit_Index(TComponent* Owner)
	: TForm(Owner)
{
assert(false);	// never called
}
//---------------------------------------------------------------------------
__fastcall TEdit_Index::TEdit_Index(TComponent* Owner, String FN, int EAM, int clip)
	: TForm(Owner)
{
assert(FN != "");
ClipNo = clip;
assert(ClipNo > 0 && ClipNo < 256);
EndAddrMax = EAM;

Caption = "Editing Clip Index #" + String(ClipNo);

StartAddrEdit = new THexEdit(this);
assert(StartAddrEdit != NULL);
StartAddrEdit->Parent = this;
StartAddrEdit->Height = XStartAddr->Height;
StartAddrEdit->Width = XStartAddr->Width;
StartAddrEdit->Top = XStartAddr->Top;
StartAddrEdit->Left = XStartAddr->Left;
StartAddrEdit->MinValue = 0x00008000;
StartAddrEdit->MaxValue = EndAddrMax;			//Highest address which can contain sound
StartAddrEdit->Digits = 6;
StartAddrEdit->MaxLength = 6;
StartAddrEdit->Text = "000000";
StartAddrEdit->Cursor = crHandPoint;
StartAddrEdit->ShowHint = true;
StartAddrEdit->Hint = "Starting Address of clip";
StartAddrEdit->Show();

EndRepAddrEdit = new THexEdit(this);
assert(EndRepAddrEdit != NULL);
EndRepAddrEdit->Parent = this;
EndRepAddrEdit->Height = XEndRepAddr->Height;
EndRepAddrEdit->Width = XEndRepAddr->Width;
EndRepAddrEdit->Top = XEndRepAddr->Top;
EndRepAddrEdit->Left = XEndRepAddr->Left;
EndRepAddrEdit->MinValue = 0x00008000;
EndRepAddrEdit->MaxValue = 0x00FFFFFF;
EndRepAddrEdit->Digits = 6;
EndRepAddrEdit->MaxLength = 6;
EndRepAddrEdit->Text = "000000";
EndRepAddrEdit->Cursor = crHandPoint;
EndRepAddrEdit->ShowHint = true;
EndRepAddrEdit->Hint = "Ending Address of Repeating Segment";
EndRepAddrEdit->Show();

StartRepAddrEdit = new THexEdit(this);
assert(StartRepAddrEdit != NULL);
StartRepAddrEdit->Parent = this;
StartRepAddrEdit->Height = XStartRepAddr->Height;
StartRepAddrEdit->Width = XStartRepAddr->Width;
StartRepAddrEdit->Top = XStartRepAddr->Top;
StartRepAddrEdit->Left = XStartRepAddr->Left;
StartRepAddrEdit->MinValue = 0x00008000;
StartRepAddrEdit->MaxValue = 0x00FFFFFF;
StartRepAddrEdit->Digits = 6;
StartRepAddrEdit->MaxLength = 6;
StartRepAddrEdit->Text = "000000";
StartRepAddrEdit->Cursor = crHandPoint;
StartRepAddrEdit->ShowHint = true;
StartRepAddrEdit->Hint = "Starting Address of Repeating Segment";
StartRepAddrEdit->Show();

SMInitEdit = new THexEdit(this);
assert(SMInitEdit != NULL);
SMInitEdit->Parent = this;
SMInitEdit->Height = XSMInit->Height;
SMInitEdit->Width = XSMInit->Width;
SMInitEdit->Top = XSMInit->Top;
SMInitEdit->Left = XSMInit->Left;
SMInitEdit->MinValue = 0x00000000;
SMInitEdit->MaxValue = 0x00FFFFFF;
SMInitEdit->Digits = 6;
SMInitEdit->MaxLength = 6;
SMInitEdit->Text = "000000";
SMInitEdit->Cursor = crHandPoint;
SMInitEdit->ShowHint = true;
SMInitEdit->Hint = "State Machine Initialization (4 bit ADPCM clips only)";
SMInitEdit->Show();

EndAddrEdit = new THexEdit(this);
assert(EndAddrEdit != NULL);
EndAddrEdit->Parent = this;
EndAddrEdit->Height = XEndAddr->Height;
EndAddrEdit->Width = XEndAddr->Width;
EndAddrEdit->Top = XEndAddr->Top;
EndAddrEdit->Left = XEndAddr->Left;
EndAddrEdit->MinValue = 0x00008000;
EndAddrEdit->MaxValue = EndAddrMax; 		//Highest address which can contain sound
EndAddrEdit->Digits = 6;
EndAddrEdit->MaxLength = 6;
EndAddrEdit->Text = "000000";
EndAddrEdit->Cursor = crHandPoint;
EndAddrEdit->ShowHint = true;
EndAddrEdit->Hint = "Ending Address of clip";
EndAddrEdit->Show();

DataByteEdit = new THexEdit(this);
assert(DataByteEdit != NULL);
DataByteEdit->Parent = this;
DataByteEdit->Height = XDataByte->Height;
DataByteEdit->Width = XDataByte->Width/2;
DataByteEdit->Top = XDataByte->Top;
DataByteEdit->Left = XDataByte->Left;
DataByteEdit->MinValue = 0x00000000;
DataByteEdit->MaxValue = 0x000000FF;
DataByteEdit->Digits = 2;
DataByteEdit->MaxLength = 2;
DataByteEdit->Text = "00";
DataByteEdit->Cursor = crHandPoint;
DataByteEdit->ShowHint = true;
DataByteEdit->Hint = "Databyte specifying format, sample rate, and category of clip";
DataByteEdit->Show();


}
//---------------------------------------------------------------------------

__fastcall TEdit_Index::~TEdit_Index()
{
// parent of dyanmically added control will free it
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Index::ShowIt(String FN, int EAM, int clip)
{
TargetFileName = FN;
assert(FN != "");
ClipNo = clip;
assert(ClipNo > 0 && ClipNo < 256);
EndAddrMax = EAM;

Caption = "Editing Clip Index #" + String(ClipNo);

HANDLE hSoundFile;
// try and open the file
if((hSoundFile = CreateFile(TargetFileName.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	ADPCM_Main->MyShowLastError("EditIndex: Unable to open file : " + TargetFileName, true);
	Close();    // cancel the operation
	} else {
	ADPCM_Main->MyShowMessage("EditIndex: Sound file to be edited : " + TargetFileName, false);
	}
if (!ParseIndex(IndexSound, hSoundFile, ClipNo)) {
	ADPCM_Main->MyShowMessage("EditIndex: Parse Index failed", true);
	CloseHandle(hSoundFile);
	Close();    // cancel the operation
	}
CloseHandle(hSoundFile);

StartAddrEdit->MaxValue = EndAddrMax;	//Highest address which can contain sound
StartAddrEdit->Text = IntToHex(IndexSound.StartAddr, 6);

EndRepAddrEdit->Text = IntToHex(IndexSound.EndRepAddr, 6);

StartRepAddrEdit->Text = IntToHex(IndexSound.StartRepAddr, 6);

SMInitEdit->Text = IntToHex(IndexSound.SMInit, 6);

EndAddrEdit->MaxValue = EndAddrMax;		//Highest address which can contain sound
EndAddrEdit->Text = IntToHex(IndexSound.EndAddr, 6);

DataByteEdit->Text = IntToHex(IndexSound.databyte, 2);

// Setup the Low Order (nybble) bit check boxes for 4 bit ADPCM clips
#ifdef NYBBLEADDRESSES
	SA_LOBit->Visible = true;
	SRA_LOBit->Visible = true;
	ERA_LOBit->Visible = true;
	EA_LOBit->Visible = true;
#endif
SA_LOBit->Checked = IndexSound.SA_LOBit;
SRA_LOBit->Checked = IndexSound.SRA_LOBit;
ERA_LOBit->Checked = IndexSound.ERA_LOBit;
EA_LOBit->Checked = IndexSound.EA_LOBit;
//
Show();
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Index::DoItButtonClick(TObject *Sender)
{
// OK
if (!ADPCM_Main->iDebug) {
	// validate entries
	if (StartAddrEdit->Value < 0x00008000) {
		ShowMessage("Start Address must be greater than 0x008000");
		return;
		}
	if (EndAddrEdit->Value <= StartAddrEdit->Value) {
		ShowMessage("End Address must be greater than Start Address");
		return;
		}

	if (StartRepAddrEdit->Value != EMPTY_FIELD) {
		if (StartRepAddrEdit->Value < StartAddrEdit->Value) {
			ShowMessage("Start Repeating Address must be greater than or equal to Start Address");
			return;
			}
		}

	if ((EndRepAddrEdit->Value != EMPTY_FIELD) && (StartRepAddrEdit->Value != EMPTY_FIELD)) {
			if (EndRepAddrEdit->Value <= StartAddrEdit->Value) {
			ShowMessage("End Repeating Address must be greater than Start Address");
			return;
			}
		if (EndRepAddrEdit->Value <= StartRepAddrEdit->Value) {
			ShowMessage("End Repeating Address must be greater than Start Repeating Address");
			return;
			}
		if (EndAddrEdit->Value < EndRepAddrEdit->Value) {
			ShowMessage("End Address must be greater than or equal to End Repeating Address");
			return;
			}
		}
	// if its a 24 bit signed big endian PCM file, check that addresses are a multiple of 3 offset
	// from the starting address
	if (DataByteEdit->Value & PCM24MASK) {
		if (((EndAddrEdit->Value - StartAddrEdit->Value) % 3) != 0) {
			ShowMessage("Warning : End Address value not modulo 3");
			// A warning only
			}
		if ((EndRepAddrEdit->Value != EMPTY_FIELD) && (StartRepAddrEdit->Value != EMPTY_FIELD)) {
			if (((EndRepAddrEdit->Value - StartAddrEdit->Value) % 3) != 0) {
				ShowMessage("Warning : End Repeating Address value not modulo 3");
				// A warning only
				}
			if (((StartRepAddrEdit->Value - StartAddrEdit->Value) % 3) != 0) {
				ShowMessage("Error : Start Repeating Address value not modulo 3 - You must fix this");
				return;
				}

			}

		}
	}

IndexSound.StartAddr = StartAddrEdit->Value;
IndexSound.EndAddr = EndAddrEdit->Value;
IndexSound.StartRepAddr = StartRepAddrEdit->Value;
IndexSound.EndRepAddr = EndRepAddrEdit->Value;
IndexSound.SMInit = SMInitEdit->Value;
IndexSound.databyte = DataByteEdit->Value;
// Low Order (nybble) bits
IndexSound.SA_LOBit = SA_LOBit->Checked;
IndexSound.ERA_LOBit = ERA_LOBit->Checked;
IndexSound.SRA_LOBit = SRA_LOBit->Checked;
IndexSound.EA_LOBit = EA_LOBit->Checked;

HANDLE hSoundFile;
// try and open the file
if((hSoundFile = CreateFile(TargetFileName.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	ADPCM_Main->MyShowLastError("EditIndex: Unable to open file : " + TargetFileName, true);
	Close(); // cancel the operation
	}
// write the sound file index entry
if (!WriteIndex(IndexSound, hSoundFile, ClipNo)) {
	ADPCM_Main->MyShowMessage("EditIndex: Could not write index to sound file", true);
	CloseHandle(hSoundFile);
	Close();    // cancel the operation
	}
CloseHandle(hSoundFile);
ADPCM_Main->FileSetDateNow(TargetFileName);
//
if (TargetFileName == ADPCM_Main->MRUComboBox->Text) {
	ADPCM_Main->AnalyzeButtonClick(Sender);
    }
Close();    // cancel the operation
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Index::CancelButtonClick(TObject *Sender)
{
Close();    // cancel the operation
}
//---------------------------------------------------------------------------


 