//---------------------------------------------------------------------------
/*
ADPCM Program for Windows

 Copyright 2020 by Mark DiVecchio

    This file is part of Remote Train Control.

    Remote Train Control is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Remote Train Control is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
    along with Remote Train Control.  If not, see <http://www.gnu.org/licenses/>.

*/
// Home Page: http://www.silogic.com/trains/ADPCM.html
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include <tchar.h>
#include "ADPCM_Player.h"
#include "Edit_the_Scale_Factor.h"
#include "assert.h"
#include <stdio.h>
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TEdit_Scale_Factor *Edit_Scale_Factor;
//---------------------------------------------------------------------------
__fastcall TEdit_Scale_Factor::TEdit_Scale_Factor(TComponent* Owner)
	: TForm(Owner)
{
assert(false);	// never called
}
//---------------------------------------------------------------------------

THexEdit * __fastcall TEdit_Scale_Factor::CreateHexEdit(TMaskEdit * TemplateEdit, int minval, int maxval)
{
THexEdit * NewHexEdit = new THexEdit(this);
assert(NewHexEdit != NULL);
NewHexEdit->Parent = this;
NewHexEdit->Height = TemplateEdit->Height;
NewHexEdit->Width = TemplateEdit->Width;
NewHexEdit->Top = TemplateEdit->Top;
NewHexEdit->Left = TemplateEdit->Left;
NewHexEdit->MinValue = minval;
NewHexEdit->MaxValue = maxval;
NewHexEdit->Digits = 4;
NewHexEdit->MaxLength = 4;
NewHexEdit->Text = "00.00";
NewHexEdit->Cursor = crHandPoint;
NewHexEdit->ShowHint = true;
NewHexEdit->Hint = TemplateEdit->Hint;
NewHexEdit->TabOrder = 0;
NewHexEdit->OnChange = XChange;
NewHexEdit->Font = TemplateEdit->Font;
NewHexEdit->Show();

return NewHexEdit;
}
//---------------------------------------------------------------------------

__fastcall TEdit_Scale_Factor::TEdit_Scale_Factor(TComponent* Owner, String FN, int p1, int p2, int PSType)
	: TForm(Owner)
{
TargetFileName = FN;
assert(FN != "");
assert(PSType == PS2 || PSType == PS3);
MyPSType = PSType;

XSFEdit = CreateHexEdit(XSF, 0x00000000, 0x0000FFFF);
XSpdEdit = CreateHexEdit(XSpd, 0x00000000, 0x0000FFFF);
XChuffEdit = CreateHexEdit(XChuff, 0x00000000, 0x0000FFFF);

WheelDiameter->Text = "00.00";
GearRatio->Text = "00.00";
NStripes->Text = 24;
ChangesMade = false;

XSFEdit->Text = "0";
XSpdEdit->Text = "0";
XChuffEdit->Text = "0";
iSF->Text = "0";

LoadFactors(TargetFileName);
}
//---------------------------------------------------------------------------

__fastcall TEdit_Scale_Factor::~TEdit_Scale_Factor()
{
// parent of dyanmically added control will free it
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Scale_Factor::ShowIt(String FN, int p1, int p2, int PSType)
{
TargetFileName = FN;
assert(FN != "");
assert(PSType == PS2 || PSType == PS3);
MyPSType = PSType;

Caption = "Editing Scale Factor";

//
Show();
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Scale_Factor::DoItButtonClick(TObject *Sender)
{
// OK
if (!ChangesMade) {
    ADPCM_Main->MyShowMessage("DoIt: No factors were changed, nothing new to write", true);
	return;
	}
HANDLE hSoundFile;
// try and open the file
if((hSoundFile = CreateFile(TargetFileName.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	ADPCM_Main->MyShowLastError("DoIt: Unable to open file : " + TargetFileName, true);
	Close(); // cancel the operation
	}
// write the Scale Factor
SetFilePointer(hSoundFile, 0x0008, NULL, FILE_BEGIN);
unsigned char buffer16[2];
DWORD byteswritten;
// convert little endian to big endian
buffer16[1] =  _tcstol(XSFEdit->Text.SubString(3,2).c_str(), NULL, 16);	// LSB
buffer16[0] =  _tcstol(XSFEdit->Text.SubString(1,2).c_str(), NULL, 16);	// MSB

WriteFile(hSoundFile, buffer16, sizeof(buffer16), &byteswritten, NULL);
if (byteswritten != sizeof(buffer16)) {
	ADPCM_Main->MyShowMessage("DoIt: Write did not write 2 bytes of scale factor", true);
	return;
	}

// write the Speed Factor
SetFilePointer(hSoundFile, 0x0004, NULL, FILE_BEGIN);
// convert little endian to big endian
buffer16[1] =  _tcstol(XSpdEdit->Text.SubString(3,2).c_str(), NULL, 16);	// LSB
buffer16[0] =  _tcstol(XSpdEdit->Text.SubString(1,2).c_str(), NULL, 16);	// MSB

WriteFile(hSoundFile, buffer16, sizeof(buffer16), &byteswritten, NULL);
if (byteswritten != sizeof(buffer16)) {
	ADPCM_Main->MyShowMessage("DoIt: Write did not write 2 bytes of speed factor", true);
	return;
	}


// write the 2 bytes at 0x18-0x19 Chuff Factor
SetFilePointer(hSoundFile, 0x0018, NULL, FILE_BEGIN);
// convert little endian to big endian
buffer16[1] =  _tcstol(XChuffEdit->Text.SubString(3,2).c_str(), NULL, 16);	// LSB
buffer16[0] =  _tcstol(XChuffEdit->Text.SubString(1,2).c_str(), NULL, 16);	// MSB

WriteFile(hSoundFile, buffer16, sizeof(buffer16), &byteswritten, NULL);
if (byteswritten != sizeof(buffer16)) {
	ADPCM_Main->MyShowMessage("DoIt: Write did not write 2 bytes of chuff factor", true);
	return;
	}

CloseHandle(hSoundFile);
ADPCM_Main->FileSetDateNow(TargetFileName);
ChangesMade = false;
Close();	// done
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Scale_Factor::CancelButtonClick(TObject *Sender)
{
Close();    // cancel the operation
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Scale_Factor::CalcScaleFactorButtonClick(TObject *Sender)
{
// Create the hex and decimal versions of the SF from
// Wheel diameter, gear ratio, and number of stripes

double WD = _tcstod(WheelDiameter->Text.c_str(), NULL);
double GR = _tcstod(GearRatio->Text.c_str(), NULL);
int NS = _tcstol(NStripes->Text.c_str(), NULL, 10);
int K;
double TS;
int ScaleF;

if (WD != 0 && GR != 0 && NS != 0) {
	// This equation calculates the scale factor, SF - parameters:
	// 5280 feet per mile
	// 12 inches per foot
	// 48 O Scale
	// 25.4 mm per inch
	// M_PI is 3.14159265358979323846264338327950288
	// NS Number of stripes on the tach strip, usually 24
	// GR Gear Ratio, eg  10.50 or 18.00
	// WD Wheel Diameter in mm, eg 22.60
	//
	K = (12 * 5280/48);	// 1320 inches/Smile
	// multiplier of 2 seems to be because both the
	// leading edge and the trailing edge of each stripe generates a count
	TS = 2 * ((K)/((WD / 25.4) * M_PI)) * NS * GR;
	// The meaning of the divisor of 11 is not known. Test runs over
	// measured distance show that it is exactly needed to generate
	// correct distance reporting by the engine.
	// Calculate Scale Factor
	ScaleF = TS / 11;
	//
	//TCHAR str[25];
	//_sntprintf(str, sizeof str, TEXT("%6d"), ScaleF);
	//iSF->Text = str;
	XSFEdit->Text = IntToHex(ScaleF, 4);
	// Convert the hex value to decimal and display it
	float fCount = _tcstol(XSFEdit->Text.c_str(), NULL, 16) / 10.0;
	iSF->Text = Format("%8.1f", ARRAYOFCONST((fCount)));
	}
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Scale_Factor::CalcSpeedFactorButtonClick(TObject *Sender)
{
// Create the hex and decimal versions of the SpdF from
// Wheel diameter, gear ratio, and number of stripes

double WD = _tcstod(WheelDiameter->Text.c_str(), NULL);
double GR = _tcstod(GearRatio->Text.c_str(), NULL);
int NS = _tcstol(NStripes->Text.c_str(), NULL, 10);
int K;
double TS;
int ScaleF, SpdF;

if (WD != 0 && GR != 0 && NS != 0) {
	// This equation calculates the scale factor, SF - parameters:
	// 5280 feet per mile
	// 12 inches per foot
	// 48 O Scale
	// 25.4 mm per inch
	// M_PI is 3.14159265358979323846264338327950288
	// NS Number of stripes on the tach strip, usually 24
	// GR Gear Ratio, eg  10.50 or 18.00
	// WD Wheel Diameter in mm, eg 22.60
	//
	K = (12 * 5280/48);	// 1320 inches/Smile
	// multiplier of 2 seems to be because both the
	// leading edge and the trailing edge of each stripe generates a count
	TS = 2 * ((K)/((WD / 25.4) * M_PI)) * NS * GR;
	// the meaning of the divisor of 11 is not known. Test runs over
	// measured distance show that it is exactly needed to generate
	// correct distance reporting by the engine.
	// Calculate Scale Factor
	ScaleF = TS / 11;
	//
	TCHAR str[25];
	// Calculate Speed Factor
	// Constant was calculated by Gear Ratio Spreadsheet
    // 31.9288 ms/tick_of_DCH
	SpdF = (4973087 * 31.9288) / (ScaleF * 11);
	XSpdEdit->Text = IntToHex(SpdF, 4);
	ChangesMade = true;
    }
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Scale_Factor::DecodeButtonClick(TObject *Sender)
{
// Convert the hex scale factor to decimal and
// if 2 out of 3 parameters are filled in, calculate the third paramter
//iSF->Text = Format("%04d", ARRAYOFCONST(((int)_tcstol(XSFEdit->Text.c_str(), NULL, 16))));

int ScaleFactor = _tcstol(iSF->Text.c_str(), NULL, 10);

double WD = _tcstod(WheelDiameter->Text.c_str(), NULL);
double GR = _tcstod(GearRatio->Text.c_str(), NULL);
int NS = _tcstol(NStripes->Text.c_str(), NULL, 10);
int K;
int SF = _tcstol(XSFEdit->Text.c_str(), NULL, 16);

if (ScaleFactor != 0) {
	// Given the scale factor, SF, and 2 of the 3 other input parameters,
	// these equations calculate the 3rd parameter
    if (WD == 0 && GR != 0 && NS != 0) {
    	// calculate Wheel Diameter
		K = (12 * 5280/48);
		WD = (25.4 * (2 * K * NS * GR))/ (SF * 11 * M_PI);
		TCHAR str[25];
		_sntprintf(str, sizeof str, TEXT("%5.2f"), WD);
		WheelDiameter->Text = str;
    	}
    if (WD != 0 && GR == 0 && NS != 0) {
    	// calculate Gear Ratio
		K = (12 * 5280/48);
		double DN = 2 * ((K)/((WD / 25.4) * M_PI)) * NS;
		GR = (SF * 11) / DN;
		TCHAR str[25];
		_sntprintf(str, sizeof str, TEXT("%5.2f"), GR);
		GearRatio->Text = str;
    	}
    if (WD != 0 && GR != 0 && NS == 0) {
    	// calculate Number of Stripes
		K = (12 * 5280/48);
		double DN = 2 * ((K)/((WD / 25.4) * M_PI)) * GR;
		NS = (SF * 11) / DN;
		TCHAR str[25];
		_sntprintf(str, sizeof str, TEXT("%2f"), NS);
		NStripes->Text = str;
		}
	}
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Scale_Factor::EncodeFromDecimalButtonClick(
      TObject *Sender)
{
// Convert the decimal scale factor into hex
int iFactor = _tcstod(iSF->Text.c_str(), NULL) * 10;
XSFEdit->Text= IntToHex(iFactor, 4);
ChangesMade = true;
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Scale_Factor::XChange(TObject *Sender)
{
ChangesMade = true;
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Scale_Factor::CalcChuffFactorButtonClick(TObject *Sender)
{
// Calculate the Chuff Factor from the gear ratio and # of stripes
//
int NS = _tcstol(NStripes->Text.c_str(), NULL, 10);
double GR = _tcstod(GearRatio->Text.c_str(), NULL);
int CF;	// Number of stripes for one rev of driving wheel * 2 = Number of DTO counts for one rev of driving wheel
// Number of stripes for one rev of driving wheel = gear_ratio * Number of stripes on tach wheel
// divided by 4 to set 4 chuffs per revolution
if (NS != 0 && GR != 0) {
	CF = NS * GR * 2 / 4;
	XChuffEdit->Text = IntToHex(CF, 4);;
	}
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Scale_Factor::Button1Click(TObject *Sender)
{
// clear the calculated values and restore original Sound File values

WheelDiameter->Text = "00.00";
GearRatio->Text = "00.00";
NStripes->Text = 24;
ChangesMade = false;

XSFEdit->Text = "0";
XSpdEdit->Text = "0";
XChuffEdit->Text = "0";
iSF->Text = "0";

LoadFactors(TargetFileName);	
}
//---------------------------------------------------------------------------

void __fastcall TEdit_Scale_Factor::LoadFactors(String TargetFileName)
{

HANDLE hSoundFile;
// try and open the file
if((hSoundFile = CreateFile(TargetFileName.c_str(), GENERIC_READ, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	ADPCM_Main->MyShowLastError("Show It: Unable to open file : " + TargetFileName, true);
	Close();    // cancel the operation
	} else {
	ADPCM_Main->MyShowMessage("Edit Scale Factors: Sound file to be edited : " + TargetFileName, false);
	}
// read the Scale Factor
SetFilePointer(hSoundFile, 0x0008, NULL, FILE_BEGIN);
unsigned char buffer16[2];
DWORD readBytes;
// read 2 bytes
ReadFile(hSoundFile, buffer16, sizeof(buffer16), &readBytes, NULL);
if (readBytes != 2) {
	ADPCM_Main->MyShowMessage("Show It: Read did not read 2 bytes of scale factor", true);
	return;
	}
// convert big endian to little endian
XSFEdit->Text = IntToHex((buffer16[0] << 8) | buffer16[1], 4);
// Convert the hex value to decimal and display it
float fCount = _tcstol(XSFEdit->Text.c_str(), NULL, 16) / 10.0;
iSF->Text = Format("%8.1f", ARRAYOFCONST((fCount)));

// read the Speed Factor
SetFilePointer(hSoundFile, 0x0004, NULL, FILE_BEGIN);
// read 2 bytes
ReadFile(hSoundFile, buffer16, sizeof(buffer16), &readBytes, NULL);
if (readBytes != 2) {
	ADPCM_Main->MyShowMessage("Show It: Read did not read 2 bytes of speed factor", true);
	return;
	}
// convert big endian to little endian
XSpdEdit->Text = IntToHex((buffer16[0] << 8) | buffer16[1], 4);

// read the 2 bytes at 0x18-0x19 Chuff Factor
SetFilePointer(hSoundFile, 0x0018, NULL, FILE_BEGIN);
// read 2 bytes
ReadFile(hSoundFile, buffer16, sizeof(buffer16), &readBytes, NULL);
if (readBytes != 2) {
	ADPCM_Main->MyShowMessage("Show It: Read did not read 2 bytes chuff factor", true);
	return;
	}
// convert big endian to little endian
XChuffEdit->Text = IntToHex((buffer16[0] << 8) | buffer16[1], 4);

CloseHandle(hSoundFile);
}
//---------------------------------------------------------------------------

