//---------------------------------------------------------------------------
/*
ADPCM Program for Windows

 Copyright 2019 by Mark DiVecchio

    This file is part of Remote Train Control.

    Remote Train Control is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Remote Train Control is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
    along with Remote Train Control.  If not, see <http://www.gnu.org/licenses/>.

*/
// Home Page: http://www.silogic.com/trains/ADPCM.html
//---------------------------------------------------------------------------//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include <tchar.h>
#include "ADPCM_Player.h"
#include "Edit_the_Softkeys.h"
#include "assert.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TSoftKey_Editor *SoftKey_Editor;
//---------------------------------------------------------------------------

__fastcall TSoftKey_Editor::TSoftKey_Editor(TComponent* Owner)
	: TForm(Owner)
{
// User Registry Data
RegFileRootKey = "Software\\Remote Control\\ADPCM";
}
//---------------------------------------------------------------------------

void __fastcall TSoftKey_Editor::ShowIt(String FN)
{
// insert FN into MRU List
// don't insert a duplicate string
int idx = MRUComboBox->Items->IndexOf(FN);
if (idx == -1) {
	MRUComboBox->Items->Insert(0, FN);
	MRUComboBox->ItemIndex = 0;
	if (MRUComboBox->Items->Count > MRUComboBox->DropDownCount) MRUComboBox->Items->Delete(MRUComboBox->DropDownCount);
	} else {
	// set ItemIndex
	MRUComboBox->ItemIndex = idx;
	}

MRUComboBox->Text = FN;
MRUComboBox->Hint = MRUComboBox->Text;
Caption = "Editing Softkeys";
SetCheckBoxes();
Show();
}
//---------------------------------------------------------------------------

void __fastcall TSoftKey_Editor::CancelButtonClick(TObject *Sender)
{
// cancel the operation
Close();    // this window
}
//---------------------------------------------------------------------------

void __fastcall TSoftKey_Editor::DoItButtonClick(TObject *Sender)
{
// Do the function
HANDLE hSoundFile;
// Open the Sound File
if (MRUComboBox->Text != "c:\\" && MRUComboBox->Text != "") {
	// try and open the file
	if((hSoundFile = CreateFile(MRUComboBox->Text.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
		ADPCM_Main->MyShowLastError("Edit Softkeys: Unable to open sound file : " + MRUComboBox->Text, true);
		// cancel the operation
		return;
		} else {
		ADPCM_Main->MyShowMessage("Edit Softkeys: Sound file to be edited : " + MRUComboBox->Text, false);
		}

	setBit(hSoundFile, CDX->Checked, 0x191F , 0x03);  // returns true (0x01) or false (0x00)

	setBit(hSoundFile, FCC->Checked, 0x1912 , 0x40);
	setBit(hSoundFile, AuxLight3onDitch->Checked, 0x1912 , 0x20);
	setBit(hSoundFile, AuxLight2onMars->Checked, 0x1912 , 0x10);
	setBit(hSoundFile, AuxLight1onInterior->Checked, 0x1912 , 0x08);
	setBit(hSoundFile, MarkerLightonMars->Checked, 0x1912 , 0x01);

	setBit(hSoundFile, FireboxGlowonDitch->Checked, 0x1911 , 0x80);
	setBit(hSoundFile, FireboxGlowonMars->Checked, 0x1911 , 0x40);
	setBit(hSoundFile, NumberBoardsonMarsandDitch->Checked, 0x1911 , 0x20);
	setBit(hSoundFile, InteriorLightonDitch->Checked, 0x1911 , 0x10);
	setBit(hSoundFile, InteriorLightonMars->Checked, 0x1911 , 0x08);
	setBit(hSoundFile, RunningLightsonMarsandDitch->Checked, 0x1911 , 0x04);
	setBit(hSoundFile, MarkerLightonDitch->Checked, 0x1911 , 0x02);
	setBit(hSoundFile, FireboxGlowonInterior->Checked, 0x1911 , 0x01);

	setBit(hSoundFile, NumberBoardsonDitch->Checked, 0x1910 , 0x80);
	setBit(hSoundFile, TrackInspectionLightonDitch->Checked, 0x1910 , 0x40);
	setBit(hSoundFile, NumberBoardsonMars->Checked, 0x1910 , 0x20);
	setBit(hSoundFile, HeadlightonInterior->Checked, 0x1910 , 0x10);
	setBit(hSoundFile, MarkerLightonInterior->Checked, 0x1910 , 0x08);
	setBit(hSoundFile, MarkerLight->Checked, 0x1910 , 0x04);
	setBit(hSoundFile, BeaconLight->Checked, 0x1910 , 0x01);

	setBit(hSoundFile, MarsLightonMars->Checked, 0x190F , 0x80);
	setBit(hSoundFile, DitchLightMenu->Checked, 0x190F , 0x40);
	setBit(hSoundFile, InteriorLightonInterior->Checked, 0x190F , 0x20);

	setBit(hSoundFile, SwingingBell->Checked, 0x190C , 0x20);
	setBit(hSoundFile, SmokeWhistle->Checked, 0x190C , 0x10);
	setBit(hSoundFile, PlayWhistle->Checked, 0x190C , 0x08);
	setBit(hSoundFile, SXS->Checked, 0x190C , 0x02);
	setBit(hSoundFile, SAH->Checked, 0x190C , 0x01);

	setBit(hSoundFile, SRS->Checked, 0x190B , 0x80);
	setBit(hSoundFile, SFS->Checked, 0x190B , 0x40);
	setBit(hSoundFile, SSD->Checked, 0x190B , 0x04);
	setBit(hSoundFile, SSU->Checked, 0x190B , 0x02);

	setBit(hSoundFile, CabChat->Checked, 0x190A , 0x20);
	setBit(hSoundFile, I01->Checked, 0x1909, 0x08);
	setBit(hSoundFile, I02->Checked, 0x1909, 0x10);
	setBit(hSoundFile, I03->Checked, 0x1909, 0x20);
	setBit(hSoundFile, I04->Checked, 0x1909, 0x40);
	setBit(hSoundFile, I05->Checked, 0x1909, 0x80);
	setBit(hSoundFile, I06->Checked, 0x190A, 0x01);
	setBit(hSoundFile, I07->Checked, 0x190A, 0x02);
	setBit(hSoundFile, I08->Checked, 0x190A, 0x04);

	setBit(hSoundFile, S01->Checked, 0x1906, 0x04);
	setBit(hSoundFile, S02->Checked, 0x1906, 0x08);
	setBit(hSoundFile, S03->Checked, 0x1906, 0x10);
	setBit(hSoundFile, S04->Checked, 0x1906, 0x20);
	setBit(hSoundFile, S05->Checked, 0x1906, 0x40);
	setBit(hSoundFile, S06->Checked, 0x1906, 0x80);
	setBit(hSoundFile, S07->Checked, 0x1907, 0x01);
	setBit(hSoundFile, S08->Checked, 0x1907, 0x02);
	setBit(hSoundFile, S09->Checked, 0x1907, 0x04);
	setBit(hSoundFile, S10->Checked, 0x1907, 0x08);

	//
	CloseHandle(hSoundFile);
	ADPCM_Main->FileSetDateNow(MRUComboBox->Text);
	//
	Close(); // this window
	}
}
//---------------------------------------------------------------------------

void __fastcall TSoftKey_Editor::BrowseforFileButtonClick(TObject *Sender)
{
OpenDialog->DefaultExt = "mth";
OpenDialog->Filter = "mth files (*.mth)|*.MTH";
OpenDialog->Options.Clear();
OpenDialog->Options << ofFileMustExist;

if (OpenDialog->Execute()) {
	// file picked
	MRUComboBox->Text = OpenDialog->FileName;
	MRUComboBox->Hint = MRUComboBox->Text;
	OpenDialog->InitialDir = ExtractFilePath(OpenDialog->FileName);
	SetCheckBoxes();

    // insert into MRU List
	// don't insert a duplicate string
    int idx = MRUComboBox->Items->IndexOf(MRUComboBox->Text);
    if (idx == -1) {
        MRUComboBox->Items->Insert(0, MRUComboBox->Text);
        MRUComboBox->ItemIndex = 0;
		if (MRUComboBox->Items->Count > MRUComboBox->DropDownCount) MRUComboBox->Items->Delete(MRUComboBox->DropDownCount);
        } else {
        // set ItemIndex
        MRUComboBox->ItemIndex = idx;
        }
	}
}
//---------------------------------------------------------------------------

void __fastcall TSoftKey_Editor::WMDropFiles(TWMDropFiles &message)
{
// Get the number of files dropped
UINT filecount = DragQueryFile((HDROP) message.Drop, 0xFFFFFFFF, NULL, 0);
if (filecount > 0) {
	String FullFileName;
	for (UINT ii = 0 ; ii < filecount ; ++ii){
    	Application->ProcessMessages();
		// Get the file name
		TCHAR FileName[MAX_PATH];
		DragQueryFile ((HDROP) message.Drop, ii, FileName, MAX_PATH);
		FullFileName = FileName;
		//Application->MessageBox(FullFileName.c_str(), TEXT("Result of DragQueryFile()"), MB_OKCANCEL | MB_DEFBUTTON1);
		if (FileExists(FullFileName.c_str())) {
			// do something for each file
			ADPCM_Main->MyShowMessage("Dropped Edit Softkeys Source file " + FullFileName, false);
			} else {
			continue;
			}
		}
	String FileName = LowerCase(ExtractFileName(FullFileName));

	String Extension = LowerCase(ExtractFileExt(FullFileName));
	if (Extension != ".mth") {
		ADPCM_Main->MyShowMessage("Edit Soft Keys: Can only edit '.mth' files", true);
		MRUComboBox->Text = "c:\\";
		MRUComboBox->Hint = "c:\\";
		} else {
		// use last dropped path/filename
		MRUComboBox->Text = FullFileName;
		MRUComboBox->Hint = MRUComboBox->Text;
		}

    // Add new filename to DropDown List
    // insert into MRU List
	// don't insert a duplicate string
    int idx = MRUComboBox->Items->IndexOf(MRUComboBox->Text);
    if (idx == -1) {
        MRUComboBox->Items->Insert(0, MRUComboBox->Text);
        MRUComboBox->ItemIndex = 0;
		if (MRUComboBox->Items->Count > MRUComboBox->DropDownCount)
        MRUComboBox->Items->Delete(MRUComboBox->DropDownCount);
		} else {
        // set ItemIndex
        MRUComboBox->ItemIndex = idx;
        }

	}

Update();	// redraw Form
DragFinish ((HDROP) message.Drop);
SetCheckBoxes();
return;
}
//---------------------------------------------------------------------------

void __fastcall TSoftKey_Editor::SetCheckBoxes(void) {
HANDLE hSoundFile;
// Open the Sound File
if (MRUComboBox->Text != "c:\\" && MRUComboBox->Text != "") {
	// try and open the file
	if((hSoundFile = CreateFile(MRUComboBox->Text.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
		ADPCM_Main->MyShowLastError("Edit Softkeys: Unable to open sound file : " + MRUComboBox->Text, true);
		// cancel the operation
		return;
		} else {
		ADPCM_Main->MyShowMessage("Edit Softkeys: Sound file to be edited : " + MRUComboBox->Text, false);
		}
	// check each soft key bit and check the boxes as needed

	CDX->Checked = getBit(hSoundFile, 0x191F , 0x03);  // returns true (0x01) or false (0x00)

	FCC->Checked = getBit(hSoundFile, 0x1912 , 0x40);
	AuxLight3onDitch->Checked = getBit(hSoundFile, 0x1912 , 0x20);
	AuxLight2onMars->Checked = getBit(hSoundFile, 0x1912 , 0x10);
	AuxLight1onInterior->Checked = getBit(hSoundFile, 0x1912 , 0x08);
	MarkerLightonMars->Checked = getBit(hSoundFile, 0x1912 , 0x01);

	FireboxGlowonDitch->Checked = getBit(hSoundFile, 0x1911 , 0x80);
	FireboxGlowonMars->Checked = getBit(hSoundFile, 0x1911 , 0x40);
	NumberBoardsonMarsandDitch->Checked = getBit(hSoundFile, 0x1911 , 0x20);
	InteriorLightonDitch->Checked = getBit(hSoundFile, 0x1911 , 0x10);
	InteriorLightonMars->Checked = getBit(hSoundFile, 0x1911 , 0x08);
	RunningLightsonMarsandDitch->Checked = getBit(hSoundFile, 0x1911 , 0x04);
	MarkerLightonDitch->Checked = getBit(hSoundFile, 0x1911 , 0x02);
	FireboxGlowonInterior->Checked = getBit(hSoundFile, 0x1911 , 0x01);

	NumberBoardsonDitch->Checked = getBit(hSoundFile, 0x1910 , 0x80);
	TrackInspectionLightonDitch->Checked = getBit(hSoundFile, 0x1910 , 0x40);
	NumberBoardsonMars->Checked = getBit(hSoundFile, 0x1910 , 0x20);
	HeadlightonInterior->Checked = getBit(hSoundFile, 0x1910 , 0x10);
	MarkerLightonInterior->Checked = getBit(hSoundFile, 0x1910 , 0x08);
	MarkerLight->Checked = getBit(hSoundFile, 0x1910 , 0x04);
	BeaconLight->Checked = getBit(hSoundFile, 0x1910 , 0x01);

	MarsLightonMars->Checked = getBit(hSoundFile, 0x190F , 0x80);
	DitchLightMenu->Checked = getBit(hSoundFile, 0x190F , 0x40);
	InteriorLightonInterior->Checked = getBit(hSoundFile, 0x190F , 0x20);

	SwingingBell->Checked = getBit(hSoundFile, 0x190C , 0x20);
	SmokeWhistle->Checked = getBit(hSoundFile, 0x190C , 0x10);
	PlayWhistle->Checked = getBit(hSoundFile, 0x190C , 0x08);
	SXS->Checked = getBit(hSoundFile, 0x190C , 0x02);
	SAH->Checked = getBit(hSoundFile, 0x190C , 0x01);

	SRS->Checked = getBit(hSoundFile, 0x190B , 0x80);
	SFS->Checked = getBit(hSoundFile, 0x190B , 0x40);
	SSD->Checked = getBit(hSoundFile, 0x190B , 0x04);
	SSU->Checked = getBit(hSoundFile, 0x190B , 0x02);

	CabChat->Checked = getBit(hSoundFile, 0x190A , 0x20);
	I01->Checked = getBit(hSoundFile, 0x1909, 0x08);
	I02->Checked = getBit(hSoundFile, 0x1909, 0x10);
	I03->Checked = getBit(hSoundFile, 0x1909, 0x20);
	I04->Checked = getBit(hSoundFile, 0x1909, 0x40);
	I05->Checked = getBit(hSoundFile, 0x1909, 0x80);
	I06->Checked = getBit(hSoundFile, 0x190A, 0x01);
	I07->Checked = getBit(hSoundFile, 0x190A, 0x02);
	I08->Checked = getBit(hSoundFile, 0x190A, 0x04);

	S01->Checked = getBit(hSoundFile, 0x1906, 0x04);
	S02->Checked = getBit(hSoundFile, 0x1906, 0x08);
	S03->Checked = getBit(hSoundFile, 0x1906, 0x10);
	S04->Checked = getBit(hSoundFile, 0x1906, 0x20);
	S05->Checked = getBit(hSoundFile, 0x1906, 0x40);
	S06->Checked = getBit(hSoundFile, 0x1906, 0x80);
	S07->Checked = getBit(hSoundFile, 0x1907, 0x01);
	S08->Checked = getBit(hSoundFile, 0x1907, 0x02);
	S09->Checked = getBit(hSoundFile, 0x1907, 0x04);
	S10->Checked = getBit(hSoundFile, 0x1907, 0x08);

	CloseHandle(hSoundFile);
	} else {
	ADPCM_Main->MyShowMessage("Edit Softkeys: Unable to open sound file : " + MRUComboBox->Text, true);
	// cancel the operation
	return;
	}
}
//---------------------------------------------------------------------------

bool __fastcall TSoftKey_Editor::getBit(HANDLE hFile, int Address, unsigned char MyMask) {
// check the byte at Address using Mask, return true if all masked bits are 1 otherwise return false
unsigned char membuf[2];
unsigned long readBytes = 0;
// read the byte
SetFilePointer(hFile, Address, NULL, FILE_BEGIN);
ReadFile(hFile, membuf, 1L, &readBytes, NULL);
if (readBytes != 1L) {
	ADPCM_Main->MyShowMessage("getBit: Could not read byte from sound file", true);
	return false;
	}

if ((membuf[0] & MyMask) == MyMask) return true;
return false;
}
//---------------------------------------------------------------------------

bool __fastcall TSoftKey_Editor::setBit(HANDLE hFile, bool Checked, int Address,  unsigned char MyMask) {
// set the bits in the byte at Address using Mask (do it for all bits in Mask)
unsigned char membuf[2];
unsigned long readBytes = 0, writtenBytes = 0;
// read the byte
SetFilePointer(hFile, Address, NULL, FILE_BEGIN);
ReadFile(hFile, membuf, 1L, &readBytes, NULL);
if (readBytes != 1L) {
	ADPCM_Main->MyShowMessage("setBit: Could not read byte from sound file", true);
	return false;
	}

if (Checked) {
	// set bits
	membuf[0] |= MyMask;
	} else {
	// reset bits
	membuf[0] &= ~MyMask;
	}

// write the byte
SetFilePointer(hFile, Address, NULL, FILE_BEGIN);
WriteFile(hFile, membuf, 1L, &writtenBytes, NULL);
if (writtenBytes != 1L) {
	ADPCM_Main->MyShowMessage("setBit: Could not write byte to sound file", true);
	return false;
	}

return true;
}
void __fastcall TSoftKey_Editor::FormCreate(TObject *Sender)
{
DragAcceptFiles (Handle, True);		//  Accept drag and drop from the Explorer

// read MRU from registry if they exist
TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
assert(RegFile != NULL);

//http://bcbjournal.org/articles/vol1/9711/Implementing_a_recent-files_list.htm
for (int idx = 0 ; idx < MRUComboBox->DropDownCount ; idx++) {
	String S = RegFile->ReadString("MRU Soft Keys", "MRU" + String(idx + 1), "");
    if (S != "") {
        MRUComboBox->Items->Append(S);
        } else {
        break;
        }
    }
// next line changes MRUComboBox->Text
MRUComboBox->ItemIndex = RegFile->ReadInteger("MRU Soft Keys", "MRUItemIndex", -1);
MRUComboBox->Hint = MRUComboBox->Text;;
// Deallocate RegFile
delete RegFile;
//
if (MRUComboBox->Text == "") {
    // use WorkingDir
	OpenDialog->InitialDir = ADPCM_Main->cwd;
	MRUComboBox->Text = OpenDialog->InitialDir + "\\*.mth";
   	MRUComboBox->Hint = MRUComboBox->Text;
    } else {
    // use directory from Script FileName
	OpenDialog->InitialDir = ExtractFilePath(MRUComboBox->Text);
   	MRUComboBox->Hint = MRUComboBox->Text;
	}
}
//---------------------------------------------------------------------------

void __fastcall TSoftKey_Editor::FormDestroy(TObject *Sender)
{
//
//
// write MRU to registry
TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
assert(RegFile != NULL);

int Cnt = MRUComboBox->Items->Count;
if (Cnt > 0) {
    for (int idx = 0 ; idx < Cnt ; idx++) {
		RegFile->WriteString("MRU Soft Keys", "MRU" + String(idx + 1), MRUComboBox->Items->Strings[idx]);
        }
    }
RegFile->WriteInteger("MRU Soft Keys", "MRUItemIndex", MRUComboBox->ItemIndex);
// Deallocate RegFile
delete RegFile;

}
//---------------------------------------------------------------------------

void __fastcall TSoftKey_Editor::CopyrightClick(TObject *Sender)
{
CabChat->Visible = true;
}
//---------------------------------------------------------------------------

void __fastcall TSoftKey_Editor::CleartheList1Click(TObject *Sender)
{
MRUComboBox->Text = "";
MRUComboBox->Items->Clear();
}
//---------------------------------------------------------------------------

void __fastcall TSoftKey_Editor::RemoveFromListClick(TObject *Sender)
{
// remove the filename from the MRUComboBox
MRUComboBox->Items->Delete(MRUComboBox->ItemIndex);
MRUComboBox->ItemIndex = -1;	// nothing selected
}
//---------------------------------------------------------------------------

