/*

BCB Hex Components: the free components for Borland C++ Builder 

Copyright 2003-20034 Jacek Szumigaj <szumi@users.sourceforge.net>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/****************************************************************************
 *File: HexEdit.cpp
 *Autor: Jacek Szumigaj
 *Version: 0.1 (listopad 2003)
 *Description: Komponent typu Edit z wbudowanymi mechanizmem
 *             kowersji zawartoci na posta hexadecymaln i odwrotnie.
 *             Do standarowych waciwoci TEdit dodano property "Value"
 *             ktre zawiera warto wpisan hexadecymalnie w okienko komponentu.
 *             Podstawienie wartoci do Value powoduje jej wywietlenie
 *             w komponencie w postaci hexadecymalnej
 *             Edycja zawartoci moe odbywa si w trybie "Overwrite"
 ****************************************************************************/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include <tchar.h>
#include "HexEdit.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------
// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
//

static inline void ValidCtrCheck(THexEdit *)
{
    new THexEdit(NULL);
}
//---------------------------------------------------------------------------
namespace Hexedit
{
void __fastcall PACKAGE Register()
	{
		 TComponentClass classes[1] = {__classid(THexEdit)};
		 RegisterComponents("Hex", classes, 0);
	}
}


//---------------------------------------------------------------------------

__fastcall THexEdit::THexEdit(TComponent* Owner)
	: TCustomEdit(Owner)
{
	FIncrement = 1;
	FDigits = 1;
	FMinValue = 0;
	FMaxValue = 0x7FFFFF;
	Text = "0";
	Overwrite = true;
	OnChangeEventDisabled = false;
	ReadOnly = false;
    AutoSelect = false;
}


//---------------------------------------------------------------------------
void __fastcall THexEdit::KeyDown(Word &VirtualKeyCode,  TShiftState Shift)
{
	if(VirtualKeyCode == VK_UP) {
		Up(this);
		VirtualKeyCode = 0;
		}
		else
		if(VirtualKeyCode == VK_DOWN) {
			Down(this);
			VirtualKeyCode = 0;
			}
			else
			if ((VirtualKeyCode == VK_INSERT) && !(Shift.Contains(ssCtrl) || Shift.Contains(ssShift) || Shift.Contains(ssAlt))) {
				FOverwrite = !FOverwrite;
				}
				else
				TCustomEdit::KeyDown(VirtualKeyCode, Shift);
}

void __fastcall THexEdit::KeyPress(Char& Key)
{
	if(Key == '\b') {
		if(FOverwrite) {
			SendMessage(Handle, WM_KEYDOWN, VK_LEFT, 0);       //Key in wParam
			Key = 0;
			return;
			}
		}
	if (!IsValidChar(Key)) {
		Key = 0;
		MessageBeep(0);
		return;
		}
//https://community.embarcadero.com/article/technical-articles/149-tools/13407-overwrite-in-tmemo-and-tedit
/*
The Windows TEdit and TMemo controls have no overwrite capability.
It is possible to emulate this behavior however, by setting the
SelLength property of the edit or memo control to one during the
processing of the KeyPress event. This causes the character at the
current position of the caret to be overwritten.
*/
if ((SelLength == 0) && FOverwrite) SelLength = 1; // emulate overwrite capability of TMemo
if (Key != 0) {
	if((Key >= 'a') && (Key <= 'f'))  Key = Key - 'a' + 'A';  //zamiana malych liter na wielkie
	if(Key) TCustomEdit::KeyPress(Key);
	}
}

bool __fastcall THexEdit::IsValidChar(Char Key)
{
	bool Result = false;
    if(((Key == ',' /*DecimalSeparator*/)       ||
        ((Key >= '0') && (Key <= '9'))  ||
        ((Key >= 'a') && (Key <= 'f'))  ||
		((Key >= 'A') && (Key <= 'F')))
		|| ((Key < 0x32) && (Key != Char(VK_RETURN)))
		)
		Result = True;

return Result;
}


void __fastcall THexEdit::Up(TObject *Sender)
{
    if(ReadOnly)
        MessageBeep(0);
		else
		Value += FIncrement;
}

void __fastcall THexEdit::Down(TObject *Sender)
{
	if(ReadOnly)
		MessageBeep(0);
		else
		Value -= FIncrement;
}

void __fastcall THexEdit::WMPaste(TWMNoParams &Message)
{
	if(ReadOnly) return;
}

void __fastcall THexEdit::WMCut(TWMNoParams &Message)
{
	if(ReadOnly) return;
}

void __fastcall THexEdit::CMExit(TWMNoParams &Message)
{
//    if(CheckValue (Value) != Value)
        SetValue (Value);
}

void __fastcall THexEdit::CMEnter(TWMNoParams &Message)
{
    if(AutoSelect && !(ControlState.Contains(csLButtonDown)))
        SelectAll();
}

unsigned long __fastcall THexEdit::GetValue(void)
{
    unsigned long Result;
    try {
        Result = AnsiString("0x"+Text).ToInt();
		}
	catch(...) {
		Value = FMinValue;
		return FMinValue;
		}
	return Result;
}

void __fastcall THexEdit::SetValue(unsigned long NewValue)
{
	int CursorPosition = SelStart;
	Text = IntToHex((__int64)CheckValue(NewValue), FDigits);
	SelStart = CursorPosition;
}

unsigned long __fastcall THexEdit::CheckValue(unsigned long NewValue)
{
	long Result;
	Result = NewValue;
	if (FMaxValue != FMinValue) {
		if (NewValue < FMinValue)
			Result = FMinValue;
		else if (NewValue > FMaxValue)
			Result = FMaxValue;
		}
	return Result;
}

void __fastcall THexEdit::SetDigits(int digits)
{
	FDigits = digits;
	long Temp = Value;
	Value = Temp;
}

void __fastcall THexEdit::SetMinValue(unsigned long value)
{
	FMinValue = value;
	if(Value < FMinValue) Value = FMinValue;
}

void __fastcall THexEdit::SetMaxValue(unsigned long value)
{
	FMaxValue = value;
	if(Value > FMaxValue) Value = FMaxValue;
}

void __fastcall THexEdit::Change(void)
{
	if(OnChangeEventDisabled) return; //nie obsluguj on change event
	TCustomEdit::Change();
}


