//---------------------------------------------------------------------------
/*
Remote Train Control Program for Windows

 Copyright 2015 by Mark DiVecchio

    This file is part of Remote Train Control.

    Remote Train Control is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Remote Train Control is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Remote Train Control.  If not, see <http://www.gnu.org/licenses/>.

*/
// Home Page: http://www.silogic.com/trains/ADPCM.html
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include <tchar.h>
#include <windows.h>
#include <mmsystem.h>
#include <stdio.h>
#include "assert.h"
#include "alloc.h"

#include "ADPCM_Player.h"
#include "ADPCM_Info.h"
#include "Quilling_Whistle.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TQuill_Index *Quill_Index;

extern bool ParseQuillIndex(QIndex &Index, HANDLE hParseFile, int index);
extern bool WriteQuillIndex(QIndex &Index, HANDLE hParseFile, int index);

extern SRSetting SR[6]; // structure containing the sample rates in various forms
extern bool Stopem;

//---------------------------------------------------------------------------
__fastcall TQuill_Index::TQuill_Index(TComponent* Owner)
	: TForm(Owner)
{
assert(false);	// never called
}
//---------------------------------------------------------------------------
__fastcall TQuill_Index::TQuill_Index(TComponent* Owner, String FN, int clip)
	: TForm(Owner)
{
ClipNo = clip;
assert(ClipNo > 0 && ClipNo < 256);
FileEdit->Text = FN;
assert(FN != "");
IndexFormat = false;

StringGrid1->Cells[QCLIPNO][0] = "Quill Clip #";
StringGrid1->Cells[QSTARTADDR][0] = "Start";
StringGrid1->Cells[QSMINIT][0] = "SM Init";	// initial values for IMA ADPCM state machine when playing the Repeat Segment
										/// 0xYYZZZZ : YY=step, ZZZZ=predicted value
StringGrid1->Cells[QENDADDR][0] = "End";
StringGrid1->Cells[QLENGTH][0] = "Length";
StringGrid1->Cells[QDATAWORD][0] = "Data";

StringGrid1->RowCount =  (0x3F00 - 0x3200) / 16;
// plus one for the Title Row
StringGrid1->RowCount += 1;

FileEdit->ShowHint = true;
FileEdit->Hint = FileEdit->Text;
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::ShowIt(String FN, int clip)
{
ClipNo = clip;
assert(ClipNo > 0 && ClipNo < 256);
FileEdit->Text = FN;
assert(FN != "");
FileEdit->Hint = FileEdit->Text;
Show();
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::ClearButtonClick(TObject *Sender)
{
ExportOneCliptomp3->Enabled = false;
PlayClipButton->Enabled = false;
ClearStringGrid();	// sets Decoded false
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::ClearStringGrid(void)
{
Decoded = false;    // signal sound file not analyzed
for(int x = StringGrid1->FixedRows; x < StringGrid1->RowCount; ++x)
	StringGrid1->Rows[x]->Clear();
StringGrid1->Repaint();
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::ExitButtonClick(TObject *Sender)
{
Close();
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::AnalyzeButtonClick(TObject *Sender)
{
HANDLE hAnalyzeFile;/* file handle */
/*
 * try and open the file
 */
if((hAnalyzeFile = CreateFile(FileEdit->Text.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	ADPCM_Main->MyShowMessage("Analyze: Unable to open file: " + FileEdit->Text, true);
	return;
	} else {
	ADPCM_Main->MyShowMessage("Analyze: Analyzing file: " + FileEdit->Text, false);
	}

/*
00003200  00 01 17 2B 20 F5 3E 00 20 F5 A1 80 00 00 00 00  ...+ >. ....
					  --------    --------   <--- Actual Sound file byte addresss
												 --------  <--- State Machine Initialization Value
00003210  23 02 06 38 20 F5 A2 00 21 05 73 80 00 00 00 00  #..8 .!.s....
							   --          -- --  <--- Constant
00003220  23 03 06 38 21 05 74 00 21 0C 82 80 00 2D FF 01  #..8!.t.!..-.
          -----------  <--- Unknown function
00003230  23 04 06 38 21 0C 83 00 21 18 28 80 00 30 FC B1  #..8!..!.(.0
00003240  23 05 06 38 21 18 29 00 21 24 C7 80 00 36 FE 6B  #..8!.).!$ǀ.6k
00003250  23 02 06 38 21 24 C8 00 21 2B DC 80 00 2F FF 84  #..8!$.!+܀./
00003260  2F 22 07 0F 21 2B DD 00 21 38 A0 80 00 33 FC 9A  /"..!+.!8.3
00003270  2F 22 08 0F 21 38 A1 00 21 41 FD 80 00 39 FD ED  /"..!8.!A.9
00003280  2F 22 09 09 21 41 FE 00 21 4C 75 80 00 35 FC 6E  /"..!A.!Lu.5n
00003290  2F 0A 0A 0F 21 4C 76 00 21 50 16 80 00 32 FD 70  /...!Lv.!P..2p
000032A0  2F 22 0B 0B 21 50 17 00 21 56 E0 80 00 3E F7 63  /"..!P..!V.>c
000032B0  2F 0C 0C 0F 21 56 E1 00 21 5D D7 80 00 3E FB 5D  /...!V.!]׀.>]
000032C0  2F 22 0D 0D 21 5D D8 00 21 65 2C 80 00 3C FA CF  /"..!].!e,.<
000032D0  2F 0E 0E 0F 21 65 2D 00 21 6E 49 80 00 3D FB 4C  /...!e-.!nI.=L
000032E0  2F 22 07 0F 21 6E 4A 00 21 72 D4 80 00 39 00 2D  /"..!nJ.!rԀ.9.-
000032F0  17 3A 2E 10 21 72 D5 00 21 81 86 80 00 38 00 4B  .:..!r.!..8.K
00003300  17 3A 2E 11 21 82 35 80 21 89 04 80 00 3B FA A3  .:..!5!..;
00003310  17 3A 2E 12 21 89 05 00 21 90 BF 80 00 3C FD A1  .:..!..!..<
00003320  13 3A 2E 13 21 90 C0 00 21 98 2F 80 00 43 FD 6B  .:..!..!/.Ck
*/

// empty String Grid
ClearStringGrid();

// Fill String Grid
QIndex LocalQIndex;

// Is there Quill Data there?
if (!ParseQuillIndex(LocalQIndex, hAnalyzeFile, 1)) {
	CloseHandle(hAnalyzeFile);
	return;	// error parsing index
	} else {
	if (LocalQIndex.StartAddr == 0x000000 || LocalQIndex.StartAddr == 0x00FFFFFF) {
		ADPCM_Main->MyShowMessage("Analyze: No Quilling Whistle Data in this sound file", true);
		CloseHandle(hAnalyzeFile);
		return;
		}
	}

// Quilling data is there, read in the index

for (int indexclipNo = 1 ; indexclipNo <= 208 ; indexclipNo++) {	// goes through the index : 1 based
	// zeorth Column of data is the Clip Number or the offset in Hex
	StringGrid1->Cells[QCLIPNO][indexclipNo] = HEXFIELDPREFIX + IntToHex(0x3200 + ((indexclipNo - 1) * 16), QOFFSETDIGITSTODISPLAY);
	// read and decode 16 bytes
	if (!ParseQuillIndex(LocalQIndex, hAnalyzeFile, indexclipNo)) {
		break;	// error parsing index
		}
	// Second column of data - Starting address of clip
	if (LocalQIndex.StartAddr == 0x000000 || LocalQIndex.StartAddr == 0x00FFFFFF) {
		StringGrid1->RowCount = indexclipNo;
		break;	// done
		}
	StringGrid1->Cells[QSTARTADDR][indexclipNo] = HEXFIELDPREFIX + IntToHex(LocalQIndex.StartAddr, QDIGITSTODISPLAY);

	if (indexclipNo == 1) ADPCM_Main->DebugBuffer->Lines->Append("Analyze: Start of Quill " +  StringGrid1->Cells[QSTARTADDR][indexclipNo]);

	// First column of data - First 4 bytes in the index
	StringGrid1->Cells[QDATAWORD][indexclipNo] = HEXFIELDPREFIX + IntToHex(LocalQIndex.dataword, QDATADIGITSTODISPLAY);


	// Fourth column of data - IMA ADPCM State Machine Initialization
	StringGrid1->Cells[QSMINIT][indexclipNo] = HEXFIELDPREFIX + IntToHex(LocalQIndex.SMInit,6);

	// THird column of data - Ending address of clip
	if (LocalQIndex.EndAddr != 0x7FFFFF) {
		//int EndAddr = BufferToAddress(&buffer16[12]);
		StringGrid1->Cells[QENDADDR][indexclipNo] = HEXFIELDPREFIX + IntToHex(LocalQIndex.EndAddr, QDIGITSTODISPLAY);
		// Sixth column  - calculated length
		int Len = (LocalQIndex.EndAddr - LocalQIndex.StartAddr) + 1;
		StringGrid1->Cells[QLENGTH][indexclipNo] = Len;
		}
	//
	}

CloseHandle(hAnalyzeFile);
Decoded = true; 	// signal sound file analyzed
ExportOneCliptomp3->Enabled = true;
// display Clip # numbers
ToggleOffsetClipNoClick(Sender);
// set push buttons
StringGrid1Click(Sender);
// focus on the StringGrid
StringGrid1->SetFocus();
return;
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::StringGrid1FixedCellClick(TObject *Sender, int ACol,
          int ARow)
{
// This toggles the values in the first column from memory addresses to/from
// numbers.
if (ACol == 0 && ARow == 0) {
	ToggleOffsetClipNoClick(Sender);
	}
if (ARow != 0) {
	StringGrid1->Row = ARow;    // hightlight selected row
	}
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::ToggleOffsetClipNoClick(TObject *Sender)
{
if (StringGrid1->Cells[QCLIPNO][1] == 1) { // if its numbers, change to addresses
	StringGrid1->Cells[QCLIPNO][0] = "Offset";
	for (int Row = 1 ; Row <= 208 ; Row++){
		StringGrid1->Cells[QCLIPNO][Row] = HEXFIELDPREFIX + IntToHex(0x3200 + ((Row - 1) * 16), IOFFSETDIGITSTODISPLAY);
		};
	} else {
	// its addresses, change to numbers
	StringGrid1->Cells[QCLIPNO][0] = "Quill Clip #";
	for (int Row = 1 ; Row <= 255 ; Row++){
		StringGrid1->Cells[QCLIPNO][Row] = Row;
		}
	}
IndexFormat = !IndexFormat;
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::StringGrid1Click(TObject *Sender)
{
//
if (StopButton1->Enabled) return;	// do nothing if playing a clip

if (StringGrid1->Cells[QSTARTADDR][StringGrid1->Row] == "") {
	PlayClipButton->Enabled = false;
	} else {
	PlayClipButton->Enabled = true;
	}
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::StringGrid1DrawCell(TObject *Sender, int ACol, int ARow,
          TRect &Rect, TGridDrawState State)
{
// default to black on white
StringGrid1->Canvas->Font->Color = clBlack;	// Normal Black on White
StringGrid1->Canvas->Brush->Color = clWhite;
// Draw the contents of each cell
if (Decoded) {
	// if the cell is selected, highlight it
	if( State.Contains(gdSelected) ) {
		StringGrid1->Canvas->Font->Color = clWhite;
		StringGrid1->Canvas->Brush->Color = clBlue;		// White on Blue
		} else {
		StringGrid1->Canvas->Font->Color = clBlack;	// Normal Black on White
		StringGrid1->Canvas->Brush->Color = clWhite;
		}
	}
// fill the rectangle
StringGrid1->Canvas->FillRect(Rect);

// Center text
String StringToAlign = StringGrid1->Cells[ACol][ARow];
const int StringStartH = (Rect.Left+Rect.Right)/2 - StringGrid1->Canvas->TextWidth(StringToAlign)/2;
const int StringStartV = (Rect.Top+Rect.Bottom)/2 - StringGrid1->Canvas->TextHeight(StringToAlign)/2;
StringGrid1->Canvas->TextOut(StringStartH, StringStartV, StringToAlign);

StringGrid1->Canvas->Font->Color = clBlack;	// Normal Black on White
StringGrid1->Canvas->Brush->Color = clWhite;
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::PlayClipButtonClick(TObject *Sender)
{
// Play the clip
if (!Decoded) {
	ADPCM_Main->MyShowMessage("Press the Analyze button first", true);
	return;
	}
// play the entire clip
// (Start Address, End Address, index, previousValue)
short index = 0;
short previousValue = 0;
//
// This is for a case seen in some PS3 file. The State Machine Initialization values
// are used even for non-repeating clips. TBD This needs to be thought about more.
//
// pick out the IMA ADPCM state machine's index and previous value
// Use _tcstol for both BCB4 and CB10          see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
index = _tcstol(StringGrid1->Cells[QSMINIT][StringGrid1->Row].c_str(), NULL, 16) >> 16;
previousValue = _tcstol(StringGrid1->Cells[QSMINIT][StringGrid1->Row].c_str(), NULL, 16) & 0x0000FFFF;
if (index < -1 || index > 88) {
	ADPCM_Main->MyShowMessage("Play Quill Clip: IMA ADPCM index out of range: " + String(index), false);
	return;
	}
//
TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;
PauseButton->Enabled = true;
RestartButton->Enabled = true;
StopButton1->Enabled = true;
// Use _tcstol for both BCB4 and CB10          see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
const int StartAddr = _tcstol(StringGrid1->Cells[QSTARTADDR][StringGrid1->Row].c_str(), NULL, 16);
const int EndAddr = _tcstol(StringGrid1->Cells[QENDADDR][StringGrid1->Row].c_str(), NULL, 16);
//ADPCM_Main->PlayClip(StartAddr, 0, 0, EndAddr, 0, index, previousValue);
ADPCM_Main->PlayClip(0, StartAddr, EndAddr, -1, 0, index, previousValue);	// each Quill clip plays as if it were a repeating segment
//
Screen->Cursor = OldCursor;
PauseButton->Enabled = false;
RestartButton->Enabled = false;
StopButton1->Enabled = false;
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::StringGrid1DblClick(TObject *Sender)
{
TPoint pt;
GetCursorPos(&pt);  // current mouse position
int Column, Row;
pt = StringGrid1->ScreenToClient(pt);
StringGrid1->MouseToCell(pt.x, pt.y, Column, Row);
if (Row == 0) return; // ignore double clicks in fixed rows and columns
if (Column == 0) return;

if (PlayClipButton->Enabled) PlayClipButtonClick(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::StringGrid1KeyPress(TObject *Sender, System::WideChar &Key)

{
if (Key == VK_RETURN) {
	StringGrid1DblClick(Sender);
    }
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::StringGrid1SelectCell(TObject *Sender, int ACol, int ARow,
          bool &CanSelect)
{
if (ACol != 0 && ARow != 0) CanSelect = true;
}

//---------------------------------------------------------------------------

void __fastcall TQuill_Index::PauseButtonClick(TObject *Sender)
{
ADPCM_Main->PauseButtonClick(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::RestartButtonClick(TObject *Sender)
{
ADPCM_Main->RestartButtonClick(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TQuill_Index::AboutADPCMClick(TObject *Sender)
{
// Splash Window Information Test
Splash = new TSplash(Application);
Splash->Memo1->Lines->Strings[0] = "ADPCM - " + String(ADPCM_Main->VersionLabel->Caption);
Splash->Show();			// Show Splash Window
Splash->Update();		// Paint Splash Window
ADPCM_Main->SS = SPLASH_OFF; 		// Turn off future Splash Window
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::CopyRowtoClipboardClick(TObject *Sender)
{
if (!Decoded) {
	ADPCM_Main->MyShowMessage("Press the Analyze button first", true);
	return;
	}
//Copy the selected row of the clip index to the clipboard
ClipEdit->Text = StringGrid1->Rows[StringGrid1->Row]->CommaText;
ClipEdit->SelectAll();
ClipEdit->CopyToClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::CopyClipIndextoClipboard1Click(TObject *Sender)
{
if (!Decoded) {
	ADPCM_Main->MyShowMessage("Press the Analyze button first", true);
	return;
	}
// Copy the entire clip index to the clipboard
MemoBuffer->Visible = false;
MemoBuffer->Clear();
MemoBuffer->Lines->Append("\"" + FileEdit->Text + "\"");
for (int i = 0 ; i < StringGrid1->RowCount  ; i++) {
	MemoBuffer->Lines->Append(StringGrid1->Rows[i]->CommaText);
	}
MemoBuffer->SelectAll();
MemoBuffer->CopyToClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::ClearthisIndexClick(TObject *Sender)
{
// Clear this index entry - this is not useful since a 0x00 start address indicates the end of the Quill clips
// must have analyzed a sound file
if (!Decoded) {
	ADPCM_Main->MyShowMessage("Press the Analyze button first", true);
	return;
	}
const int clip = StringGrid1->Row;
// read the sound file index entry
HANDLE hSoundFile;
QIndex LocalQIndex;
// try and open the file
if((hSoundFile = CreateFile(FileEdit->Text.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	ADPCM_Main->MyShowMessage("Clear Quill Index: Unable to open file: " + FileEdit->Text, true);
	// cancel the operation
	return;
	} else {
	ADPCM_Main->MyShowMessage("Clear Quill Index: Sound file to be edited: " + FileEdit->Text, false);
	}
if (!ParseQuillIndex(LocalQIndex, hSoundFile, clip)) {
	ADPCM_Main->MyShowMessage("Clear Index: Parse Quill Index failed", true);
	CloseHandle(hSoundFile);
	return;
	}

ADPCM_Main->MyShowMessage("Clear Quill Index: Clear Quill Index #" + String(clip), false);

LocalQIndex.Clip = clip;
LocalQIndex.StartAddr = 0x000000;
LocalQIndex.SMInit = 0x000000;
LocalQIndex.EndAddr = 0x000000;
LocalQIndex.dataword = 0x00000000;

// write the sound file index entry
if (!WriteQuillIndex(LocalQIndex, hSoundFile, clip)) {
	ADPCM_Main->MyShowMessage("Clear Quill Index: Could not write quill index to sound file", true);
	}
CloseHandle(hSoundFile);
AnalyzeButtonClick(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::CopyIndexClick(TObject *Sender)
{
//Copy this index entry to another index
// must have analyzed a sound file
if (!Decoded) {
	ADPCM_Main->MyShowMessage("Press the Analyze button first", true);
	return;
	}
const int clipFrom = StringGrid1->Row;
String FromCaption = "Copy Index from #" + String(clipFrom);

// read the sound file index entry
HANDLE hSoundFile;
QIndex LocalQIndex;
// try and open the file
if((hSoundFile = CreateFile(FileEdit->Text.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	ADPCM_Main->MyShowMessage("Copy Quill Index: Unable to open file: " + FileEdit->Text, true);
	// cancel the operation
	return;
	} else {
	ADPCM_Main->MyShowMessage("Copy Quill Index: Sound file to be edited: " + FileEdit->Text, false);
	}
if (!ParseQuillIndex(LocalQIndex, hSoundFile, clipFrom)) {
	ADPCM_Main->MyShowMessage("Copy Quill Index: Parse Quill Index failed", true);
	CloseHandle(hSoundFile);
	return;
	}

String Answer = 0;
if(InputQuery(FromCaption, "Enter Destination Quill Index Number: ", Answer)) {
	const int clipTo = Answer.ToInt();
	ADPCM_Main->MyShowMessage("Copy Quill Index: Copy Index #" + String(clipFrom) + " to #" + String(clipTo), false);
	// write the sound file index entry
	if (!WriteQuillIndex(LocalQIndex, hSoundFile, clipTo)) {
		ADPCM_Main->MyShowMessage("Copy Quill Index: Could not write quill index to sound file", true);
		}
	}
CloseHandle(hSoundFile);
AnalyzeButtonClick(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::ClipInfoClick(TObject *Sender)
{
// must have analyzed a sound file
if (!Decoded) {
	ADPCM_Main->MyShowMessage("Press the Analyze button first", true);
	return;
	}
// Display Clip info depending on clip type
const int StartAddr = _tcstol(StringGrid1->Cells[QSTARTADDR][StringGrid1->Row].c_str(), NULL, 16);    // Start
if (StartAddr == 0 || StartAddr == 0x00FFFFFF) return;	// NOP
const int EndAddr = _tcstol(StringGrid1->Cells[QENDADDR][StringGrid1->Row].c_str(), NULL, 16);    // End
const int Length =  StringGrid1->Cells[QLENGTH][StringGrid1->Row].ToInt();    // Length
const int dataword =  _tcstol(StringGrid1->Cells[QDATAWORD][StringGrid1->Row].c_str(), NULL, 16);    // data word
//
// display filename
//
ADPCM_Main->MyShowMessage(FileEdit->Text, false);
const int index =  _tcstol(StringGrid1->Cells[QCLIPNO][StringGrid1->Row].c_str(), NULL, 16);
if (IndexFormat == true) {
	ADPCM_Main->MyShowMessage("Quill Clip # = " + String(index), false);
	} else {
	String MAS = HEXFIELDPREFIX + IntToHex((int)index, 6);
	ADPCM_Main->MyShowMessage("Quill Clip # = " + MAS, false);
	}

ADPCM_Main->MyShowMessage("StartAddr = " + String(StartAddr), false);
ADPCM_Main->MyShowMessage("EndAddr = " + String(EndAddr), false);
ADPCM_Main->MyShowMessage("Length = " + String(Length), false);
ADPCM_Main->MyShowMessage("Data Word = " + String(dataword), false);
ADPCM_Main->MyShowMessage(" ", false);
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::Export1Cliptomp3Click(TObject *Sender)
{
// Export one Clip
ExportOneClipFlag  = true;
ExportClip(StringGrid1->Row);
ExportOneClipFlag  = false;
}
//---------------------------------------------------------------------------
void __fastcall TQuill_Index::ExportClip(int clipNo)
{
// Export one clip
ADPCM_Main->RecordFlag = true;
// set index into StringGrid
StringGrid1->Row = clipNo;
// Use _tcstol for both BCB4 and CB10          see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
const int StartAddr = _tcstol(StringGrid1->Cells[QSTARTADDR][StringGrid1->Row].c_str(), NULL, 16);
const int EndAddr = _tcstol(StringGrid1->Cells[QENDADDR][StringGrid1->Row].c_str(), NULL, 16);
if (StartAddr == 0x00 || StartAddr == 0x00FFFFFF) {
	ADPCM_Main->RecordFlag = false;
	return;	// empty index slot
	}
// create temporary filename
// create file
String NewExtension =  "_" + FormatFloat("000", clipNo) + "_Quill_16bit_little_endian_PCM_" + SR[ADPCM_Main->SampleGroup->ItemIndex].Hz + ".raw";
String RecordFileName = ChangeFileExt(FileEdit->Text, NewExtension);

// Use _tcstol for both BCB4 and CB10         see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
const short index = _tcstol(StringGrid1->Cells[QSMINIT][StringGrid1->Row].c_str(), NULL, 16) >> 16;
const short previousValue = _tcstol(StringGrid1->Cells[QSMINIT][StringGrid1->Row].c_str(), NULL, 16) & 0x0000FFFF;

// open temporary output file which will be created as 16 bit little endian signed PCM
if((ADPCM_Main->hRawFile = CreateFile(RecordFileName.c_str(), GENERIC_WRITE, NULL, NULL, CREATE_ALWAYS, 0, NULL)) == INVALID_HANDLE_VALUE) {
	ADPCM_Main->MyShowMessage("Export Quill Clip: Unable to open file: " + RecordFileName, true);
	ADPCM_Main->RecordFlag = false;
	return;
	} else {
	ADPCM_Main->MyShowMessage("Export Quill Clip: Exporting to file: " + RecordFileName, false);
	}
// play it and write the output to a file (always use SMInit)
// (Start Address, Start Repeat Address, End Repeat Address, End Address, state.index, state.previousValue)
ADPCM_Main->PlayClip(0, StartAddr, EndAddr, -1, 0, index, previousValue);

// close temporary output file
CloseHandle(ADPCM_Main->hRawFile);

// Build Command Line
NewExtension = "_" + FormatFloat("000", clipNo) + "_Quill.mp3";
String CmdLine = "\"" + String(ADPCM_Main->cwd) + "\\lame.exe\"" ;
String OutputFileName = ChangeFileExt(FileEdit->Text, NewExtension);
// Adding -V 0 creates a larger output file with not much impovement in sound
// output for audio at 5512 Hz is always upsampled to 8000 Hz,
// output at & over 8000 Hz is sampled at its actual rate
// use --resample 22.05  to get higher sample rate for all clips (makes much larger files)
String ParamLine = " --silent -r -s " + String(SR[ADPCM_Main->SampleGroup->ItemIndex].Kz) + " --bitwidth 16 --signed -m m " +
	"\"" + RecordFileName + "\" \"" + OutputFileName + "\"";

ADPCM_Main->MyShowMessage("Export Quill Clip: Executing: " + CmdLine + " " + ParamLine, false);

// delete output file if present
if (FileExists(OutputFileName))
	DeleteFile(OutputFileName);

// working directory
String WD = ExtractFilePath(FileEdit->Text);

// run lame.exe
SHELLEXECUTEINFO ShExecInfo = {0};
ShExecInfo.cbSize = sizeof(SHELLEXECUTEINFO);
ShExecInfo.fMask = SEE_MASK_NOCLOSEPROCESS;
ShExecInfo.hwnd = NULL;
ShExecInfo.lpVerb = NULL;
ShExecInfo.lpFile = CmdLine.c_str();
ShExecInfo.lpParameters = ParamLine.c_str();
ShExecInfo.lpDirectory = WD.c_str();
ShExecInfo.nShow = SW_HIDE;
ShExecInfo.hInstApp = NULL;

ShellExecuteEx(&ShExecInfo);

WaitForSingleObject(ShExecInfo.hProcess,INFINITE);

ADPCM_Main->RecordFlag = false;
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::ExportAllClipstomp3Click(TObject *Sender)
{
// export all clips to mp3 files
/*
Why does LAME add silence to the beginning each song?

DECODER DELAY AT START OF FILE:

All decoders I have tested introduce a delay of 528 samples. That is, after
decoding an mp3 file, the output will have 528 samples of 0's appended to the
front. This is because the standard MDCT/filterbank routines used by the ISO
have a 528 sample delay. It would be possible to write a MDCT/filterbank routine
with a 0 sample delay (see description of Takehiro's MDCT/filterbank routine
used in LAME encoding below) but I dont know that anyone has done this.
Furthermore, because of the overlapped nature of MDCT frames, the first half of
the first granule (1 granule=576 samples) doesn't have a previous frame to
overlap with, resulting in attenuation of the first N samples. The value of N
depends on the window type. For "STOP_TYPE" and "SHORT_TYPE", N=96, while for
"START_TYPE" and "NORMAL_TYPE", N=288. The first frame produced by LAME 3.56
and up will always be of STOP_TYPE or SHORT_TYPE.

ENCODER DELAY AT START OF FILE:

ISO based encoders (BladeEnc, 8hz-mp3, etc) use a MDCT/filterbank routine
similar to the one used in decoding, and thus also introduce their own 528
sample delay. A .wav file encoded & decoded will have a 1056 sample delay
(1056 samples will be appended to the beginning).
*/
// put them into the same directory as the .mth file

if (!FileExists(String(ADPCM_Main->cwd) + "\\lame.exe") || !FileExists(String(ADPCM_Main->cwd) + "\\lame_enc.dll")) {
	ADPCM_Main->MyShowMessage("Exporting requires that 'lame.exe' and 'lame_enc.dll' be in the ADPCM program directory. Download: http://www.rarewares.org/mp3-lame-bundle.php", true);
	return;
	}

// Loop through 255 sounds 1 to 256 (command "n1" through "n255")

Stopem = false; // press the STOP button to terminate loop early
const int LastClip = StringGrid1->RowCount;

for (int clip = 1 ; clip < LastClip ; clip++) {
	if (Stopem) {
		ADPCM_Main->RecordFlag = false;	// records as 16 bit little endian signed PCM
        return;		// exit
		}
	ExportClip(clip);  // Export
	}
ADPCM_Main->MyShowMessage("Export All Quill Clips: Complete", true);
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::StopButton1Click(TObject *Sender)
{
ADPCM_Main->StopButton2Click(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TQuill_Index::DumpEntireCliptoDiskClick(TObject *Sender)
{
// Dump clip to disk file - exact byte by byte dump. Works for both ADPCM and PCM clips
// created file is either 4 bit IMA-ADPCM or 24 bit big endian signed PCM
//
char * membuf;
DWORD readBytes, writtenBytes;
long fp;
// must have analyzed a sound file
if (!Decoded) {
	ADPCM_Main->MyShowMessage("Press the Analyze button first", true);
	return;
	}
const int clip = StringGrid1->Row;
const int StartAddr = _tcstol(StringGrid1->Cells[QSTARTADDR][clip].c_str(), NULL, 16);    // Start
DWORD Length = StringGrid1->Cells[QLENGTH][clip].ToInt();    // Length
const int databyte = _tcstol(ADPCM_Main->StringGrid1->Cells[IDATABYTE][ClipNo].c_str(), NULL, 16);    // data byte from Clip Index

if (databyte & PCM24MASK) {   // 24 pit PCM - length must be modulo 3
	if ((Length % 3) != 0) Length++;
	if ((Length % 3) != 0) Length++;
	}

HANDLE hSoundFile, hRawFile;
// try and open the file
if((hSoundFile = CreateFile(FileEdit->Text.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	ADPCM_Main->MyShowLastError("Dump Quill Clip: Unable to open file: " + FileEdit->Text, true);
	// cancel the operation
	return;
	} else {
	ADPCM_Main->MyShowMessage("Dump Quill Clip: Sound file to be processed: " + FileEdit->Text, false);
	}

// allocate memory of the correct size for a buffer
membuf = (char *) malloc(Length);
if (membuf == NULL) {
	ADPCM_Main->MyShowMessage("Dump Quill Clip: membuf malloc failed", true);
	CloseHandle(hSoundFile);
	return;
	}
// read from sound file from StartAddr for Length
fp = StartAddr;
SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
// read length bytes
ReadFile(hSoundFile, membuf, Length, &readBytes, NULL);
if (readBytes != Length) {
	ADPCM_Main->MyShowMessage("Dump Quill Clip: Could not read clip from sound file", true);
	CloseHandle(hSoundFile);
	free(membuf);
	return;
	}
CloseHandle(hSoundFile);
// write to default named disk file
// create file
String NewExtension = "_" + FormatFloat("000", clip) + "_" + SR[ADPCM_Main->SampleGroup->ItemIndex].Hz + ((databyte & PCM24MASK)? "_quill_be.raw": "_quill.ima");
String RecordFileName = ChangeFileExt(FileEdit->Text, NewExtension);

// open output file
if((ADPCM_Main->hRawFile = CreateFile(RecordFileName.c_str(), GENERIC_WRITE, NULL, NULL, CREATE_ALWAYS, 0, NULL)) == INVALID_HANDLE_VALUE) {
	ADPCM_Main->MyShowLastError("Dump Quill Clip: Unable to open file: " + RecordFileName, true);
	return;
	} else {
	ADPCM_Main->MyShowMessage("Dump Quill Clip: Dumping to file: " + RecordFileName, false);
	}
// write to disk file
SetFilePointer(hSoundFile, 0L, NULL, FILE_BEGIN);
SetEndOfFile(ADPCM_Main->hRawFile);
// write length bytes
WriteFile(ADPCM_Main->hRawFile, membuf, Length, &writtenBytes, NULL);
if (writtenBytes != Length) {
	ADPCM_Main->MyShowMessage("Dump Quill Clip: Could not write clip to sound file", true);
	CloseHandle(ADPCM_Main->hRawFile);
	free (membuf);
	return;
	}

CloseHandle(ADPCM_Main->hRawFile);
free(membuf);
}
//---------------------------------------------------------------------------



