#pragma once

#include <Windows.h>
#include <assert.h>
#include <conio.h>
#include <MMSystem.h>

struct MDT_PCM_WAV_HEADER
{
	MDT_PCM_WAV_HEADER()
	{
		//initialize what we can by default
		RIFF[0] = 'R';
		RIFF[1] = 'I';
		RIFF[2] = 'F';
		RIFF[3] = 'F';

		WAVE[0] = 'W';
		WAVE[1] = 'A';
		WAVE[2] = 'V';
		WAVE[3] = 'E';

		fmt[0] = 'f';
		fmt[1] = 'm';
		fmt[2] = 't';
		fmt[3] = ' ';
		fmtSize = 16;	//it's PCM

		pcmfmt.wf.wFormatTag = WAVE_FORMAT_PCM;

		data[0] = 'd';
		data[1] = 'a';
		data[2] = 't';
		data[3] = 'a';

		//the rest of the fields will be completed as needed
	}

	//header
	char RIFF[4];	//here should be "RIFF"
	DWORD dwChunkSize;	//here should be filesize-8
	char WAVE[4];	//here should be "WAVE"

	//the "fmt " chunk
	char fmt[4];	//here should be "fmt "
	DWORD fmtSize;	//the size of the "fmt " chunk (should be 16 for PCM wavs)
	PCMWAVEFORMAT pcmfmt;

	//the "data" subchunk
	char data[4];	//here should be "data"
	DWORD dataSize;	//the size of the "data" chunk = NumSamples * NumChannels * BitsPerSample/8
					//also, it's the filesize-sizeof(MDT_PCM_WAV_HEADER)
};

const int VOX_DECODE_BUFFER_SIZE = 64*1024;
const int VOX_NR_STEPS = 49;
const short VOX_STEP_SIZE[VOX_NR_STEPS] = 
{
	16, 17, 19, 21, 23, 25, 28, 31, 34, 37,
	41, 45, 50, 55, 60, 66, 73, 80, 88, 97,
	107, 118, 130, 143, 157, 173, 190, 209, 230, 253,
	279, 307, 337, 371, 408, 449, 494, 544, 598, 658,
	724, 796, 876, 963, 1060, 1166, 1282, 1411, 1552,
};
const short VOX_STEP_ADJUST[8] = 
{
	-1,
	-1,
	-1,
	-1,
	2,
	4,
	6,
	8,
};

//THIS IS NOT MULTITHREADED!!!!!
class CVox2Wav
{
public:
	CVox2Wav(void);
	~CVox2Wav(void);

	BOOL Convert(LPCTSTR szVoxFileName, LPCTSTR szWavFileName, int nSamplesPerSec, int nWavBitsPerSample);
	void GetLastError(LPTSTR szErrorBuff, int nErrorBuffLen);

protected:
	TCHAR*	m_szLastError;

	MDT_PCM_WAV_HEADER m_sWavHeader;

	int	m_nCurrentStep;
	short	m_nLastSample;

	//Methods
	short Decode(BYTE nSample);
	BOOL WriteDecodedSamples(HANDLE hWav, int nWavBitsPerSample, short* pSamples, int nrSamples);
};

